// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "base.h"
#include "fitsimage.h"
#include "blt.h"
#include "ps.h"
#include "tkpostscript.h"

// Debug
int DebugAST= 0;
int DebugBin= 0;
int DebugCompress= 0;
int DebugGZ= 0;
int DebugMosaic= 0;
int DebugPerf= 0;
int DebugRGB= 0;
int DebugWCS= 0;

// Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer frFlexLexer
#include <FlexLexer.h>

#include "parser.H"

YYSTYPE* frlval;
extern int frparse(Base*, frFlexLexer*);

int frlex(void* vval, frFlexLexer* ll)
{
  frlval = (YYSTYPE*)vval;
  return ll ? ll->yylex() : 0;
}

void frerror(Base* fr, frFlexLexer* ll, const char* m)
{
  fr->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    fr->error(": ");
    fr->error(cmd);
  }
}

// Base Member Functions

Base::Base(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : Widget(i, c, item)
{
  baseXImage = NULL;
  basePixmap = 0;
  needsUpdate = NOUPDATE;

  // these should all be set in the derived classes
  keyFits_ = NULL;
  channelFits_ = NULL;
  currentFits_ = NULL;

  currentMosaicCount = NULL;
  currentSliceCount = NULL;

  currentScale = NULL;
  currentMosaic = NULL;
  currentMosaicSystem = NULL;

  orientation = NORMAL;
  zoom_ = Vector(1,1);
  rotation = 0;

  rotateRotation = 0;

  preservePan = 0;

  slice = 1;

  panPM = 0;
  panGCXOR = NULL;
  pannerPixmap = 0;
  pannerXImage = NULL;
  pannerWidth = 0;
  pannerHeight = 0;
  pannerName[0] = '\0';
  pannerSystem = WCS;
  pannerSky = FK5;
  usePanner = 0;

  magnifierPixmap = 0;
  magnifierXImage = NULL;
  magnifierWidth = 0;
  magnifierHeight = 0;
  magnifierZoom_ = 4;
  magnifierName[0] = '\0';
  useMagnifier = 0;
  useMagnifierGraphics = 1;
  useMagnifierCursor = 1;
  magnifierGC = NULL;

  wcsAlignment = 0;
  wcsSystem = WCS;
  wcsSky = FK5;
  wcsOrientation = NORMAL;
  wcsRotation = 0;
  irafOrientation = (Orientation)-1;

  mask.deleteAll();
  maskColorName = dupstr("red");
  maskAlpha = 1;
  maskMark = 1;

  invert = 0;

  binFunction_ = FitsHist::SUM;
  binFactor_ = Vector(1,1);
  binBufferSize_ = 1024;
  binDepth_ = 1;

  doSmooth_ = 0;
  smoothFunction_ = FitsImage::GAUSSIAN;
  smoothRadius_ = 3;

  highliteGC = NULL;
  useHighlite = 0;

  useCrosshair = 0;
  crosshairGC = NULL;
  crosshairGCXOR = NULL;

  showMarkers = 1;
  showMarkersText = 1;
  centroidAuto = 0;
  centroidIteration = 5;
  centroidRadius = 10;
  preserveMarkers = 0;

  markerGC = NULL;
  markerGCXOR = NULL;
  selectGCXOR = NULL;

  grid = NULL;
  gridGC = NULL;

  contourLevel = NULL;
  currentcontour = NULL;
  contourGC = NULL;

  bgColorName = dupstr("white");
  bgColor = getXColor("white");
  memset(bgTrueColor_,255,4);

  nanColorName = dupstr("white");
  nanColor = getXColor("white");
  memset(nanTrueColor_,255,4);

  imageToData = Translate(-.5, -.5);
  dataToImage = Translate( .5,  .5);

  imageToData3d = Translate3d(-.5, -.5, -.5);
  dataToImage3d = Translate3d( .5,  .5, .5);
}

Base::~Base()
{
  if (basePixmap)
    Tk_FreePixmap(display, basePixmap);

  if (baseXImage)
    XDestroyImage(baseXImage);

  if (pannerPixmap)
    Tk_FreePixmap(display, pannerPixmap);

  if (pannerXImage)
    XDestroyImage(pannerXImage);

  if (magnifierPixmap)
    Tk_FreePixmap(display, magnifierPixmap);

  if (magnifierXImage)
    XDestroyImage(magnifierXImage);

  if (magnifierGC)
    XFreeGC(display, magnifierGC);

  if (highliteGC)
    XFreeGC(display, highliteGC);

  if (maskColorName)
    delete [] maskColorName;

  if (crosshairGC)
    XFreeGC(display, crosshairGC);

  if (crosshairGCXOR)
    XFreeGC(display, crosshairGCXOR);

  if (markerGC)
    XFreeGC(display, markerGC);

  if (markerGCXOR)
    XFreeGC(display, markerGCXOR);

  if (selectGCXOR)
    XFreeGC(display, selectGCXOR);

  if (grid)
    delete grid;

  if (gridGC)
    XFreeGC(display, gridGC);

  if (contourLevel)
    delete [] contourLevel;

  if (contourGC)
    XFreeGC(display, contourGC);

  if (bgColorName)
    delete [] bgColorName;

  if (nanColorName)
    delete [] nanColorName;
}

void Base::analysisFits()
{
  {
    FitsImage* ptr = *channelFits_;
    while (ptr) {
      FitsImage* sptr = ptr;
      while (sptr) {
	sptr->analysis();
	sptr = sptr->nextSlice();
      }
      ptr = ptr->nextMosaic();
    }
  }

  resetClip(1);
}

void Base::bltHist(char* xname, char* yname)
{
  int num = 256;

  if (!currentFits_)
    return;

  if (DebugPerf)
    cerr << "bltHist...";

  double* histogramX = currentScale->histogramX();
  double* histogramY = currentScale->histogramY();

  if (!histogramX || !histogramY) {
    histogramX = currentScale->initHistogramX(num);
    histogramY = currentScale->initHistogramY(num);

    for (int i=0; i<num; i++)
      histogramX[i] = (double)i/(num-1)*
	(currentScale->max()-currentScale->min()) + currentScale->min();
    
    for (int j=0; j<num; j++)
      histogramY[j] = 0;

    if (currentScale->max() > currentScale->min()) {
      FitsImage* ptr = *channelFits_;
      while (ptr) {
	FitsImage* sptr = ptr;
	while (sptr) {
	  sptr->bin(histogramY, num, currentScale->min(), currentScale->max(), 
		    currentScale->scanMode());
	  sptr = sptr->nextSlice();
	}
	ptr = ptr->nextMosaic();
      }
    }
  }

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, histogramX, num, num*sizeof(double), TCL_STATIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, histogramY, num, num*sizeof(double), TCL_STATIC) 
      != TCL_OK)
    goto error;

  if (DebugPerf)
    cerr << "end" << endl;

  return;

 error:
  result = TCL_ERROR;
  return;
}

void Base::calcAlignWCS(FitsImage* fits1, 
			CoordSystem sys, SkyFrame sky,
			Orientation* orientation, Matrix* oo,
			double* rotation)
{
  // init
  *orientation = NORMAL;
  oo->identity();
  *rotation = 0;

  if (!fits1 || !fits1->hasWCS(sys))
    return;

  // orientation
  *orientation = fits1->getWCSOrientation(sys,sky);
  switch (*orientation) {
  case NORMAL:
    oo->identity();
    break;
  case XX:
    *oo = FlipX();
    break;
  }

  // rotation
  *rotation = fits1->getWCSRotation(sys,sky);
}

void Base::calcAlignWCS(FitsImage* fits1, FitsImage* fits2, 
			CoordSystem sys, SkyFrame sky,
			Orientation* orientation, Matrix* oo,
			double* rotation, Vector* zoom)
{
  // init
  *orientation = NORMAL;
  oo->identity();
  *rotation = 0;

  if ((!fits1 || !fits2) || 
      (fits1 == fits2) ||
      !(fits1->hasWCS(sys)) ||
      !(fits2->hasWCS(sys)))
    return;

  Vector org1 = fits1->getWCScrpix(sys);
  Vector orval1 = fits1->pix2wcs(org1,sys,sky);

  // orientation
  *orientation = fits2->getWCSOrientation(sys,sky);
  switch (*orientation) {
  case NORMAL:
    oo->identity();
    break;
  case XX:
    *oo = FlipX();
    break;
  }

  // rotation
  Vector orpix = fits2->wcs2pix(orval1, sys, sky);
  Vector delta = fits2->getWCScdelt(sys).abs();
  Vector npix = 
    fits2->wcs2pix(Vector(orval1[0],orval1[1]+delta[1]),sys,sky);
  Vector north = npix - orpix;

  if ((fits2->getWCS(sys))->imflip)
    *rotation = (north.angle()-M_PI/2);
  else
    *rotation = -(north.angle()-M_PI/2);

  // zoom
  Vector cd1 = fits1->getWCScdelt(sys);
  Vector cd2 = fits2->getWCScdelt(sys);
  *zoom = Vector((*zoom)[0]/fabs(cd1[0]/cd2[0]), 
		 (*zoom)[1]/fabs(cd1[1]/cd2[1]));
}

Matrix Base::calcAlignWCS(FitsImage* fits1, FitsImage* fits2, 
			  CoordSystem sys, SkyFrame sky)
{
  if ((!fits1 || !fits2) || 
      (fits1 == fits2) ||
      !(fits1->hasWCS(sys)) ||
      !(fits2->hasWCS(sys)))
    return Matrix();

  Vector org1 = fits1->getWCScrpix(sys);
  Vector orval1 = fits1->pix2wcs(org1,sys,sky);

  Vector org2 = fits2->getWCScrpix(sys);
  Vector orval2 = fits2->pix2wcs(org2,sys,sky);

  // orientation
  Orientation orientation =NORMAL;
  {
    Orientation o1 = fits1->getWCSOrientation(sys,sky);
    Orientation o2 = fits2->getWCSOrientation(sys,sky);

    switch (o1) {
    case NORMAL:
      {
	switch (o2) {
	case NORMAL:
	  orientation = NORMAL;
	  break;
	case XX:
	  orientation = XX;
	  break;
	}
      }
      break;
    case XX:
      {
	switch (o2) {
	case NORMAL:
	  orientation = XX;
	  break;
	case XX:
	  orientation = NORMAL;
	  break;
	}
      }
      break;
    }
  }

  // zoom
  Vector zoom =Vector(1,1);
  {
    Vector cd1 = fits1->getWCScdelt(sys);
    Vector cd2 = fits2->getWCScdelt(sys);
    zoom = Vector(cd2[0]/cd1[0], cd2[1]/cd1[1]).abs();
  }
  
  // rotation
  double rotation =0;
  {
    double r1 =0;
    {
      Vector orpix = fits1->wcs2pix(orval1, sys, sky);
      Vector delta = fits1->getWCScdelt(sys).abs();
      Vector npix = 
	fits1->wcs2pix(Vector(orval1[0],orval1[1]+delta[1]),sys,sky);
      Vector north = npix - orpix;

      if ((fits1->getWCS(sys))->imflip)
	r1 = (north.angle()-M_PI/2);
      else
	r1 = -(north.angle()-M_PI/2);
    }

    double r2 =0;
    {
      Vector orpix = fits2->wcs2pix(orval1, sys, sky);
      Vector delta = fits2->getWCScdelt(sys).abs();
      Vector npix = 
	fits2->wcs2pix(Vector(orval1[0],orval1[1]+delta[1]),sys,sky);
      Vector north = npix - orpix;

      if ((fits2->getWCS(sys))->imflip)
	r2 = (north.angle()-M_PI/2);
      else
	r2 = -(north.angle()-M_PI/2);
    }
    rotation = r1-r2;
  }

  // origin
  Vector origin;
  {
    Vector orpix1 = fits1->wcs2pix(orval2,sys,sky) * imageToData;
    Vector orpix2 = fits2->wcs2pix(orval2,sys,sky) * imageToData;
    origin = orpix1 - orpix2;
  }

  // matrix
  {
    // if the segment is flipped, we can have a discontinuity at
    // the edges, due to round off errors, so we 'nudge' it

    double rr = degToRad(180);
    if ((rotation > rr-FLT_EPSILON && rotation < rr+FLT_EPSILON) || 
	(-rotation > rr-FLT_EPSILON && -rotation < rr+FLT_EPSILON)) {
      switch (orientation) {
      case NORMAL:
	origin -= Vector(0,FLT_EPSILON);
	break;
      case XX:
	origin -= Vector(FLT_EPSILON,FLT_EPSILON);
	break;
      }
    }
    else {
      switch (orientation) {
      case NORMAL:
	break;
      case XX:
	origin -= Vector(FLT_EPSILON,0);
	break;
      }
    }

    Matrix flip;
    switch (orientation) {
    case NORMAL:
      break;
    case XX:
      flip = FlipX();
      break;
    }

    Vector orpix2 = fits2->wcs2pix(orval2,sys,sky) * imageToData;
    return
      Translate(-orpix2) *
      flip *
      Scale(zoom) *
      Rotate(rotation) *
      Translate(orpix2) *
      Translate(origin);
  }
}

void Base::calcHistEqu()
{
  // if we don't have any data, bail
  if (!*channelFits_) {
    currentScale->deleteHistequ();
    return;
  }

  // if we already have it, bail
  if (currentScale->histequ())
    return;

  if (DebugPerf)
    cerr << "calcHistEqu...";

  // create pdf or histogram
  double* pdf = new double[HISTEQUSIZE];
  memset(pdf,0,HISTEQUSIZE*sizeof(double));

  FitsImage* ptr = *channelFits_;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->bin(pdf, HISTEQUSIZE, currentScale->low(), currentScale->high(), 
		currentScale->scanMode());
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }

  // find a total/average
  double total, average;
  {
    total = 0;
    for (int i=0; i<HISTEQUSIZE; i++)
      total += pdf[i];
    average = total/HISTEQUSIZE;
  }

  // build transfer function (cdf)
  double* histequ = currentScale->initHistequ(HISTEQUSIZE);

  double bin = 0;
  int color,level;
  for (color=level=0; level<HISTEQUSIZE && color<HISTEQUSIZE; level++) {
    histequ[level] = (double)color/HISTEQUSIZE;
    bin += pdf[level];
    while (bin>=average && color<HISTEQUSIZE) {
      bin -= average;
      color++;
    }
  }
  while (level<HISTEQUSIZE)
    histequ[level++] = (double)(HISTEQUSIZE-1)/HISTEQUSIZE;

  delete [] pdf;

  if (DebugPerf)
    cerr << "end" << endl;
}

void Base::coordToTclArray(FitsImage* ptr, const Vector& v, 
			   InternalSystem in, CoordSystem out, 
			   Precision p, const char* var, const char* index)
{
  Vector r = ptr->mapFromRef(mapToRef(v,in), out);
  doubleToTclArray(r[0], p, var, index, "x");
  doubleToTclArray(r[1], p, var, index, "y");
}

void Base::doubleToTclArray(double d, Precision p, const char* var, 
			    const char* base, const char* mod)
{
  char index[64];
  strcpy(index,mod);
  strcat(index,base);

  ostringstream str;
  switch (p) {
  case DEFAULT:
    str << d << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) 
	<< d << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << d << ends;
    break;
  case INTEGER:
    str << (int)d << ends;
    break;
  }

  Tcl_SetVar2(interp, (char*)var, index, str.str().c_str(), 0);
}

Vector Base::getClip(FrScale::ClipMode cm, float ac)
{
  // save current scale
  FrScale* sav = currentScale;
  currentScale = new FrScale(*sav);

  // change values
  currentScale->setClipMode(cm);
  currentScale->setAutoCutPer(ac);
  updateClip(0);

  // this routine will change current settings, calculate, then reset
  double ll = currentScale->low();
  double hh = currentScale->high();

  // restore currentScale
  delete currentScale;
  currentScale = sav;
  updateClip(0);

  return Vector(ll,hh);
}

void Base::getInfo(const Vector& vv, InternalSystem ref,
		   Matrix& (FitsImage::*matrixToData)(),
		   Matrix& (FitsImage::*matrixToImage)(),
		   SkyFrame sky, SkyFormat format, char* var)
{
  // clear values
  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"value,red","",0);
  Tcl_SetVar2(interp,var,"value,green","",0);
  Tcl_SetVar2(interp,var,"value,blue","",0);

  FitsImage* sptr = currentFits_;
  int mosaic;
  FitsBound* params;

  // make sure we have an image
  if (!sptr)
    goto noFits;

  mosaic = isMosaic();

  params = sptr->getDataParams(currentScale->scanMode());

  if (!mosaic) {
    Tcl_SetVar2(interp,var,"filename",(char*)sptr->getRootBaseFileName(),0);
    Tcl_SetVar2(interp,var,"objectname",(char*)sptr->getObjectName(),0);
    Tcl_SetVar2(interp,var,"min",(char*)sptr->getMin(),0);
    Tcl_SetVar2(interp,var,"max",(char*)sptr->getMax(),0);
    Tcl_SetVar2(interp,var,"low",(char*)sptr->getLow(),0);
    Tcl_SetVar2(interp,var,"high",(char*)sptr->getHigh(),0);
  }

  do {
    Vector img = vv * (sptr->*matrixToData)();

    if (img[0]>=params->xmin && img[0]<params->xmax && 
	img[1]>=params->ymin && img[1]<params->ymax) {

      if (mosaic) {
	Tcl_SetVar2(interp,var,"filename",(char*)sptr->getRootBaseFileName(),0);
	Tcl_SetVar2(interp,var,"objectname",(char*)sptr->getObjectName(),0);
	Tcl_SetVar2(interp,var,"min",(char*)sptr->getMin(),0);
	Tcl_SetVar2(interp,var,"max",(char*)sptr->getMax(),0);
	Tcl_SetVar2(interp,var,"low",(char*)sptr->getLow(),0);
	Tcl_SetVar2(interp,var,"high",(char*)sptr->getHigh(),0);
      }

      Tcl_SetVar2(interp,var,"value",(char*)sptr->getValue(img),0);

      coordToTclArray(sptr,vv,ref,IMAGE,FIXED,var,"Image");
      coordToTclArray(sptr,vv,ref,PHYSICAL,FIXED,var,"Physical");

      if (hasATMV())
	coordToTclArray(sptr,vv,ref,AMPLIFIER,FIXED,var,"Amplifier");
      else {
	Tcl_SetVar2(interp,var,"xAmplifier","",0);
	Tcl_SetVar2(interp,var,"yAmplifier","",0);
      }

      if (hasDTMV())
	coordToTclArray(sptr,vv,ref,DETECTOR,FIXED,var,"Detector");
      else {
	Tcl_SetVar2(interp,var,"xDetector","",0);
	Tcl_SetVar2(interp,var,"yDetector","",0);
      }

      for (int i=0; i<MULTWCS; i++) {
	char buf[64];
	char w = !i ? NULL : '`'+i;
	CoordSystem www = (CoordSystem)(WCS+i);

	if (hasWCS(www)) {
	  char buff[128];
	  sptr->pix2wcs(vv*(sptr->*matrixToImage)(),www,sky,format,buff,128);

	  int argc;
	  const char** argv;
	  Tcl_SplitList(interp, buff, &argc, &argv);

	  if (argc > 0 && argv && argv[0])
	    Tcl_SetVar2(interp,var,varcat(buf,"xWCS,",w),argv[0],0);
	  else
	    Tcl_SetVar2(interp,var,varcat(buf,"xWCS,",w),"",0);

	  if (argc > 1 && argv && argv[1])
	    Tcl_SetVar2(interp,var,varcat(buf,"yWCS,",w),argv[1],0);
	  else
	    Tcl_SetVar2(interp,var,varcat(buf,"yWCS,",w),"",0);

	  char* wcsname = (char*)sptr->getWCSName(www);
	  if (wcsname)
	    Tcl_SetVar2(interp,var,varcat(buf,"sysWCS,",w),wcsname,0);
	  else if (argc > 2 && argv && argv[2])
	    Tcl_SetVar2(interp,var,varcat(buf,"sysWCS,",w),argv[2],0);
	  else
	    Tcl_SetVar2(interp,var,varcat(buf,"sysWCS,",w),"",0);
	    
	  if (argc > 3 && argv && argv[3])
	    Tcl_SetVar2(interp,var,varcat(buf,"xUnitWCS,",w),argv[3],0);
	  else
	    Tcl_SetVar2(interp,var,varcat(buf,"xUnitWCS,",w),"",0);

	  if (argc > 4 && argv && argv[4])
	    Tcl_SetVar2(interp,var,varcat(buf,"yUnitWCS,",w),argv[4],0);
	  else
	    Tcl_SetVar2(interp,var,varcat(buf,"yUnitWCS,",w),"",0);

	  Tcl_Free((char*)argv);
	}
	else {
	  Tcl_SetVar2(interp,var,varcat(buf,"xWCS,",w),"",0);
	  Tcl_SetVar2(interp,var,varcat(buf,"yWCS,",w),"",0);
	  Tcl_SetVar2(interp,var,varcat(buf,"sysWCS,",w),"",0);
	  Tcl_SetVar2(interp,var,varcat(buf,"xUnitWCS,",w),"",0);
	  Tcl_SetVar2(interp,var,varcat(buf,"yUnitWCS,",w),"",0);
	}
      }

      return;
    }
    else {
      if (mosaic) {
	sptr = sptr->nextMosaic();
	if (sptr)
	  params = sptr->getDataParams(currentScale->scanMode());
      }
      else
	goto noImage;
    }
  }
  while (mosaic && sptr);

  // else, return blanks
 noFits:
  Tcl_SetVar2(interp,var,"objectname","",0);
  Tcl_SetVar2(interp,var,"filename","",0);
  Tcl_SetVar2(interp,var,"min","",0);
  Tcl_SetVar2(interp,var,"max","",0);
  Tcl_SetVar2(interp,var,"low","",0);
  Tcl_SetVar2(interp,var,"high","",0);

 noImage:
  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"value,red","",0);
  Tcl_SetVar2(interp,var,"value,green","",0);
  Tcl_SetVar2(interp,var,"value,blue","",0);

  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

  for (int i=0; i<MULTWCS; i++) {
    char buf[64];
    char w = !i ? NULL : '`'+i;
    Tcl_SetVar2(interp,var,varcat(buf,"xWCS,",w),"",0);
    Tcl_SetVar2(interp,var,varcat(buf,"yWCS,",w),"",0);
    Tcl_SetVar2(interp,var,varcat(buf,"sysWCS,",w),"",0);
    Tcl_SetVar2(interp,var,varcat(buf,"xUnitWCS,",w),"",0);
    Tcl_SetVar2(interp,var,varcat(buf,"yUnitWCS,",w),"",0);
  }
}

Orientation Base::getWCSOrientation(CoordSystem sys, SkyFrame sky)
{
  return currentFits_ ? currentFits_->getWCSOrientation(sys, sky) : NORMAL;
}

double Base::getWCSRotation(CoordSystem sys, SkyFrame sky)
{
  return currentFits_ ? currentFits_->getWCSRotation(sys,sky) : 0;
}

int Base::hasATMV()
{
  return currentFits_ && currentFits_->hasATMV();
}

int Base::hasDTMV()
{
  return currentFits_ && currentFits_->hasDTMV();
}

int Base::hasLTMV()
{
  return currentFits_ && currentFits_->hasLTMV();
}

int Base::hasWCS(CoordSystem sys)
{ 
  return currentFits_ && currentFits_->hasWCS(sys);
}

int Base::hasWCSEqu(CoordSystem sys)
{ 
  return currentFits_ && currentFits_->hasWCSEqu(sys);
}

int Base::hasWCSLinear(CoordSystem sys)
{ 
  return currentFits_ && currentFits_->hasWCSLinear(sys);
}

InverseScale* Base::inverseScale(int cnt, double ll, double hh, 
				      FrScale::ColorScaleType type)
{
  switch (type) {
  case FrScale::IISSCALE:
  case FrScale::LINEARSCALE:
    return new LinearInverseScale(cnt, ll, hh);
  case FrScale::LOGSCALE:
    return new LogInverseScale(cnt, ll, hh);
  case FrScale::POWSCALE:
    return new PowInverseScale(cnt, ll, hh);
  case FrScale::SQRTSCALE:
    return new SqrtInverseScale(cnt, ll, hh);
  case FrScale::SQUAREDSCALE:
    return new SquaredInverseScale(cnt, ll, hh);
  case FrScale::HISTEQUSCALE:
    calcHistEqu();
    return new HistEquInverseScale(cnt, ll, hh, currentScale->histequ(),
				   HISTEQUSIZE);
  }
}

void Base::invalidPixmap()
{
  Widget::invalidPixmap();

  if (basePixmap)
    Tk_FreePixmap(display, basePixmap);
  basePixmap = 0;

  if (baseXImage)
    XDestroyImage(baseXImage);
  baseXImage = NULL;

  needsUpdate = MATRIX;
}

Orientation Base::IRAFOrientation(Orientation oo)
{
  if (irafOrientation == (Orientation)-1)
    irafOrientation = oo;
  return irafOrientation;
}

int Base::parse(istringstream& istr)
{
  result = TCL_OK;
  frFlexLexer* ll = new frFlexLexer(&istr);
  frparse(this, ll);
  delete ll;

  return result;
}

FitsHead* Base::parseWCS(istream& str)
{
  if (!currentFits_)
    return NULL;

  FitsHead* head = currentFits_->getHead();
  FitsHead* r = new FitsHead(head->naxis(1), head->naxis(2),
			     head->naxis(3), head->bitpix());

  while (!str.eof()) {
    char buf[256];
    str.getline(buf,256,'\n');

    if (strlen(buf) > 0) {
      string x(buf);
      istringstream sstr(x);

      char keyword[256];
      sstr >> keyword;

      if (strchr(buf,'=')) {
	char dummy;
	sstr >> dummy;
      }

      if (strchr(buf,'\'')) {
	char string[32];
	sstr >> string;

	// remove both '
	string[strlen(string)-1] = '\0';
	r->appendString(keyword, string+1, "");
      }
      else {
	double val;
	sstr >> val;
	r->appendReal(keyword, val, 9, "");
      }
    }
    else
      break;
  }

  return r;
}

void Base::printCoordSystem(CoordSystem sys)
{
  Tcl_AppendResult(interp, coordSystemStr_[sys-IMAGE], NULL);
}

void Base::printDouble(double d, Precision p)
{
  ostringstream str;
  switch (p) {
  case DEFAULT:
    str << d << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) << d << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << d  << ends;
    break;
  case INTEGER:
    str << (int)d << ends;
    break;
  }
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::printInteger(int i)
{
  ostringstream str;
  str << i << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::printSkyFrame(SkyFrame sky)
{
  Tcl_AppendResult(interp, skyFrameStr_[sky-FK4], NULL);
}

void Base::printSkyFormat(SkyFormat format)
{
  Tcl_AppendResult(interp, skyFormatStr_[format], NULL);
}

void Base::printVector(const Vector& v, Precision p)
{
  ostringstream str;
  switch (p) {
  case DEFAULT:
    str << setprecision(8) << v << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) << v << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << v << ends;
    break;
  case INTEGER:
    {
      Vector z = v;
      str << z.round() << ends;
      break;
    }
  }
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::ps(double* (FitsImage::*getMatrix)())

{
  TkPostscriptInfo* psInfo = (TkPostscriptInfo*)(((TkCanvas*)canvas)->psInfo);

  Vector org = psOrigin();

  // Image
  if (*channelFits_) {
    // save graphics state
    Tcl_AppendResult(interp, "gsave\n", NULL);

    double scale = psInfo->scale * psResolution / 72.;
    int width = (int)(options->width*scale);
    int height = (int)(options->height*scale);

    // scale
    {
      ostringstream str;
      str << org[0] << ' ' << org[1] << " translate" << endl
	  << setprecision(12)
	  << 1/scale << ' ' << 1/scale << " scale" << endl
	  << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
    }

    // header
    psHead(width, height);

    switch (psLevel) {
    case 1: 
      {
	NoCompressAsciiHex filter;
	psImage(filter, psColorSpace, width, height, scale, getMatrix);
      }
      break;
    case 2:
      {
	RLEAscii85 filter;
	psImage(filter, psColorSpace, width, height, scale, getMatrix);
      }
      break;
    }

    // restore gstate
    Tcl_AppendResult(interp, "grestore\n", NULL);
  }

  // Markers & Contours & Grids
  // clip path
  {
    ostringstream str;
    str << org[0] << ' ' << org[1] << ' ' 
    	<< options->width << ' ' << options->height
    	<< " rectclip" << endl << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::psImage(Filter& filter, PSColorSpace mode, 
		   int width, int height, float scale, 
		   double* (FitsImage::*getMatrix)())
{
  updatePS(scale, width, height);

  unsigned char* img = fillImage(width, height, getMatrix);
  if (!img)
    return;
  unsigned char* ptr = img;

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++, ptr+=3) {
      unsigned char red = ptr[0];
      unsigned char green = ptr[1];
      unsigned char blue = ptr[2];

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);

  if (img)
    delete [] img;
}

void Base::renderHighLite()
{
  if (useHighlite)
    XDrawRectangle(display, pixmap, highliteGC, 1, 1, 
		   options->width-2, options->height-2);
}

void Base::reset()
{
  invert = 0;
  update(MATRIX);
}

void Base::resetClip(int force)
{
  if (DebugPerf)
    cerr << "resetClip" << endl;

  updateClip(force);

  // init histogram
  currentScale->deleteHistogramX();
  currentScale->deleteHistogramY();
  currentScale->deleteHistequ();

  // always (for HISTEQU and LOG)
  updateColorScale();
}

void Base::setColorScale(FrScale::ColorScaleType s)
{
  currentScale->setColorScaleType(s);
  updateColorScale();
}

void Base::setScanModeIncr(LoadMethod lm)
{
  if (lm == INCR)
    switch (currentScale->scanMode()) {
    case FrScale::NODATASEC:
    case FrScale::UNODATASEC:
    case FrScale::RESET:
      currentScale->setScanMode(FrScale::UNODATASEC);
      break;
    case FrScale::DATASEC:
    case FrScale::UDATASEC:
      currentScale->setScanMode(FrScale::UDATASEC);
      break;
    }
}

void Base::unloadFits()
{
  irafOrientation = (Orientation)-1;

  mask.deleteAll();

  if (grid)
    delete grid;
  grid = NULL;

  // delete any auxcontour
  auxcontours.deleteAll();

  setCurrentFits(NULL);
  slice = 1;
  *currentMosaicCount = 0;
  *currentSliceCount = 0;
  *currentMosaic = NOMOSAIC;

  resetClip(0);
}

void Base::update(UpdateType flag)
{
  if (DebugPerf)
    cerr << "update" << endl;

  // Be careful, someone may have already set the flag at a lower level
  // therefor, only change the flag if we need more to be done

  if (flag < needsUpdate)
    needsUpdate = flag;
  redraw();
}

void Base::update(UpdateType flag, BBox bb)
{
  if (DebugPerf)
    cerr << "update" << bb << endl;

  // bb is in canvas coords

  if (flag < needsUpdate)
    needsUpdate = flag;
  redraw(bb);
}

void Base::updateBase()
{
  int& width = options->width;
  int& height = options->height;

  if (!basePixmap) {
    if (!(basePixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				    width, height, depth))){
      cerr << "Frame Internal Error: Unable to Create Pixmap" << endl;
      exit(1);
    }

    // Geometry has changed, redefine our marker GCs including clip regions
    updateGCs();
  }

  if (!baseXImage) {
    if (!(baseXImage = XGetImage(display, basePixmap, 0, 0, width, height, 
				 AllPlanes, ZPixmap))) {
      cerr << "Frame Internal Error: Unable to Create XImage" << endl;
      exit(1);
    }

    // we have to wait until now, since the encodings depend on baseXImage
    encodeTrueColor(bgColor, bgTrueColor_);
    encodeTrueColor(nanColor, nanTrueColor_);

    // we have a race condition. Some Truecolor ColorScales need to know the 
    // bytes per pixel, RGB masks, and byte order, from XImage struct.
    // yet, we receive colormap commands well before we render to the 
    // screen and have a valid XImage. 
    // So, we put off creating a colorscale until we are ready to render.

    if (!validColorScale())
      updateColorScale();
  }
}

void Base::updateBinFileNames()
{
  // we only want to do this to binn'd 3D cubes
  if (!(*channelFits_)->isHist())
    return;

  char* zcol = (char*)(*channelFits_)->getHistZ();
  int bd = (*channelFits_)->binDepth();
  if (bd>1 && zcol) {
    Vector zlim = (*channelFits_)->getHistColMinMax(zcol);
    double zlen = zlim[1]-zlim[0];
    double zdelta = zlen/bd;

    double zptr = zlim[0];
    FitsImage* ptr = *channelFits_;
    for (int i=0; i<(*channelFits_)->depth();i++,ptr=ptr->nextSlice(),zptr+=zdelta) {
      ostringstream str;
      str << zcol << ">=" << zptr << '&' << zcol << '<' << zptr+zdelta << ends;
      ptr->setBinSliceFilter(str.str().c_str());
      ptr->updateFileName();
    } 
  }
  else {
    (*channelFits_)->setBinSliceFilter(NULL);
    (*channelFits_)->updateFileName();
  }
}

void Base::updateClip(int force)
{
  if (DebugPerf)
    cerr << "updateClip" << endl;

  if (!currentScale->preserve() || force) {
    currentScale->setMin(DBL_MAX);
    currentScale->setMax(-DBL_MAX);

    currentScale->setLow(DBL_MAX);
    currentScale->setHigh(-DBL_MAX);

    if (*channelFits_) {
      // find min/max
      {
	FitsImage* ptr = *channelFits_;
	while (ptr) {
	  FitsImage* sptr = ptr;
	  while (sptr) {
	    sptr->updateClip(currentScale);

	    // find over-all min/max
	    if (currentScale->min() > sptr->getMinDouble())
	      currentScale->setMin(sptr->getMinDouble());
      
	    if (currentScale->max() <= sptr->getMaxDouble())
	      currentScale->setMax(sptr->getMaxDouble());

	    // find low/high
	    if (currentScale->low() > sptr->getLowDouble())
	      currentScale->setLow(sptr->getLowDouble());

	    if (currentScale->high() <= sptr->getHighDouble())
	      currentScale->setHigh(sptr->getHighDouble());

	    sptr = sptr->nextSlice();
	  }

	  ptr = ptr->nextMosaic();
	}
      }

      // set global
      {
	if ((isCube() || isMosaic()) && 
	    currentScale->clipScope() == FrScale::GLOBAL) {
	  FitsImage* ptr = *channelFits_;
	  while (ptr) {
	    FitsImage* sptr = ptr;
	    while (sptr) {
	      sptr->setClip(currentScale->low(), currentScale->high());
	      sptr = sptr->nextSlice();
	    }
	    ptr = ptr->nextMosaic();
	  }
	}
      }
    }
    else {
      if (currentScale->clipMode() != FrScale::USERCLIP) {
	currentScale->setLow(DEFAULTLOW);
	currentScale->setHigh(DEFAULTHIGH);
      }
      else {
	currentScale->setLow(currentScale->uLow());
	currentScale->setHigh(currentScale->uHigh());
      }
    }
  }
  else {
    FitsImage* ptr = *channelFits_;
    while (ptr) {
      FitsImage* sptr = ptr;
      while (sptr) {
	sptr->updateClip(currentScale);
	sptr->setClip(currentScale->low(), currentScale->high());
	sptr = sptr->nextSlice();
      }
      ptr = ptr->nextMosaic();
    }
  }
}

void Base::updateGCs()
{
  // widget clip region
  BBox bbWidget = BBox(0, 0, options->width, options->height);
  Vector sizeWidget = bbWidget.size();

  rectWidget[0].x = (int)bbWidget.ll[0];
  rectWidget[0].y = (int)bbWidget.ll[1];
  rectWidget[0].width = (int)sizeWidget[0];
  rectWidget[0].height = (int)sizeWidget[1];

  // window clip region
  BBox bbWindow = bbWidget * widgetToWindow;
  Vector sizeWindow = bbWindow.size();

  rectWindow[0].x = (int)bbWindow.ll[0];
  rectWindow[0].y = (int)bbWindow.ll[1];
  rectWindow[0].width = (int)sizeWindow[0];
  rectWindow[0].height = (int)sizeWindow[1];

  if (!highliteGC) {
    highliteGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, highliteGC, getColor("blue"));
    XSetLineAttributes(display, highliteGC, 2, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, highliteGC, 0, 0, rectWidget, 1, Unsorted);

  // crosshairGC
  if (!crosshairGC) {
    crosshairGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, crosshairGC, getColor("green"));
    XSetLineAttributes(display, crosshairGC, 1, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, crosshairGC, 0, 0, rectWidget, 1, Unsorted);

  // crosshairGCXOR
  if (!crosshairGCXOR) {
    crosshairGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, crosshairGCXOR, getColor("green"));
    XSetLineAttributes(display, crosshairGCXOR, 1, LineSolid, CapButt, 
		       JoinMiter);
  }
  XSetClipRectangles(display, crosshairGCXOR, 0, 0, rectWindow, 1, Unsorted);

  // markerGC
  if (!markerGC)
    markerGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, markerGC, 0, 0, rectWidget, 1, Unsorted);

  if (!markerGCXOR) {
    markerGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, markerGCXOR, getColor("white"));
  }
  XSetClipRectangles(display, markerGCXOR, 0, 0, rectWindow, 1, Unsorted);

  if (!selectGCXOR) {
    selectGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, selectGCXOR, getColor("white"));
    char dlist[] = {8,3};
    XSetLineAttributes(display, selectGCXOR, 1, LineOnOffDash, 
                     CapButt, JoinMiter);
    XSetDashes(display, selectGCXOR, 0, dlist, 2);
  }
  XSetClipRectangles(display, selectGCXOR, 0, 0, rectWindow, 1, Unsorted);

  // gridGC
  if (!gridGC)
    gridGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, gridGC, 0, 0, rectWidget, 1, Unsorted);

  // contourGC
  if (!contourGC) {
    contourGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetLineAttributes(display, contourGC, 1, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, contourGC, 0, 0, rectWidget, 1, Unsorted);
}

void Base::updateMagnifier()
{
  updateMagnifier(magnifierCursor);
}

void Base::updateMagnifier(const Vector& v)
{
  // v is in CANVAS coords
  // save it, we may need it later
  magnifierCursor = v;

  if (useMagnifier && magnifierXImage && magnifierGC && doRender()) {

    // clear XImage data
    memset(XImageData(magnifierXImage), 0, magnifierXImage->bytes_per_line * 
	   magnifierHeight);

    // fill XImage
    updateMagnifierZoom(v); // pixmap is updated inside

    // render cursor
    if (useMagnifierCursor)
      renderMagnifierCursor(v);

    // notify the magnifier widget
    ostringstream str;
    str << magnifierName << " update " << magnifierPixmap << ends;
    Tcl_Eval(interp, str.str().c_str());
  }
}

void Base::updateMatrices()
{
  // refToUser, userToWidget needs to be set prior
  widgetToCanvas = Translate(originX, originY);
  canvasToWidget = widgetToCanvas.invert();

  short x, y;
  Tk_CanvasWindowCoords(canvas, 0, 0, &x, &y);
  canvasToWindow = Translate(x, y);
  windowToCanvas = canvasToWindow.invert();

  widgetToWindow = widgetToCanvas * canvasToWindow;
  windowToWidget = widgetToWindow.invert();

  // These are derived Transformation Matrices
  refToWidget = refToUser * userToWidget;
  widgetToRef = refToWidget.invert();

  refToCanvas = refToUser * userToWidget * widgetToCanvas;
  canvasToRef = refToCanvas.invert();

  refToWindow = refToUser * userToWidget * widgetToCanvas * canvasToWindow;
  windowToRef = refToWindow.invert();

  userToCanvas = userToWidget * widgetToCanvas;
  canvasToUser = userToCanvas.invert();
}

void Base::updateNow(UpdateType flag)
{
  if (DebugPerf)
    cerr << "updateNow" << endl;

  if (flag < needsUpdate)
    needsUpdate = flag;
  redrawNow();
}

void Base::updateNow(UpdateType flag, BBox bb)
{
  if (DebugPerf)
    cerr << "updateNow BB" << endl;

  // bb is in canvas coords

  if (flag < needsUpdate)
    needsUpdate = flag;
  redrawNow(bb);
}

int Base::updatePixmap(const BBox& bb)
{
  // bbox is in canvas coords

  // Note: lack of breaks-- on purpose. 
  // If Matrices are update, both Base and Pixmap
  // also need to be updated. Same for Base-- ie, pixmap is also updated.

  switch (needsUpdate) {
  case MATRIX:
    updateMatrices();
  case BASE:
    updateBase();
  case PIXMAP:
    updatePM(bb);
    break;
  }

  needsUpdate = NOUPDATE;
  return TCL_OK;
}

void Base::updatePM(const BBox& bbox)
{
  // bbox is in Canvas Coords

  int& width = options->width;
  int& height = options->height;

  if (!pixmap) {
    if (!(pixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				width, height, depth))) {
      cerr << "Frame Internal Error: Unable to Create Pixmap" << endl;
      exit(1);
    }
  }

  if (!bbox.isEmpty()) {
    BBox bb = bbox * canvasToWidget;
    int x0 = (int)bb.ll[0] > 0 ? (int)bb.ll[0] : 0;
    int y0 = (int)bb.ll[1] > 0 ? (int)bb.ll[1] : 0;
    int x1 = (int)bb.ur[0] < width ? (int)bb.ur[0] : width;
    int y1 = (int)bb.ur[1] < height ? (int)bb.ur[1] : height;
    int sx = x1-x0;
    int sy = y1-y0;

    if (DebugPerf)
      cerr << ' ' << x0 << ' ' << y0 << ' ' << x1 << ' ' << y1 << ' ';

    XCopyArea(display, basePixmap, pixmap, gc, x0, y0, sx, sy, x0, y0);
  }
}

char* Base::varcat(char* buf, char* var, char r)
{
  int l = strlen(var);
  strcpy(buf,var);
  buf[l++] = r;
  buf[l++] = NULL;
  return buf;
}

void Base::ximageToPixmap(Pixmap pmap, XImage* xmap, 
			  double* (FitsImage::*getMatrix)())
{
  buildXImage(xmap, getMatrix);
  XPutImage(display, pmap, gc, xmap, 0, 0, 0, 0, xmap->width, xmap->height);
}

#ifdef _MACOSX
void Base::macosx(double* (FitsImage::*getMatrix)())
{
  // clip rect
  XRectangle* rr = rectWidget;
  Vector v1 = Vector(rr->x, rr->y) * widgetToCanvas;
  Vector v2 = Vector(rr->x+rr->width, rr->y+rr->height) * widgetToCanvas;
  macosxClip(v1,v2-v1);

  // scale
  double scale = 150 / 72.;
  int width = (int)(options->width*scale);
  int height = (int)(options->height*scale);

  // image
  macosxImage(scale, width, height, v1, v2-v1, getMatrix);
}

void Base::macosxImage(float scale, int width, int height, 
		       const Vector& v, const Vector& s,
		       double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  // update PS Matrix
  updatePS(scale, width, height);

  // source
  unsigned char* src = fillImage(width, height, getMatrix);
  if (!src)
    return;

  // destination
  unsigned char* dst = new unsigned char[width*height*4];
  if (!dst)
    return;

  unsigned char* sptr = src;
  unsigned char* dptr = dst;
  for (int ii=0; ii<width*height; ii++) {
    *dptr++ = *sptr++;
    *dptr++ = *sptr++;
    *dptr++ = *sptr++;
    *dptr++ = 0;
  }

  if (src)
    delete [] src;

  macosxBitmapCreate(dst, width, height, v, s);

  if (dst)
    delete [] dst;
}

#endif

#ifdef _WIN32
void Base::win32(double* (FitsImage::*getMatrix)())
{
  // clip rect
  XRectangle* rr = rectWidget;
  Vector v1 = Vector(rr->x, rr->y) * widgetToCanvas;
  Vector v2 = Vector(rr->x+rr->width, rr->y+rr->height) * widgetToCanvas;
  win32Clip(v1,v2-v1);

  // scale
  double scale = 1.0;
  int width = (int)(options->width*scale);
  int height = (int)(options->height*scale);

  // image
  win32Image(scale, width, height, v1, v2-v1, getMatrix);
}

void Base::win32Image(float scale, int width, int height, 
		     const Vector& v, const Vector& s,
		     double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  // update PS Matrix
  updatePS(scale, width, height);

  // source
  unsigned char* src = fillImage(width, height, getMatrix);
  if (!src)
    return;

  // destination
  unsigned char* dst = new unsigned char[width*height*3];
  if (!dst)
    return;

  unsigned char* sptr = src;
  unsigned char* dptr = dst;
  unsigned char red, green, blue;
  for (int ii=0; ii<width*height; ii++) {
    red   = *sptr++;
    green = *sptr++;
    blue  = *sptr++;
    *dptr++ = blue;
    *dptr++ = green;
    *dptr++ = red;
  }

  if (src)
    delete [] src;

  win32BitmapCreate(dst, width, height, v, s);

  if (dst)
    delete [] dst;
}
#endif
