// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "baseellipse.h"
#include "framebase.h"
#include "fitsimage.h"

BaseEllipse::BaseEllipse(const BaseEllipse& a) : BaseMarker(a) {}

BaseEllipse::BaseEllipse(FrameBase* p, const Vector& ctr, 
			 double ang,
			 const char* clr, int* dsh, 
			 int w, const char* f, 
			 const char* t, unsigned short prop, const char* c,
			 const List<Tag>& tag, const List<CallBack>& cb)
  : BaseMarker(p, ctr, ang, clr, dsh, w, f, t, prop, c, tag, cb) {}

BaseEllipse::~BaseEllipse() {}

int BaseEllipse::isIn(const Vector& v)
{
  // v is in canvas coords
  Vector r = annuli_[numAnnuli_-1];

  // zero radius
  if (!r[0] || !r[1])
    return 0;

  Matrix mm = bckRefMatrix();
  Vector p = v * mm;

  if ((p[0]*p[0])/(r[0]*r[0]) + (p[1]*p[1])/(r[1]*r[1]) <= 1)
    return 1;
  else
    return 0;
}

void BaseEllipse::updateHandles()
{
  // handles are in canvas coords
  // we can't garantee that the annuli_ have been sorted yet

  if (handle)
    delete [] handle;
  handle = new Vector[numHandle];

  Vector max;
  for(int i=0; i<numAnnuli_; i++)
    if (max[0]<annuli_[i][0])
      max = annuli_[i];
  Vector r = max;

  Matrix mm = fwdCanvasMatrix();

  handle[0] = Vector(-r[0],-r[1]) * mm;
  handle[1] = Vector( r[0],-r[1]) * mm;
  handle[2] = Vector( r[0], r[1]) * mm;
  handle[3] = Vector(-r[0], r[1]) * mm;

  // the rest are annuli_
  for (int i=4; i<numHandle; i++)
    handle[i] = Vector(annuli_[i-4][0],0) * mm;
}

Vector BaseEllipse::intersect(Vector rr, double aa)
{
  double tt = rr[1]*rr[1]*cos(aa)*cos(aa)+rr[0]*rr[0]*sin(aa)*sin(aa);
  double ss =0;
  if (tt>0) 
    ss = rr[0]*rr[1]/sqrt(tt);

  return Vector(ss*cos(aa),ss*sin(aa));
}

// this routine maps the desired angle to an angle to be used with XDrawArc
double BaseEllipse::xyz(Vector rr, double aa)
{
  // just in case
  if (!rr[0] || !rr[1]) 
    return aa;

  int flip=0;
  while (aa>M_PI) {
    aa -= M_PI;
    flip++;
  }

  double tt = rr[1]*rr[1]*cos(aa)*cos(aa)+rr[0]*rr[0]*sin(aa)*sin(aa);
  double ss =0;
  if (tt>0) 
    ss = 1./sqrt(tt);

  double bb = rr[1]*ss*cos(aa);
  return acos(bb)+M_PI*flip;
}

// X11

void BaseEllipse::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  double ang = calcAngle();
  Vector r = annuli_[numAnnuli_-1];
  Vector z = parent->zoom();

  int isOrient = parent->getOrientation() == NORMAL && 
    parent->getWCSOrientation() == NORMAL;
  int isRound = r[0] == r[1];
  int isScale = z[0] == z[1];
  int isAngle = (ang > -FLT_EPSILON && ang < FLT_EPSILON) ||
    (ang > M_PI-FLT_EPSILON && ang < M_PI+FLT_EPSILON);

  if (isRound && isScale && isOrient)
    renderXCircle(drawable, mx, ang, mode);
  else if (!isRound && isScale && isAngle && isOrient)
    renderXEllipse(drawable, mx, ang, mode);
  else
    renderXEllipseCurve(drawable, mx, mode);
}

void BaseEllipse::renderXCircle(Drawable drawable, const Matrix& mx,
				double ang, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  // this routine is only valid for circles with equal zoom in x & y
  Matrix mm = fwdRefMatrix() * mx;
  Vector c = center * mx;

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    Vector ur = (r * mm).round();
    double l = (ur-c).length() * cos(M_PI/4);

    Vector rr(l,l);
    Vector st = (c-rr).round();
    Vector size = (rr*2).round();

    // Verify size is positive
    // XDrawArc is sensative to bad data, and may hang the XServer
    if (!size[0] || !size[1])
      continue;

    double a1 = startAng_ + radToDeg(ang);
    double a2 = stopAng_ + radToDeg(ang);
    while (a1>360 && a1>0) {
      a1 -= 360;
      a2 -= 360;
    }
    if (a2<=a1)
      a2 += 360;

    XDrawArc(display, drawable, lgc, st[0], st[1], size[0], size[1], 
	     a1*64, (a2-a1)*64);
  }

  renderXInclude(drawable, mx, mode);
}

void BaseEllipse::renderXEllipse(Drawable drawable, const Matrix& mx, 
				 double ang, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  // this routine is only valid for equal zoom in x & y and 
  // angles of 0,90,180,and 270
  Matrix mm = fwdRefMatrix() * mx;

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    Vector ur = ( r * mm).round();
    Vector ll = (-r * mm).round();
    Vector size = (ur - ll).abs();

    // Verify size is positive
    // XDrawArc is sensative to bad data, and may hang the XServer
    if (!size[0] || !size[1])
      continue;

    Vector st(ur[0]>ll[0]?ll[0]:ur[0], ur[1]>ll[1]?ll[1]:ur[1]);

    double a1 = radToDeg(xyz(r,degToRad(startAng_) + ang));
    double a2 = radToDeg(xyz(r,degToRad(stopAng_ ) + ang));
    while (a1>360 && a1>0) {
      a1 -= 360;
      a2 -= 360;
    }
    if (a2<=a1)
      a2 += 360;

    XDrawArc(display, drawable, lgc, st[0], st[1], size[0], size[1],
	     a1*64, (a2-a1)*64);
  }

  renderXInclude(drawable, mx, mode);
}

void BaseEllipse::renderXEllipseCurve(Drawable drawable, const Matrix& mx, 
				       RenderMode mode)
{
  double a1 = degToRad(startAng_);
  double a2 = degToRad(stopAng_);

  if (a2 <= a1)
    a2 += 2*M_PI;

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    int s1 =0;
    int s2 =0;
    for (int i=0; i<8; i++) {
      double b1 = i*M_PI/2;
      double b2 = (i+1)*M_PI/2;
      if (!s1 && a1>=b1 && a1<b2)
	s1 =1;
      if (!s2 && a2>b1 && a2<=b2)
	s2 =1;

      if ((s1 && !s2) || (s1 && s2))
	renderXEllipsePrep(drawable, mx, mode, a1,a2,b1,b2,r);

      if (s1&&s2)
	s1=s2=0;
    }
  }

  renderXInclude(drawable, mx, mode);
}

void BaseEllipse::renderXEllipsePrep(Drawable drawable, const Matrix& mx,
				     RenderMode mode,
				     double a1, double a2, 
				     double b1, double b2, 
				     Vector& r)
{
  if (!(a1 >= b1 && a1 <= b2))
    a1 = b1;
  if (!(a2 >= b1 && a2 <= b2))
    a2 = b2;

  if (a1>a2) {
    renderXEllipseArc(drawable, mx, mode, b1,a2,r);
    renderXEllipseArc(drawable, mx, mode, a1,b2,r);
  }
  else
    renderXEllipseArc(drawable, mx, mode, a1,a2,r);
}				

void BaseEllipse::renderXEllipseArc(Drawable drawable, const Matrix& mx,
				    RenderMode mode,
				    double a1, double a2, Vector& rr)

{
  // don't render zero length arcs
  if ((a1 > a2-FLT_EPSILON) && (a1 < a2+FLT_EPSILON))
    return;

  if (((rr[0] > -FLT_EPSILON) && (rr[0] < FLT_EPSILON)) || 
      ((rr[1] > -FLT_EPSILON) && (rr[1] < FLT_EPSILON)))
    return;

  // bezier curve, valid for arcs of <M_PI/2
  Matrix mm = FlipY() * fwdRefMatrix() * mx;

  double aa1 = xyz(rr,a1);
  double aa2 = xyz(rr,a2);
  double bcp = 4.0/3*(1-cos((aa2-aa1)/2))/sin((aa2-aa1)/2);
  
  Vector t0 = intersect(rr,a1);
  Vector x1 = Vector(rr[0]*(cos(aa1)-bcp*sin(aa1)),
		     rr[1]*(sin(aa1)+bcp*cos(aa1)));
  Vector x2 = Vector(rr[0]*(cos(aa2)+bcp*sin(aa2)),
		     rr[1]*(sin(aa2)-bcp*cos(aa2)));
  Vector t1 = intersect(rr,a2);

  Vector tt0 = t0*mm;
  Vector xx1 = x1*mm;
  Vector xx2 = x2*mm;
  Vector tt1 = t1*mm;

  XDrawCurve(drawable, mode, tt0,xx1,xx2,tt1);
}

void BaseEllipse::XDrawCurve(Drawable drawable, RenderMode mode,
			     Vector& t1, Vector& x1,
			     Vector& x2, Vector& t2)
{
  // if dashed, fake it
  GC lgc;
  if ((properties & SOURCE) && !(properties & DASH))
    lgc = renderXGC(mode);
  else {
    // set width, color, dash
    switch (mode) {
    case SRC:
      XSetForeground(display, gc, color); 
      renderXLineNoDash(gc);
      lgc = gc;
      break;
    case XOR:
      renderXLineNoDash(gcxor);
      lgc = gcxor;
      break;
    }
  }

  float t1x = t1[0];
  float t1y = t1[1];
  float t2x = t2[0];
  float t2y = t2[1];

  float x1x = x1[0];
  float x1y = x1[1];
  float x2x = x2[0];
  float x2y = x2[1];

  float rx = t1x;
  float ry = t1y;

  Vector dd = (t2-t1).abs();
  float max = dd[0]>dd[1] ? dd[0] : dd[1];

  int bbx = parent->options->width;
  int bby = parent->options->height;

  // calculate incr
  // this is a crude attempt to limit the number of iterations
  // we want a min for very small segments, but not that large for
  // high zoom or elongated curves
  float aa = int(log(max))*5;
  float incr = 1./(aa > 2 ? aa : 2);

  int dash=0;
  float tt = incr;
  while (tt<=1+FLT_EPSILON) {
    float xx = pow(tt,3)*(t2x+3*(x1x-x2x)-t1x)
      +3*pow(tt,2)*(t1x-2*x1x+x2x)
      +3*tt*(x1x-t1x)+t1x;
    float yy = pow(tt,3)*(t2y+3*(x1y-x2y)-t1y)
      +3*pow(tt,2)*(t1y-2*x1y+x2y)
      +3*tt*(x1y-t1y)+t1y;

    // crude attempt to clip unwanted drawlines
    if ((rx>0 && rx<bbx && ry>0 && ry<bby) ||
	(xx>0 && xx<bbx && yy>0 && yy<bby)) {
      if ((properties & SOURCE) && !(properties & DASH))
	XDrawLine(display, drawable, lgc, rx, ry, xx, yy);
      else {
	if (dash)
	  XDrawLine(display, drawable, lgc, rx, ry, xx, yy);    
	dash = !dash;
      }
    }

    rx = xx;
    ry = yy;

    tt += incr;
  }
}

void BaseEllipse::renderXInclude(Drawable drawable, const Matrix& mx, 
				RenderMode mode)
{
  if (!(properties & INCLUDE)) {
    Matrix mm = fwdRefMatrix() * mx;
    double theta = degToRad(45);

    Vector r1 = (Vector(annuli_[numAnnuli_-1][0]*cos(theta), 
			annuli_[numAnnuli_-1][1]*sin(theta)) * mm).round();
    Vector r2 = (Vector(-annuli_[numAnnuli_-1][0]*cos(theta), 
			-annuli_[numAnnuli_-1][1]*sin(theta)) * mm).round();
  
    GC lgc = renderXGC(mode);
    if (mode == SRC)
      XSetForeground(display, gc, parent->getColor("red"));

    XDrawLine(display, drawable, lgc, r1[0], r1[1], r2[0], r2[1]);    
  }
}

// PS

void BaseEllipse::renderPS(int mode) {
  Vector r = annuli_[numAnnuli_-1];
  Vector z = parent->zoom();

  int isOrient = parent->getOrientation() == NORMAL && 
    parent->getWCSOrientation() == NORMAL;
  int isRound = r[0] == r[1];
  int isScale = z[0] == z[1];

  if (isRound && isScale && isOrient)
    renderPSCircle(mode);
  else
    renderPSEllipse(mode);
}

void BaseEllipse::renderPSCircle(int mode)
{
  renderPSGC(mode);

  Vector cc = center * parent->refToCanvas;
  Matrix mm = fwdCanvasMatrix();
  double ang = calcAngle();

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    Vector ur = r * mm;
    double l = (ur-cc).length() * cos(M_PI/4);

    // don't render zero length arcs
    if (!l)
      continue;

    double a1 = startAng_ + radToDeg(ang);
    double a2 = stopAng_ + radToDeg(ang);
    while (a1>360 && a1>0) {
      a1 -= 360;
      a2 -= 360;
    }
    if (a2<=a1)
      a2 += 360;

    ostringstream str;
    str << "newpath " << cc.TkCanvasPs(parent->canvas) << l << ' '
	<< a1 << ' ' << a2 << " arc stroke " << endl << ends;

    Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
  }

  renderPSInclude(mode);
}

void BaseEllipse::renderPSEllipse(int mode)
{
  renderPSGC(mode);

  double a1 = degToRad(startAng_);
  double a2 = degToRad(stopAng_);

  if (a2 <= a1)
    a2 += 2*M_PI;

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    int s1 =0;
    int s2 =0;
    for (int i=0; i<8; i++) {
      double b1 = i*M_PI/2;
      double b2 = (i+1)*M_PI/2;
      if (!s1 && a1>=b1 && a1<b2)
	s1 =1;
      if (!s2 && a2>b1 && a2<=b2)
	s2 =1;

      if ((s1 && !s2) || (s1 && s2))
	renderPSEllipsePrep(a1,a2,b1,b2,r);

      if (s1&&s2)
	s1=s2=0;
    }
  }

  renderPSInclude(mode);
}

void BaseEllipse::renderPSEllipsePrep(double a1, double a2, 
				      double b1, double b2, 
				      Vector& r)
{
  if (!(a1 >= b1 && a1 <= b2))
    a1 = b1;
  if (!(a2 >= b1 && a2 <= b2))
    a2 = b2;

  if (a1>a2) {
    renderPSEllipseArc(b1,a2,r);
    renderPSEllipseArc(a1,b2,r);
  }
  else
    renderPSEllipseArc(a1,a2,r);
}				

void BaseEllipse::renderPSEllipseArc(double a1, double a2, Vector& rr)

{
  // don't render zero length arcs
  if (a1 == a2)
    return;

  if (!rr[0] || !rr[1])
    return;

  // bezier curve, valid for arcs of <M_PI/2
  Matrix mm = FlipY() * fwdCanvasMatrix();

  double aa1 = xyz(rr,a1);
  double aa2 = xyz(rr,a2);
  double bcp = 4.0/3*(1-cos((aa2-aa1)/2))/sin((aa2-aa1)/2);
  
  Vector t0 = intersect(rr,a1);
  Vector x1 = Vector(rr[0]*(cos(aa1)-bcp*sin(aa1)),
		     rr[1]*(sin(aa1)+bcp*cos(aa1)));
  Vector x2 = Vector(rr[0]*(cos(aa2)+bcp*sin(aa2)),
		     rr[1]*(sin(aa2)-bcp*cos(aa2)));
  Vector t1 = intersect(rr,a2);

  Vector tt0 = t0*mm;
  Vector xx1 = x1*mm;
  Vector xx2 = x2*mm;
  Vector tt1 = t1*mm;

  ostringstream str;
  str <<
    "newpath " <<
    tt0.TkCanvasPs(parent->canvas) << "moveto " <<
    xx1.TkCanvasPs(parent->canvas) << 
    xx2.TkCanvasPs(parent->canvas) << 
    tt1.TkCanvasPs(parent->canvas) << 
    " curveto stroke" << 
    endl;

  str << ends;

  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
}

void BaseEllipse::renderPSInclude(int mode)
{
  if (!(properties & INCLUDE)) {
    Matrix mm = fwdCanvasMatrix();
    double theta = degToRad(45);
    Vector r1 = Vector(annuli_[numAnnuli_-1][0]*cos(theta), 
		       annuli_[numAnnuli_-1][1]*sin(theta)) * mm;
    Vector r2 = Vector(-annuli_[numAnnuli_-1][0]*cos(theta), 
		       -annuli_[numAnnuli_-1][1]*sin(theta)) * mm;

    renderPSColor(mode, parent->getXColor("red"));

    ostringstream str;
    str << "newpath " 
	<< r1.TkCanvasPs(parent->canvas) << "moveto"
	<< r2.TkCanvasPs(parent->canvas) << "lineto"
	<< " stroke" << endl << ends;
    Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
  }
}

// MacOSX

#ifdef _MACOSX
void BaseEllipse::renderMACOSX() {
  Vector r = annuli_[numAnnuli_-1];
  Vector z = parent->zoom();

  int isOrient = parent->getOrientation() == NORMAL && 
    parent->getWCSOrientation() == NORMAL;
  int isRound = r[0] == r[1];
  int isScale = z[0] == z[1];

  if (isRound && isScale && isOrient)
    renderMACOSXCircle();
  else
    renderMACOSXEllipse();
}

void BaseEllipse::renderMACOSXCircle()
{
  renderMACOSXGC();

  Vector cc = center * parent->refToCanvas;
  Matrix mm = fwdCanvasMatrix();
  double ang = calcAngle();

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    Vector ur = r * mm;
    double l = (ur-cc).length() * cos(M_PI/4);

    // don't render zero length arcs
    if (!l)
      continue;

    double a1 = degToRad(startAng_) + ang;
    double a2 = degToRad(stopAng_) + ang;

    if (a2<=a1)
      a2 += 2*M_PI;

    macosxDrawArc(cc, l, a1, a2);
  }

  renderMACOSXInclude();
}

void BaseEllipse::renderMACOSXEllipse()
{
  renderMACOSXGC();

  double a1 = degToRad(startAng_);
  double a2 = degToRad(stopAng_);

  if (a2 <= a1)
    a2 += 2*M_PI;

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    int s1 =0;
    int s2 =0;
    for (int i=0; i<8; i++) {
      double b1 = i*M_PI/2;
      double b2 = (i+1)*M_PI/2;
      if (!s1 && a1>=b1 && a1<b2)
	s1 =1;
      if (!s2 && a2>b1 && a2<=b2)
	s2 =1;

      if ((s1 && !s2) || (s1 && s2))
	renderMACOSXEllipsePrep(a1,a2,b1,b2,r);

      if (s1&&s2)
	s1=s2=0;
    }
  }

  renderMACOSXInclude();
}

void BaseEllipse::renderMACOSXEllipsePrep(double a1, double a2, 
				      double b1, double b2, 
				      Vector& r)
{
  if (!(a1 >= b1 && a1 <= b2))
    a1 = b1;
  if (!(a2 >= b1 && a2 <= b2))
    a2 = b2;

  if (a1>a2) {
    renderMACOSXEllipseArc(b1,a2,r);
    renderMACOSXEllipseArc(a1,b2,r);
  }
  else
    renderMACOSXEllipseArc(a1,a2,r);
}				

void BaseEllipse::renderMACOSXEllipseArc(double a1, double a2, Vector& rr)

{
  // don't render zero length arcs
  if (a1 == a2)
    return;

  if (!rr[0] || !rr[1])
    return;

  // bezier curve, valid for arcs of <M_PI/2
  Matrix mm = FlipY() * fwdCanvasMatrix();

  double aa1 = xyz(rr,a1);
  double aa2 = xyz(rr,a2);
  double bcp = 4.0/3*(1-cos((aa2-aa1)/2))/sin((aa2-aa1)/2);
  
  Vector t0 = intersect(rr,a1);
  Vector x1 = Vector(rr[0]*(cos(aa1)-bcp*sin(aa1)),
		     rr[1]*(sin(aa1)+bcp*cos(aa1)));
  Vector x2 = Vector(rr[0]*(cos(aa2)+bcp*sin(aa2)),
		     rr[1]*(sin(aa2)-bcp*cos(aa2)));
  Vector t1 = intersect(rr,a2);

  Vector tt0 = t0*mm;
  Vector xx1 = x1*mm;
  Vector xx2 = x2*mm;
  Vector tt1 = t1*mm;

  macosxDrawCurve(tt0, xx1, xx2, tt1);
}

void BaseEllipse::renderMACOSXInclude()
{
  if (!(properties & INCLUDE)) {
    Matrix mm = fwdCanvasMatrix();
    double theta = degToRad(45);
    Vector r1 = Vector(annuli_[numAnnuli_-1][0]*cos(theta), 
		       annuli_[numAnnuli_-1][1]*sin(theta)) * mm;
    Vector r2 = Vector(-annuli_[numAnnuli_-1][0]*cos(theta), 
		       -annuli_[numAnnuli_-1][1]*sin(theta)) * mm;

    macosxColor(parent->getXColor("red"));
    macosxDrawLine(r1,r2);
  }
}

#endif

// Win32

#ifdef _WIN32
void BaseEllipse::renderWIN32() {
  Vector r = annuli_[numAnnuli_-1];
  Vector z = parent->zoom();

  int isOrient = parent->getOrientation() == NORMAL && 
    parent->getWCSOrientation() == NORMAL;
  int isRound = r[0] == r[1];
  int isScale = z[0] == z[1];

  if (isRound && isScale && isOrient)
    renderWIN32Circle();
  else
    renderWIN32Ellipse();
}

void BaseEllipse::renderWIN32Circle()
{
  renderWIN32GC();

  Vector cc = center * parent->refToCanvas;
  Matrix mm = fwdCanvasMatrix();
  double ang = calcAngle();

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    Vector ur = r * mm;
    double l = (ur-cc).length() * cos(M_PI/4);

    // don't render zero length arcs
    if (!l)
      continue;

    double a1 = degToRad(startAng_) + ang;
    double a2 = degToRad(stopAng_) + ang;

    if (a2<=a1)
      a2 += 2*M_PI;

    win32DrawArc(cc, l, a1, a2);
  }

  renderWIN32Include();
}

void BaseEllipse::renderWIN32Ellipse()
{
  renderWIN32GC();

  double a1 = degToRad(startAng_);
  double a2 = degToRad(stopAng_);

  if (a2 <= a1)
    a2 += 2*M_PI;

  for (int i=0; i<numAnnuli_; i++) {
    Vector r = annuli_[i];

    int s1 =0;
    int s2 =0;
    for (int i=0; i<8; i++) {
      double b1 = i*M_PI/2;
      double b2 = (i+1)*M_PI/2;
      if (!s1 && a1>=b1 && a1<b2)
	s1 =1;
      if (!s2 && a2>b1 && a2<=b2)
	s2 =1;

      if ((s1 && !s2) || (s1 && s2))
	renderWIN32EllipsePrep(a1,a2,b1,b2,r);

      if (s1&&s2)
	s1=s2=0;
    }
  }

  renderWIN32Include();
}

void BaseEllipse::renderWIN32EllipsePrep(double a1, double a2, 
				      double b1, double b2, 
				      Vector& r)
{
  if (!(a1 >= b1 && a1 <= b2))
    a1 = b1;
  if (!(a2 >= b1 && a2 <= b2))
    a2 = b2;

  if (a1>a2) {
    renderWIN32EllipseArc(b1,a2,r);
    renderWIN32EllipseArc(a1,b2,r);
  }
  else
    renderWIN32EllipseArc(a1,a2,r);
}				

void BaseEllipse::renderWIN32EllipseArc(double a1, double a2, Vector& rr)

{
  // don't render zero length arcs
  if (a1 == a2)
    return;

  if (!rr[0] || !rr[1])
    return;

  // bezier curve, valid for arcs of <M_PI/2
  Matrix mm = FlipY() * fwdCanvasMatrix();

  double aa1 = xyz(rr,a1);
  double aa2 = xyz(rr,a2);
  double bcp = 4.0/3*(1-cos((aa2-aa1)/2))/sin((aa2-aa1)/2);
  
  Vector t0 = intersect(rr,a1);
  Vector x1 = Vector(rr[0]*(cos(aa1)-bcp*sin(aa1)),
		     rr[1]*(sin(aa1)+bcp*cos(aa1)));
  Vector x2 = Vector(rr[0]*(cos(aa2)+bcp*sin(aa2)),
		     rr[1]*(sin(aa2)-bcp*cos(aa2)));
  Vector t1 = intersect(rr,a2);

  Vector tt0 = t0*mm;
  Vector xx1 = x1*mm;
  Vector xx2 = x2*mm;
  Vector tt1 = t1*mm;

  win32DrawCurve(tt0, xx1, xx2, tt1);
}

void BaseEllipse::renderWIN32Include()
{
  if (!(properties & INCLUDE)) {
    Matrix mm = fwdCanvasMatrix();
    double theta = degToRad(45);
    Vector r1 = Vector(annuli_[numAnnuli_-1][0]*cos(theta), 
		       annuli_[numAnnuli_-1][1]*sin(theta)) * mm;
    Vector r2 = Vector(-annuli_[numAnnuli_-1][0]*cos(theta), 
		       -annuli_[numAnnuli_-1][1]*sin(theta)) * mm;

    win32Color(parent->getXColor("red"));
    win32DrawLine(r1,r2);
  }
}

#endif
