// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "colorscale.h"

ColorScale::ColorScale(int s)
{
  size_ = s;
  psColors_ = new unsigned char[s*3];
  psIndex_ = new unsigned short[s];

  memset(psColors_, '0', size_*3);
  memset(psIndex_, '0', size_*sizeof(unsigned short));
}

ColorScale::~ColorScale()
{
  if (psColors_)
    delete [] psColors_;

  if (psIndex_)
    delete [] psIndex_;
}

LinearScale::LinearScale(int s, unsigned short* indexCells, 
			 unsigned char* colorCells, int count) : ColorScale(s)
{
  for (int i=0; i<s; i++) {
    double a = double(i)/s;
    int l = (int)(a * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}

LogScale::LogScale(int s, unsigned short* indexCells, 
		   unsigned char* colorCells, int count, 
		   double low, double high) : ColorScale(s)
{
  CheckScaleLog(&low,&high);

  for (int i=0; i<s; i++) {
    double a = log10(high/low*double(i)/s +1) / log10(high/low +1);
    int l = (int)(a * count);

    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3, 3);
  }
}

PowScale::PowScale(int s, unsigned short* indexCells, 
		   unsigned char* colorCells, int count, 
		   double low, double high) : ColorScale(s)
{
  CheckScaleLog(&low,&high);

  for (int i=0; i<s; i++) {
    double a = ::pow(high/low, double(i)/s) *low/high;
    //special case
    if (a==1)
      a = 0;
    int l = (int)(a * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3, 3);
  }
}

SqrtScale::SqrtScale(int s, unsigned short* indexCells, 
		     unsigned char* colorCells, int count) : ColorScale(s)
{
  for(int i=0; i<s; i++) {
    double a = double(i)/s;
    int l = (int)(sqrt(a) * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}

SquaredScale::SquaredScale(int s, unsigned short* indexCells, 
			   unsigned char* colorCells, int count) 
  : ColorScale(s)
{
  for(int i=0; i<s; i++) {
    double a = double(i)/s;
    int l = (int)(a*a * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}

HistEquScale::HistEquScale(int s, unsigned short* indexCells, 
			   unsigned char* colorCells, int count, 
			   double* hist, int histsize)
  : ColorScale(s)
{
  // if no histogram, return linear distribution
  if (!hist)
    for (int i=0; i<s; i++) {
      double a = double(i)/s;
      int l = (int)(a * count);
      psIndex_[i] = indexCells[l];
      memcpy(psColors_+i*3, colorCells+l*3,3);
    }
  else {
    for (int i=0; i<s; i++) {
      double a = hist[i*histsize/s];
      int l = (int)(a * count);
      psIndex_[i] = indexCells[l];
      memcpy(psColors_+i*3, colorCells+l*3,3);
    }
  }
}

IISScale::IISScale(unsigned short* indexCells, 
		   unsigned char* colorCells, int count) : ColorScale(IISSIZE)
{
  {
    for (int i=0; i<IISCOLORS; i++) {
      double a = double(i)/IISCOLORS;
      int l = (int)(a * count);
      psIndex_[i] = indexCells[l];
      memcpy(psColors_+i*3, colorCells+l*3,3);
    }
  }

  {
    for (int i=IISCOLORS; i<IISSIZE; i++)
      psIndex_[i] = i;

    unsigned char* ptr = psColors_+200*3;
    // 200 black
    *ptr++ = 0;
    *ptr++ = 0;
    *ptr++ = 0;
    // 201 black
    *ptr++ = 0;
    *ptr++ = 0;
    *ptr++ = 0;
    // 202 white
    *ptr++ = 255;
    *ptr++ = 255;
    *ptr++ = 255;
    // 203 red
    *ptr++ = 0;
    *ptr++ = 0;
    *ptr++ = 255;
    // 204 green
    *ptr++ = 0;
    *ptr++ = 255;
    *ptr++ = 0;
    // 205 blue
    *ptr++ = 255;
    *ptr++ = 0;
    *ptr++ = 0;
    // 206 yellow
    *ptr++ = 0;
    *ptr++ = 255;
    *ptr++ = 255;
    // 207 cyan
    *ptr++ = 255;
    *ptr++ = 255;
    *ptr++ = 0;
    // 208 magenta
    *ptr++ = 255;
    *ptr++ = 0;
    *ptr++ = 255;
    // 209 coral
    *ptr++ = 80;
    *ptr++ = 127;
    *ptr++ = 255;
    // 210 maroon
    *ptr++ = 96;
    *ptr++ = 48;
    *ptr++ = 176;
    // 211 orange
    *ptr++ = 0;
    *ptr++ = 165;
    *ptr++ = 255;
    // 212 khaki
    *ptr++ = 140;
    *ptr++ = 230;
    *ptr++ = 240;
    // 213 orchid
    *ptr++ = 214;
    *ptr++ = 112;
    *ptr++ = 218;
    // 214 turquoise
    *ptr++ = 208;
    *ptr++ = 224;
    *ptr++ = 64;
    // 215 violet
    *ptr++ = 238;
    *ptr++ = 130;
    *ptr++ = 238;
    // 216 wheat
    *ptr++ = 179;
    *ptr++ = 222;
    *ptr++ = 245;
  }
}

