// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "contour.h"
#include "framebase.h"

Contour::Contour(FrameBase* p, const char* c, int w, int d, int m) 
  : parent(p), lineWidth(w), dash(d)
{
  colorName = dupstr(c);
  color = parent->getColor(colorName);
  scale = NULL;
  method = (Method)m;

  dashlist[0] =8;
  dashlist[1] =3;

  previous_ = NULL;
  next_ = NULL;
}

Contour::Contour(FrameBase* p, const char* c, int w, int d,
		 const List<Vertex>& cont) 
  : parent(p), lineWidth(w), dash(d)
{
  contours_ = cont;
  colorName = dupstr(c);
  color = parent->getColor(colorName);
  scale = NULL;
  method = BLOCK;

  dashlist[0] =8;
  dashlist[1] =3;

  previous_ = NULL;
  next_ = NULL;
}

Contour::~Contour()
{
  if (colorName)
    delete [] colorName;

  if (scale)
    delete scale;
}

void Contour::render(Pixmap pmap, const Matrix& mx, int width, int height)
{
  if (contours_.head()) {
    XSetForeground(parent->display, parent->contourGC, color);
    int ww = lineWidth>=1 ? lineWidth : 1;

    if (!dash)
      XSetLineAttributes(parent->display, parent->contourGC, 
			 ww, LineSolid, CapButt, JoinMiter);
    else {
      char dlist[2];
      dlist[0] = dashlist[0];
      dlist[1] = dashlist[1];
      XSetDashes(parent->display, parent->contourGC, 0, dlist, 2);
      XSetLineAttributes(parent->display, parent->contourGC, 
			 ww, LineOnOffDash, CapButt, JoinMiter);
    }

    BBox bb = BBox(0, 0, width, height);

    Vector u1 = contours_.current()->vector;
    while (contours_.next()) {
      Vector u2 = contours_.current()->vector;
      if (u1[0] != DBL_MAX && u2[0] != DBL_MAX) {
	Vector v1 = (u1 * mx).round();
	Vector v2 = (u2 * mx).round();

 	if (bb.isIn(v1) || bb.isIn(v2))
 	  XDrawLine(parent->display, pmap, parent->contourGC, 
 		    (int)v1[0], (int)v1[1], (int)v2[0], (int)v2[1]);
      }

      u1 = u2;
    }
  }
}

void Contour::ps(int mode)
{
  if (contours_.head()) {
    {
      ostringstream str;
      switch ((Widget::PSColorSpace)mode) {
      case Widget::BW:
      case Widget::GRAY:
	psColorGray(parent->getXColor(colorName), str);
	str << " setgray";
	break;
      case Widget::RGB:
	psColorRGB(parent->getXColor(colorName), str);
	str << " setrgbcolor";
	break;
      case Widget::CMYK:
	psColorCMYK(parent->getXColor(colorName), str);
	str << " setcmykcolor";
	break;
      }
      str << endl << ends;
      Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
    }  
  
    if (!dash) {
      ostringstream str;
      str << lineWidth << " setlinewidth" << endl;
      str << "[] 0 setdash" << endl << ends;
      Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
    }
    else {
      ostringstream str;
      str << lineWidth << " setlinewidth" << endl;
      str << "[" << dashlist[0] << " " << dashlist[1] << "] 0 setdash" << endl << ends;
      Tcl_AppendResult(parent->interp, (char*)str.str().c_str(), NULL);
    }

    // output stroke paths

    int first = 1;
    do {
    ostringstream str;
      Vector u = contours_.current()->vector;
      if (u[0] != DBL_MAX) {
	Vector v = u * parent->refToCanvas;
	if (first) {
	  str << "newpath " << endl
	      << v.TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
	  first = 0;
	}
	else
	  str << v.TkCanvasPs(parent->canvas) << " lineto" << endl << ends;
      }
      else {
	str << "stroke" << endl << ends;
	first = 1;
      }
      Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
    } while (contours_.next());
  }
}

#ifdef _MACOSX
void Contour::macosx()
{
  if (contours_.head()) {
    // color, width, and dash
    macosxColor(parent->getXColor(colorName));
    macosxWidth(lineWidth);
    if (dash) {
      float dlist[2];
      dlist[0] = dashlist[0];
      dlist[1] = dashlist[1];
      macosxDash(dlist,2);
    }

    Vector u1 = contours_.current()->vector;
    while (contours_.next()) {
      Vector u2 = contours_.current()->vector;
      if (u1[0] != DBL_MAX && u2[0] != DBL_MAX) {
	Vector v1 = u1 * parent->refToCanvas;
	Vector v2 = u2 * parent->refToCanvas;
	
	macosxDrawLine(v1,v2);
      }

      u1 = u2;
    }
  }
}
#endif

#ifdef _WIN32
void Contour::win32()
{
  if (contours_.head()) {
    // color and width
    win32Color(parent->getXColor(colorName));
    win32Width(lineWidth);
    if (dash) {
      float dlist[2];
      dlist[0] = dashlist[0];
      dlist[1] = dashlist[1];
      win32Dash(dlist,2);
    }

    Vector u1 = contours_.current()->vector;
    while (contours_.next()) {
      Vector u2 = contours_.current()->vector;
      if (u1[0] != DBL_MAX && u2[0] != DBL_MAX) {
	Vector v1 = u1 * parent->refToCanvas;
	Vector v2 = u2 * parent->refToCanvas;
	
	win32DrawLine(v1,v2);
      }

      u1 = u2;
    }
  }
}
#endif

void Contour::updateCoords(const Matrix& mx)
{
  if (contours_.head())
    do {
      Vector& v = contours_.current()->vector;
      if (v[0] != DBL_MAX && v[0] != DBL_MAX)
	v *= mx;
    } while (contours_.next());
}

void Contour::setLineWidth(int w)
{
  lineWidth = w;
}

void Contour::setDash(int d)
{
  dash = d;
}

void Contour::setColor(const char* clr)
{
  if (colorName)
    delete [] colorName;

  colorName = dupstr(clr);
  color = parent->getColor(colorName);
}

