// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "coord.h"
#include "util.h"

char *coordSystemStr_[] = {
  "image",
  "physical",
  "amplifier",
  "detector",
  "wcs",
  "wcsA",
  "wcsB",
  "wcsC",
  "wcsD",
  "wcsE",
  "wcsF",
  "wcsG",
  "wcsH",
  "wcsI",
  "wcsJ",
  "wcsK",
  "wcsL",
  "wcsM",
  "wcsN",
  "wcsO",
  "wcsP",
  "wcsQ",
  "wcsR",
  "wcsS",
  "wcsT",
  "wcsU",
  "wcsV",
  "wcsW",
  "wcsX",
  "wcsY",
  "wcsZ",
  "wcs0"
};

char *skyFrameStr_[] = { 
  "fk4",
  "fk5",
  "icrs",
  "galactic",
  "ecliptic"
};

char *skyFormatStr_[] = { 
  "degrees",
  "sexagesimal",
  "hms",
  "arcmin",
  "arcsec"
};

void listCoordSystem(ostream& str, CoordSystem sys, SkyFrame sky, 
		     int wcs, int wcsequ)
{
  switch (sys) {
  case IMAGE:
    str << "image";
    return;
  case PHYSICAL:
    str << "physical";
    return;
  case DETECTOR:
    str << "detector";
    return;
  case AMPLIFIER:
    str << "amplifier";
    return;
  default:
    {
      if (wcs)
	if (sys == WCS0)
	  str << "wcs0" << ';';
	else if (sys > WCS)
	  str << "wcs" << (char)(sys-WCS+'`') << ';';
      
      if (wcsequ)
	switch (sky) {
	case FK4:
	  str << "fk4";
	  return;
	case FK5:
	  str << "fk5";
	  return;
	case ICRS:
	  str << "icrs";
	  return;
	case GALACTIC:
	  str << "galactic";
	  return;
	case ECLIPTIC:
	  str << "ecliptic";
	  return;
	}
      else {
	str << "linear";
	return;
      }
    }
  }
}

void listDistSystem(ostream& str, CoordSystem sys, SkyFormat format, int wcsequ)
{
  switch (sys) {
  case IMAGE:
    str << "image";
    return;
  case PHYSICAL:
    str << "physical";
    return;
  case DETECTOR:
    str << "detector";
    return;
  case AMPLIFIER:
    str << "amplifier";
    return;
  default:
    if (wcsequ)
      switch (format) {
      case DEGREES:
	str << "degrees";
	return;
      case ARCMIN:
	str << "arcmin";
	return;
      case ARCSEC:
	str << "arcsec";
	return;
      }
    else {
      str << "pixels";
      return;
    }
  }
}

void listProsCoordSystem(ostream& str, CoordSystem sys, SkyFrame sky)
{
  switch (sys) {
  case IMAGE:
    str << "logical";
    return;
  case PHYSICAL:
    str << "physical";
    return;
  case DETECTOR:
    str << "detector";
    return;
  case AMPLIFIER:
    str << "amplifier";
    return;
  default:
    switch (sky) {
    case FK4:
      str << "b1950";
      return;
    case FK5:
      str << "j2000";
      return;
    case ICRS:
      str << "icrs";
      return;
    case GALACTIC:
      str << "galactic";
      return;
    case ECLIPTIC:
      str << "ecliptic";
      return;
    default:
      str << "linear";
      return;
    }
  }
}

void strToCoordSystem(const char* str, CoordSystem* sys, SkyFrame* sky)
{
  if (!str) {
    *sys = PHYSICAL;
    *sky = FK5;
  }
  else if (STRCMP(str,"image",5)) {
    *sys = IMAGE;
    *sky = FK5;
    }
  else if (STRCMP(str,"physical",8)) {
    *sys = PHYSICAL;
    *sky = FK5;
  }
  else if (STRCMP(str,"amplifier",9)) {
    *sys = AMPLIFIER;
    *sky = FK5;
  }
  else if (STRCMP(str,"detector",9)) {
    *sys = DETECTOR;
    *sky = FK5;
  }
  else if (STRCMP(str,"fk4",3)) {
    *sys = WCS;
    *sky = FK4;
  }
  else if (STRCMP(str,"fk5",3)) {
    *sys = WCS;
    *sky = FK5;
  }
  else if (STRCMP(str,"icrs",4)) {
    *sys = WCS;
    *sky = ICRS;
  }
  else if (STRCMP(str,"galactic",8)) {
    *sys = WCS;
    *sky = GALACTIC;
  }
  else if (STRCMP(str,"ecliptic",8)) {
    *sys = WCS;
    *sky = ECLIPTIC;
  }
  else if (STRCMP(str,"linear",6)) {
    *sys = WCS;
    *sky = FK5;
  }
  else {
    *sys = PHYSICAL;
    *sky = FK5;
  }
}

void strToSkyFormat(const char* str, SkyFormat* format)
{
  if (!str)
    *format = DEGREES;
  else if (STRCMP(str,"degrees",7))
    *format = DEGREES;
  else if (STRCMP(str,"degree",6))
    *format = DEGREES;
  else if (STRCMP(str,"deg",3))
    *format = DEGREES;
  else if (STRCMP(str,"hms",3))
    *format = SEXAGESIMAL;
  else if (STRCMP(str,"h:m:s",5))
    *format = SEXAGESIMAL;
  else if (STRCMP(str,"\"h:m:s\"",7))
    *format = SEXAGESIMAL;
  else if (STRCMP(str,"dms",3))
    *format = SEXAGESIMAL;
  else if (STRCMP(str,"d:m:s",5))
    *format = SEXAGESIMAL;
  else if (STRCMP(str,"\"d:m:s\"",7))
    *format = SEXAGESIMAL;
  else if (STRCMP(str,"arcminute",9))
    *format = ARCMIN;
  else if (STRCMP(str,"arcmin",6))
    *format = ARCMIN;
  else if (STRCMP(str,"arcsecond",9))
    *format = ARCSEC;
  else if (STRCMP(str,"arcsec",6))
    *format = ARCSEC;
  else
    *format = DEGREES;
}

void strToAngleFormat(const char* str, AngleFormat* format)
{
  if (!str)
    *format = DEGREE;
  else if (STRCMP(str,"degrees",7))
    *format = DEGREE;
  else if (STRCMP(str,"degree",6))
    *format = DEGREE;
  else if (STRCMP(str,"deg",3))
    *format = DEGREE;
  else if (STRCMP(str,"radian",6))
    *format = RADIAN;
  else if (STRCMP(str,"rad",3))
    *format = RADIAN;
  else
    *format = DEGREE;
}

void strToDistSystem(const char* str, CoordSystem* sys, SkyFormat* format)
{
  if (!str) {
    *sys = PHYSICAL;
    *format = DEGREES;
  }
  else if (STRCMP(str,"image",5)) {
    *sys = IMAGE;
    *format = DEGREES;
    }
  else if (STRCMP(str,"physical",8)) {
    *sys = PHYSICAL;
    *format = DEGREES;
  }
  else if (STRCMP(str,"amplifier",9)) {
    *sys = AMPLIFIER;
    *format = DEGREES;
  }
  else if (STRCMP(str,"detector",8)) {
    *sys = DETECTOR;
    *format = DEGREES;
  }
  else if (STRCMP(str,"degrees",7)) {
    *sys = WCS;
    *format = DEGREES;
  }
  else if (STRCMP(str,"arcmin",6)) {
    *sys = WCS;
    *format = ARCMIN;
  }
  else if (STRCMP(str,"arcsec",6)) {
    *sys = WCS;
    *format = ARCSEC;
  }
  else if (STRCMP(str,"pixels",6)) {
    *sys = PHYSICAL;
    *format = DEGREES;
  }
  else {
    *sys = PHYSICAL;
    *format = DEGREES;
  }
}

