// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "fitsimage.h"
#include "framebase.h"

#include "mmap.h"
#include "smmap.h"
#include "mmapincr.h"
#include "alloc.h"
#include "allocgz.h"
#include "channel.h"
#include "share.h"
#include "sshare.h"
#include "socket.h"
#include "socketgz.h"
#include "var.h"
#include "iis.h"
#include "hist.h"
#include "compress.h"
#include "analysis.h"

// wcs coordinate system strings (for use with wcssubs)

// this is kluge to speed up doug minks wcssubs 'ksearch' routine
extern "C" {
  FitsHead* wcshead = NULL;
  char* ksearchh(char*, char*);
  char* ksearchh(char* hstring, char* keyword)
  /* Find entry for keyword keyword in FITS header string hstring.
     (the keyword may have a maximum of eight letters)
     NULL is returned if the keyword is not found */

  /*const char *hstring; character string containing fits-style header
		information in the format <keyword>= <value> {/ <comment>}
		the default is that each entry is 80 characters long;
		however, lines may be of arbitrary length terminated by
		nulls, carriage returns or linefeeds, if packed is true.  */
  /*const char *keyword; character string containing the name of the variable
		to be returned.  ksearch searches for a line beginning
		with this string.  The string may be a character
		literal or a character variable terminated by a null
		or '$'.  it is truncated to 8 characters. */
  {
    const char *headlast;
    char *loc, *headnext, *pval, *lc, *line;
    int icol, nextchar, lkey, nleft, lhead, lmax;

#ifdef USE_SAOLIB
    int iel=1, ip=1, nel, np, ier;
    char *get_fits_head_str();

    if( !use_saolib ){
#endif

    pval = 0;
    static int lhead0 = 0;
    /* Find current length of header string */
    if (lhead0)
      lmax = lhead0;
    else
      lmax = 256000;
    for (lhead = 0; lhead < lmax; lhead++) {
      if (hstring[lhead] == (char) 0)
        break;
    }

    /* Search header string for variable name */
    headlast = hstring + lhead;
    headnext = (char *) hstring;
    pval = NULL;
    while (headnext < headlast) {
      nleft = headlast - headnext;
      loc = strncsrch (headnext, keyword, nleft);

      /* Exit if keyword is not found */
      if (loc == NULL) {
        break;
      }

      icol = (loc - hstring) % 80;
      lkey = strlen (keyword);
      nextchar = (int) *(loc + lkey);

      /* If this is not in the first 8 characters of a line, keep searching */
      if (icol > 7)
        headnext = loc + 1;

      /* If parameter name in header is longer, keep searching */
      else if (nextchar != 61 && nextchar > 32 && nextchar < 127)
        headnext = loc + 1;

      /* If preceeding characters in line are not blanks, keep searching */
      else {
        line = loc - icol;
	for (lc = line; lc < loc; lc++) {
	  if (*lc != ' ')
            headnext = loc + 1;
        }

	/* Return pointer to start of line if match */
	if (loc >= headnext) {
	  pval = line;
	  break;
	}
      }
    }

    /* Return pointer to calling program */
    return (pval);

#ifdef USE_SAOLIB
    }
    else {
      if (get_fits_head_str(keyword,iel,ip,&nel,&np,&ier,hstring) != NULL)
        return(hstring);
      else
        return(NULL);
    }
#endif
  }

  char* findit(char* cards, char* key)
  {
    if (wcshead)
      return wcshead->find(key);
    else
      return ksearchh(cards, key);
  }
}

FitsImage::FitsImage(Base* p)
{
  parent = p;
  objectName = NULL;
  rootBaseFileName = NULL;
  fullBaseFileName = NULL;
  rootFileName = NULL;
  fullFileName = NULL;
  iisFileName = NULL;

  fits_ = NULL;
  compress_ = NULL;
  hist_ = NULL;

  base_ = NULL;
  basedata_ = NULL;

  smooth_ = NULL;
  smoothdata_ = NULL;

  image_ = NULL;
  data_ = NULL;

  nextMosaic_ = NULL;
  nextSlice_ = NULL;

  width_ = 0;
  height_ = 0;
  depth_ = 1;
  bitpix_ = 0;

  doSmooth_ = 0;
  smoothFunction_ = GAUSSIAN;
  smoothRadius_ = 3;

  binFunction_ = FitsHist::SUM;
  binFactor_ = Vector(1,1);
  binBufferSize_ = 1024;
  binDepth_ = 1;

  keyLTMV = 0;
  keyATMV = 0;
  keyDTMV = 0;
  keyDATASEC = 0;

  imageToData = Translate(-.5, -.5);
  dataToImage = Translate( .5,  .5);

  wcs = NULL;

  iisMode = 0;
  iiszt = 0;

  mWidgetToData = widgetToData.mm();
  mPannerToData = pannerToData.mm();
  mMagnifierToData = magnifierToData.mm();
  mPSToData = PSToData.mm();

  mWidgetToData3d = widgetToData3d.mm();
  mPannerToData3d = pannerToData3d.mm();
  mPSToData3d = PSToData3d.mm();
}

FitsImage::~FitsImage()
{
  if (objectName)
    delete [] objectName;

  if (rootBaseFileName)
    delete [] rootBaseFileName;

  if (fullBaseFileName)
    delete [] fullBaseFileName;

  if (rootFileName)
    delete [] rootFileName;

  if (fullFileName)
    delete [] fullFileName;

  if (iisFileName)
    delete [] iisFileName;

  if (fits_)
    delete fits_;

  if (compress_)
    delete compress_;

  if (hist_)
    delete hist_;

  if (basedata_)
    delete basedata_;

  if (smooth_)
    delete smooth_;

  if (smoothdata_)
    delete smoothdata_;

  if (wcs) {
    for (int i=0; i<MULTWCS; i++)
      if (wcs[i])
	wcsfree(wcs[i]);
    delete [] wcs;
  }  
}

// Fits

FitsImageFitsAlloc::FitsImageFitsAlloc(Base* p, 
				       const char* ch, const char* fn, 
				       FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsFitsAlloc(ch, FitsFile::RELAX, flush);
  process(fn);
}

FitsImageFitsAllocGZ::FitsImageFitsAllocGZ(Base* p, 
					   const char* ch, const char* fn, 
					   FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsFitsAllocGZ(ch, FitsFile::RELAX, flush);
  process(fn);
}

FitsImageFitsChannel::FitsImageFitsChannel(Base* p, Tcl_Interp* interp, 
					   const char* ch, const char* fn, 
					   FitsFile::FlushMode flush) 
  : FitsImage(p)
{
  fits_ = new FitsFitsChannel(interp, ch, fn, FitsFile::RELAX, flush);
  process(fn);
}

FitsImageFitsMMap::FitsImageFitsMMap(Base* p, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsFitsMMap(fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsSMMap::FitsImageFitsSMMap(Base* p, const char* hdr, 
				       const char* fn) : FitsImage(p)
{
  fits_ = new FitsFitsSMMap(hdr, fn);
  process(fn);
}

FitsImageFitsMMapIncr::FitsImageFitsMMapIncr(Base* p, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsFitsMMapIncr(fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsShare::FitsImageFitsShare(Base* p, Base::ShmType type, 
				       int id, const char* fn)
  : FitsImage(p)
{
  switch (type) {
  case Base::SHMID:
    fits_ = new FitsFitsShareID(id, fn, FitsFile::RELAX);
    break;
  case Base::KEY:
    fits_ = new FitsFitsShareKey(id, fn, FitsFile::RELAX);
    break;
  }
  process(fn);
}

FitsImageFitsSShare::FitsImageFitsSShare(Base* p, Base::ShmType type,
					 int hdr, int id, const char* fn)
  : FitsImage(p)
{
  switch (type) {
  case Base::SHMID:
    fits_ = new FitsFitsSShareID(hdr, id, fn);
    break;
  case Base::KEY:
    fits_ = new FitsFitsSShareKey(hdr, id, fn);
    break;
  }
  process(fn);
}

FitsImageFitsSocket::FitsImageFitsSocket(Base* p, int s, const char* fn, 
					 FitsFile::FlushMode flush) 
  : FitsImage(p)
{
  fits_ = new FitsFitsSocket(s, fn, FitsFile::RELAX, flush);
  process(fn);
}

FitsImageFitsSocketGZ::FitsImageFitsSocketGZ(Base* p, int s, 
					     const char* fn, 
					     FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsFitsSocketGZ(s, fn, FitsFile::RELAX, flush);
  process(fn);
}

FitsImageFitsVar::FitsImageFitsVar(Base* p, Tcl_Interp* interp, 
				   const char* var, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsFitsVar(interp, var, fn, FitsFile::RELAX);
  process(fn);
}

// Fits Next

FitsImageFitsNextAlloc::FitsImageFitsNextAlloc(Base* p, 
					       const char* fn,
					       FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextAlloc(prev);
  process(fn);
}

FitsImageFitsNextAllocGZ::FitsImageFitsNextAllocGZ(Base* p, 
						   const char* fn,
						   FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextAllocGZ(prev);
  process(fn);
}

FitsImageFitsNextChannel::FitsImageFitsNextChannel(Base* p, 
						   const char* fn,
						   FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextChannel(prev);
  process(fn);
}

FitsImageFitsNextMMap::FitsImageFitsNextMMap(Base* p, 
					     const char* fn,
					     FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextMMap(prev);
  process(fn);
}

FitsImageFitsNextSMMap::FitsImageFitsNextSMMap(Base* p, 
					       const char* fn,
					       FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextSMMap(prev);
  process(fn);
}

FitsImageFitsNextMMapIncr::FitsImageFitsNextMMapIncr(Base* p, 
						     const char* fn,
						     FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextMMapIncr(prev);
  process(fn);
}

FitsImageFitsNextShare::FitsImageFitsNextShare(Base* p, 
					       const char* fn,
					       FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextShare(prev);
  process(fn);
}

FitsImageFitsNextSShare::FitsImageFitsNextSShare(Base* p, 
						 const char* fn,
						 FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextSShare(prev);
  process(fn);
}

FitsImageFitsNextSocket::FitsImageFitsNextSocket(Base* p, 
						 const char* fn,
						 FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextSocket(prev);
  process(fn);
}

FitsImageFitsNextSocketGZ::FitsImageFitsNextSocketGZ(Base* p, 
						     const char* fn, 
						     FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextSocketGZ(prev);
  process(fn);
}

FitsImageFitsNextVar::FitsImageFitsNextVar(Base* p, 
					   const char* fn,
					   FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsFitsNextVar(prev);
  process(fn);
}

FitsImageFitsNextHist::FitsImageFitsNextHist(Base* p, 
					     FitsImage* fi,
					     FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsHistNext(prev);
  process(NULL);

  fits_->setpFilter(fi->getHistFilter());
  rootBaseFileName = dupstr(fi->getRootBaseFileName());
  fullBaseFileName = dupstr(fi->getFullBaseFileName());
  iisFileName = dupstr(fi->getFullBaseFileName());
}

FitsImageFitsNextCompress::FitsImageFitsNextCompress(Base* p, 
						     FitsImage* fi,
						     FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsCompressNext(prev);
  process(NULL);

  rootBaseFileName = dupstr(fi->getRootBaseFileName());
  fullBaseFileName = dupstr(fi->getFullBaseFileName());
  iisFileName = dupstr(fi->getFullBaseFileName());
}

// Array

FitsImageArrAlloc::FitsImageArrAlloc(Base* p, 
				     const char* ch, const char* fn,
				     FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsArrAlloc(ch, flush);
  process(fn);
}

FitsImageArrAllocGZ::FitsImageArrAllocGZ(Base* p, 
					 const char* ch, const char* fn,
					 FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsArrAllocGZ(ch, flush);
  process(fn);
}

FitsImageArrChannel::FitsImageArrChannel(Base* p, Tcl_Interp* interp, 
					 const char* ch, const char* fn,
					 FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsArrChannel(interp, ch, fn, flush);
  process(fn);
}

FitsImageArrMMap::FitsImageArrMMap(Base* p, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsArrMMap(fn);
  process(fn);
}

FitsImageArrMMapIncr::FitsImageArrMMapIncr(Base* p, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsArrMMapIncr(fn);
  process(fn);
}

FitsImageArrShare::FitsImageArrShare(Base* p, Base::ShmType type, 
				     int id, const char* fn) 
  : FitsImage(p)
{
  switch (type) {
  case Base::SHMID:
    fits_ = new FitsArrShareID(id, fn);
    break;
  case Base::KEY:
    fits_ = new FitsArrShareKey(id, fn);
    break;
  }
  process(fn);
}

FitsImageArrSocket::FitsImageArrSocket(Base* p, int s, const char* fn,
				       FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsArrSocket(s, fn, flush);
  process(fn);
}

FitsImageArrSocketGZ::FitsImageArrSocketGZ(Base* p, int s, const char* fn,
					   FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsArrSocketGZ(s, fn, flush);
  process(fn);
}

FitsImageArrVar::FitsImageArrVar(Base* p, Tcl_Interp* interp, 
				 const char* var, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsArrVar(interp, var, fn);
  process(fn);
}

// Mosaic

FitsImageMosaicAlloc::FitsImageMosaicAlloc(Base* p, 
					   const char* ch, 
					   const char* fn,
					   FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsMosaicAlloc(ch, flush);
  process(fn);
}

FitsImageMosaicAllocGZ::FitsImageMosaicAllocGZ(Base* p, 
					       const char* ch, 
					       const char* fn,
					       FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsMosaicAllocGZ(ch, flush);
  process(fn);
}

FitsImageMosaicChannel::FitsImageMosaicChannel(Base* p, 
					       Tcl_Interp* interp,
					       const char* ch, 
					       const char* fn,
					       FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsMosaicChannel(interp, ch, flush);
  process(fn);
}

FitsImageMosaicMMap::FitsImageMosaicMMap(Base* p, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsMosaicMMap(fn);
  process(fn);
}

FitsImageMosaicMMapIncr::FitsImageMosaicMMapIncr(Base* p, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsMosaicMMapIncr(fn);
  process(fn);
}

FitsImageMosaicShare::FitsImageMosaicShare(Base* p, 
					   Base::ShmType type,
					   int id, const char* fn)
  : FitsImage(p)
{
  switch (type) {
  case Base::SHMID:
    fits_ = new FitsMosaicShareID(id);
    break;
  case Base::KEY:
    fits_ = new FitsMosaicShareKey(id);
    break;
  }
  process(fn);
}

FitsImageMosaicSocket::FitsImageMosaicSocket(Base* p, 
					     int s, const char* fn,
					     FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsMosaicSocket(s, flush);
  process(fn);
}

FitsImageMosaicSocketGZ::FitsImageMosaicSocketGZ(Base* p, 
						 int s, const char* fn,
						 FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsMosaicSocketGZ(s, flush);
  process(fn);
}

FitsImageMosaicVar::FitsImageMosaicVar(Base* p, Tcl_Interp* interp,
				       const char* var, const char* fn)
  : FitsImage(p)
{
  fits_ = new FitsMosaicVar(interp, var, fn);
  process(fn);
}

// Mosaic Next

FitsImageMosaicNextAlloc::FitsImageMosaicNextAlloc(Base* p, 
						   const char* fn,
						   FitsFile* prev,
						   FitsFile::FlushMode flush)

  : FitsImage(p)
{
  fits_ = new FitsMosaicNextAlloc(prev, flush);
  process(fn);
}

FitsImageMosaicNextAllocGZ::FitsImageMosaicNextAllocGZ(Base* p,
						       const char* fn,
						       FitsFile* prev,
						     FitsFile::FlushMode flush)

  : FitsImage(p)
{
  fits_ = new FitsMosaicNextAllocGZ(prev, flush);
  process(fn);
}

FitsImageMosaicNextChannel::FitsImageMosaicNextChannel(Base* p, 
						       const char* fn, 
						       FitsFile* prev,
						     FitsFile::FlushMode flush)
  : FitsImage(p)
{
  fits_ = new FitsMosaicNextChannel(prev, flush);
  process(fn);
}

FitsImageMosaicNextMMap::FitsImageMosaicNextMMap(Base* p, const char* fn, 
						 FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsMosaicNextMMap(prev);
  process(fn);
}

FitsImageMosaicNextMMapIncr::FitsImageMosaicNextMMapIncr(Base* p,
							 const char* fn, 
							 FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsMosaicNextMMapIncr(prev);
  process(fn);
}

FitsImageMosaicNextShare::FitsImageMosaicNextShare(Base* p, 
						   const char* fn,
						   FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsMosaicNextShare(prev);
  process(fn);
}

FitsImageMosaicNextSocket::FitsImageMosaicNextSocket(Base* p, 
						     const char* fn, 
						     FitsFile* prev,
						     FitsFile::FlushMode flush)

  : FitsImage(p)
{
  fits_ = new FitsMosaicNextSocket(prev, flush);
  process(fn);
}

FitsImageMosaicNextSocketGZ::FitsImageMosaicNextSocketGZ(Base* p,
							 const char* fn, 
							 FitsFile* prev,
						     FitsFile::FlushMode flush)

  : FitsImage(p)
{
  fits_ = new FitsMosaicNextSocketGZ(prev, flush);
  process(fn);
}

FitsImageMosaicNextVar::FitsImageMosaicNextVar(Base* p, 
					       const char* fn, 
					       FitsFile* prev)
  : FitsImage(p)
{
  fits_ = new FitsMosaicNextVar(prev);
  process(fn);
}

// IIS

FitsImageIIS::FitsImageIIS(Base* p, int w, int h) : FitsImage(p)
{
  fits_ = new FitsIIS(w, h);
  process("");
  iisMode = 1;
}

void FitsImageIIS::iisErase()
{
  if (fits_)
    ((FitsIIS*)fits_)->erase();
}

const char* FitsImageIIS::iisGet(int x, int y)
{
  if (fits_)
    return ((FitsIIS*)fits_)->get(x,y);
}

void FitsImageIIS::iisSet(int x, int y, const char* d, int l)
{
  if (fits_)
    ((FitsIIS*)fits_)->set(x,y,d,l);
}

void FitsImageIIS::iisWCS(const Matrix& mx, const Vector& z, int zt)
{
  dataToImage = Translate(-.5,-.5) * mx;
  imageToData = dataToImage.invert();
  iisz = z;
  iiszt = zt;
}

// FitsImage

void FitsImage::analysis()
{
  image_ = base_;
  data_ = basedata_;

  FitsHead* head = image_->head();
  width_ = head->naxis(1);
  height_ = head->naxis(2);
  depth_ = head->naxis(3);
  if (!depth_)
    depth_ = 1;
  bitpix_ = head->bitpix();

 if (smooth_)
    delete smooth_;
  smooth_ = NULL;
  if (smoothdata_)
    delete smoothdata_;
  smoothdata_ = NULL;

  if (parent->doSmooth_) {
    smooth_ = new FitsAnalysis(image_);
    if (smooth_->isValid()) {
      smoothdata_ = new FitsDatam<float>(smooth_, parent);

      if (smooth()) {
	image_ = smooth_;
	data_ = smoothdata_;

	FitsHead* head = image_->head();
	width_ = head->naxis(1);
	height_ = head->naxis(2);
	depth_ = head->naxis(3);
	if (!depth_)
	  depth_ = 1;
	bitpix_ = head->bitpix();
      }
      else {
	delete smooth_;
	smooth_ = NULL;
	delete smoothdata_;
	smoothdata_ = NULL;
      }
    }
    else {
      delete smooth_;
      smooth_ = NULL;
    }
  }
}

void FitsImage::appendWCS(FitsHead* app)
{
  // process OBJECT keyword

  char* obj = app->getString("OBJECT");
  if (obj) {
    if (objectName)
      delete [] objectName;
    objectName = obj;
  }

  // Process WCS keywords
  FitsHead* head = image_->head();

  // append wcs keywords to the end of the header
  int ll = head->headbytes()+app->headbytes();
  char* cards = new char[ll];

  // copy default wcs
  memcpy(cards, head->cards(), head->headbytes());

  // find first END and zero out
  for (int i=0; i<head->headbytes(); i+=80)
    if (!strncmp(cards+i,"END",3)) {
      memcpy(cards+i, "   ",3);
      break;
    }

  // copy appended wcs
  memcpy(cards+head->headbytes(), app->cards(), app->headbytes());


  FitsHead* hhead = new FitsHead(cards,ll,FitsHead::EXTERNAL);
  initWCS(hhead);
  delete hhead;
  delete [] cards;
}

char* FitsImage::display(FitsHead* head)
{
  int size = head->ncard() * (FTY_CARDLEN+1);
  char* lbuf = new char[size+1];

  char* lptr = lbuf;
  char* cptr = head->cards();
  for (int i=0; i<head->ncard(); i++,cptr+=FTY_CARDLEN) {
    memcpy(lptr, cptr, FTY_CARDLEN);
    lptr+=FTY_CARDLEN;
    *(lptr++) = '\n';
  }

  lbuf[size] = '\0';
  return lbuf;
}

char* FitsImage::displayHeader()
{
  if (DebugBin || DebugCompress)
    return display(image_->head());
  else
    return display(fits_->head());
}

char* FitsImage::displayPrimary()
{
  return display(fits_->primary());
}

int FitsImage::findKeyword(const char* key)
{
  return fits_->find(key);
}

char* FitsImage::getCards()
{
  if (image_) {
    FitsHead* head = image_->head();
    if (head)
      return head->cards();
  }
  return NULL;
}

FitsBound* FitsImage::getDataParams(FrScale::ScanMode which)
{
  // return in 0 to width_, 0 to height_
  switch (which) {
  case FrScale::NODATASEC:
    return &iparams;
  case FrScale::DATASEC:
    return &dparams;
  case FrScale::UNODATASEC:
    return &irparams;
  case FrScale::UDATASEC:
    return &drparams;
  }
}

FitsHead* FitsImage::getHead()
{
  if (image_)
    return image_->head();
  else    
    return NULL;
}

BBox FitsImage::getIRAFBBox()
{
  // return in IMAGE coords
  // expand by .5 to move from center of pixel to edge
  Vector ll = (detsize.ll-Vector(.5,.5)) * 
    imageToData * Scale(ccdsum) * dataToImage;
  Vector ur = (detsize.ur+Vector(.5,.5)) * 
    imageToData * Scale(ccdsum) * dataToImage;
  return BBox(ll,ur);
}

char* FitsImage::getKeyword(const char* key)
{
  return fits_->getKeyword(key);
}

FitsBound* FitsImage::getMinMaxParams(FrScale::ScanMode which)
{
  // return in IMAGE coords
  switch (which) {
  case FrScale::NODATASEC:
    return &iparams;
  case FrScale::DATASEC:
    return &dparams;
  case FrScale::UNODATASEC:
    return &imparams;
  case FrScale::UDATASEC:
    return &dmparams;
  }
}

int FitsImage::getNCards()
{
  if (image_) {
    FitsHead* head = image_->head();
    if (head)
      return head->ncard();
  }
  return 0;
}

const char* FitsImage::getValue(const Vector& v)
{
  if (!isIIS() || (iiszt != 1))
    return data_->getValue(v);
  else {
    double value = data_->getValueDouble(v);

    ostringstream str;
    if (value <= 1)
      str << '<' << iisz[0] << ends;
    else if (value >= 200)
      str << '>' << iisz[1] << ends;
    else
      str << ((value-1) * (iisz[1]-iisz[0]))/199 + iisz[0] << ends;
    memcpy(buf,str.str().c_str(), str.str().length());
    return buf;
  }
}

Vector FitsImage::getWCScdelt(CoordSystem sys)
{
  if (hasWCS(sys)) {
    int i = sys-WCS;

    if (!wcs[i]->coorflip)
      return Vector(wcs[i]->cdelt[0], wcs[i]->cdelt[1]);
    else
      return Vector(wcs[i]->cdelt[1], wcs[i]->cdelt[0]);
  }
  else
    return Vector();
}

Vector FitsImage::getWCScrpix(CoordSystem sys)
{
  if (hasWCS(sys))
      return Vector(wcs[sys-WCS]->crpix[0], wcs[sys-WCS]->crpix[1]);
  else
    return Vector();
}

Vector FitsImage::getWCScrval(CoordSystem sys)
{
  if (hasWCS(sys)) {
    int i = sys-WCS;

    if (!wcs[i]->coorflip)
      return Vector(wcs[i]->crval[0], wcs[i]->crval[1]);
    else
      return Vector(wcs[i]->crval[1], wcs[i]->crval[0]);
  }
  else
    return Vector();
}

Orientation FitsImage::getWCSOrientation(CoordSystem sys, SkyFrame sky)
{
  if (hasWCS(sys)) {
    Vector orval = pix2wcs(getWCScrpix(sys), sys, sky);
    Vector orpix = wcs2pix(orval, sys, sky);
    Vector delta = getWCScdelt(sys).abs();
    Vector npix = wcs2pix(Vector(orval[0],orval[1]+delta[1]), sys, sky);
    Vector north = npix - orpix;
    Vector epix = wcs2pix(Vector(orval[0]+delta[0],orval[1]), sys, sky);
    Vector east = epix - orpix;

    // take the cross product and see which way the 3rd axis is pointing
    double w = east[0]*north[1]-east[1]*north[0];

    if (!hasWCSEqu(sys))
      return w>0 ? NORMAL : XX;
    else
      return w<0 ? NORMAL : XX;
  }

  return NORMAL;
}

double FitsImage::getWCSRotation(CoordSystem sys, SkyFrame sky)
{
  if (hasWCS(sys)) {
    Vector orval = pix2wcs(getWCScrpix(sys), sys, sky);
    Vector orpix = wcs2pix(orval, sys, sky);
    Vector delta = getWCScdelt(sys).abs();
    Vector npix = wcs2pix(Vector(orval[0],orval[1]+delta[1]), sys, sky);
    Vector north = npix - orpix;

    if (wcs[sys-WCS]->imflip)
      return  (north.angle()-M_PI/2);
    else
      return -(north.angle()-M_PI/2);
  }

  return 0;
}

int FitsImage::hasWCS(CoordSystem sys)
{
  return (sys>=WCS && wcs && wcs[sys-WCS]) ? 1 : 0;
}

int FitsImage::hasWCSEqu(CoordSystem sys)
{
  if (hasWCS(sys)) {
    const char* str = ::getradecsys(wcs[sys-WCS]);

    if (!strncmp("FK4", str, 3))
      return 1;
    else if (!strncmp("FK5", str, 3))
      return 1;
    else if (!strncmp("ICR", str, 3))
      return 1;
    else if (!strncmp("GAL", str, 3))
      return 1;
    else if (!strncmp("ECL", str, 3))
      return 1;
    else if (!strncmp("NPOLE", str, 4)) // special north polar angle
      return 1;
    else
      return 0;
  }
  else
    return 0;
}

int FitsImage::hasWCSLinear(CoordSystem sys)
{
  
  if (hasWCS(sys))
    return hasWCSEqu(sys) ? 0 : 1;
  else
    return 0;
}

void FitsImage::iisSetFileName(const char* fn)
{
  if (iisFileName)
    delete [] iisFileName;
  iisFileName = dupstr(fn);
}

void FitsImage::initWCS(FitsHead* head)
{
  if (wcs) {
    for (int i=0; i<MULTWCS; i++)
      if (wcs[i])
	wcsfree(wcs[i]);
    delete [] wcs;
  }

  // wcsinit is sloooowwww! so try to figure it out first
  wcs = new WorldCoor*[MULTWCS];
  for (int i=0; i<MULTWCS; i++)
    wcs[i] = NULL;
  
  // look first for default WCS. Let wcsinit figure it out since there can
  // be many different non-standard wcs's present
  wcshead = head;
  wcs[0] = wcsinit(head->cards());
  wcshead  = NULL;

  // now look for WCSA - WCSZ
  // we can take a short cut here, since only valid FITS wcs's are available
  for (int i=1; i<MULTWCS; i++) {
    char str[] = "CTYPE1 ";
    str[6] = '@'+i;
    if (head->find(str)) {
      wcshead = head;
      wcs[i] = wcsinitc(head->cards(),str+6);
      wcshead  = NULL;
    }
  }

  // now see if we have a 'physical' wcs, if so, set LTMV keywords
  for (int i=1; i<MULTWCS; i++) {
    if (wcs[i] && wcs[i]->wcsname && !strcmp(wcs[i]->wcsname,"PHYSICAL")) {
      keyLTMV = 1;

      double ltm11 = wcs[i]->cd[0] != 0 ? 1/wcs[i]->cd[0] : 0;
      double ltm12 = wcs[i]->cd[1] != 0 ? 1/wcs[i]->cd[1] : 0;
      double ltm21 = wcs[i]->cd[2] != 0 ? 1/wcs[i]->cd[2] : 0;
      double ltm22 = wcs[i]->cd[3] != 0 ? 1/wcs[i]->cd[3] : 0;

      double ltv1 = wcs[i]->crpix[0] -
	wcs[i]->crval[0]*ltm11 - wcs[i]->crval[1]*ltm21;
      double ltv2 = wcs[i]->crpix[1] -
	wcs[i]->crval[0]*ltm12 - wcs[i]->crval[1]*ltm22;

      physicalToImage = Matrix(ltm11, ltm12, ltm21, ltm22, ltv1, ltv2);
      imageToPhysical = physicalToImage.invert();
    }
  }

  if (DebugWCS) {
    for (int i=0; i<MULTWCS; i++) {
      if (wcs[i]) {
	int n = wcs[i]->naxes;
	int nn = n*n;

	cerr << "wcs " << (char)(!i ? ' ' : '@'+i) << endl;
	cerr << "wcs->wcsname " << (wcs[i]->wcsname ? wcs[i]->wcsname : "")
	     << endl;
	cerr << "wcs->naxes=" << wcs[i]->naxes << endl;
	cerr << "wcs->naxis=" << wcs[i]->naxis << endl;

	cerr << "wcs->equinox=" << wcs[i]->equinox << endl;
	cerr << "wcs->epoch=" << wcs[i]->epoch << endl;

	cerr << "wcs->c1type=" << wcs[i]->c1type << endl;
	cerr << "wcs->c2type=" << wcs[i]->c2type << endl;
	cerr << "wcs->ptype=" << wcs[i]->ptype << endl;

	for (int j=0; j<n; j++)
	  cerr << "wcs->crpix[" << j << "]=" << wcs[i]->crpix[j] << endl;
	for (int j=0; j<n; j++)
	  cerr << "wcs->crval[" << j << "]=" << wcs[i]->crval[j] << endl;
	for (int j=0; j<n; j++)
	  cerr << "wcs->cdelt[" << j << "]=" << wcs[i]->cdelt[j] << endl;

	for (int j=0; j<4; j++)
	  cerr << "wcs->cd[" << j << "]=" << wcs[i]->cd[j] << endl;
	for (int j=0; j<nn; j++)
	  cerr << "wcs->pc[" << j << "]=" << wcs[i]->pc[j] << endl;

	cerr << "wcs->longpole=" << wcs[i]->longpole << endl;
	cerr << "wcs->latpole=" << wcs[i]->latpole << endl;
	cerr << "wcs->prjcode=" << wcs[i]->prjcode << endl;

	cerr << "wcs->rot=" << wcs[i]->rot << endl;
	cerr << "wcs->imrot=" << wcs[i]->imrot << endl;
	cerr << "wcs->imflip=" << wcs[i]->imflip << endl;
	cerr << "wcs->pa_north=" << wcs[i]->pa_north << endl;
	cerr << "wcs->pa_east=" << wcs[i]->pa_east << endl;
	cerr << "wcs->coorflip=" << wcs[i]->coorflip << endl;

	cerr << "wcs->syswcs=" << wcs[i]->syswcs << endl;
	cerr << "wcs->wcsproj=" << wcs[i]->wcsproj << endl;
	cerr << "wcs->distcode=" << wcs[i]->distcode << endl;
      }
    }
  }
}

void FitsImage::initWCS0(const Vector& pix)
{
  if (wcs[0]) {
    Vector cc = mapFromRef(pix, IMAGE, FK5);
    WorldCoor* ww = wcs[0];
    wcs[WCS0-WCS] = wcskinit(ww->nxpix, ww->nypix, "RA---TAN", "DEC--TAN", cc[0], cc[1], 0, 0, ww->cd, 0, 0, 0, 2000, 0);
  }
}

void FitsImage::load()
{
  if (compress_)
    base_ = compress_;
  else if (hist_)
    base_ = hist_;
  else
    base_ = fits_;

  if (basedata_)
    delete basedata_;

  switch (base_->head()->bitpix()) {
  case 8:
    basedata_ = new FitsDatam<unsigned char>(base_, parent);
    break;
  case 16:
    basedata_ = new FitsDatam<short>(base_, parent);
    break;
  case -16:
    basedata_ = new FitsDatam<unsigned short>(base_, parent);
    break;
  case 32:
    basedata_ = new FitsDatam<int>(base_, parent);
    break;
  case 64:
    basedata_ = new FitsDatam<long long>(base_, parent);
    break;
  case -32:
    basedata_ = new FitsDatam<float>(base_, parent);
    break;
  case -64:
    basedata_ = new FitsDatam<double>(base_, parent);
    break;
  }

  analysis();

  resetWCS();

  iparams.reset();
  dparams.reset();
  imparams.reset();
  irparams.reset();
  dmparams.reset();
  drparams.reset();

  if (!processKeywords()) {
    reset();
    return;
  }
}

int FitsImage::parseSection(char* lbuf, Vector* v1, Vector* v2)
{
  double x1, y1, x2, y2;
  char d1,d2,d3,d4,d5; // dummy char
  string x(lbuf);
  istringstream str(x);
  str >> d1 >> x1 >> d2 >> x2 >> d3 >> y1 >> d4 >> y2 >> d5;

  // verify input
  if (!(d1=='[' && d2==':' && d3==',' && d4==':' && d5==']'))
    return 0;

  // it looks ok
  *v1 = Vector(x1,y1);
  *v2 = Vector(x2,y2);

  return 1;
}

char* FitsImage::pix2wcs(Vector in, CoordSystem sys, SkyFrame sky,
			 SkyFormat format, char* lbuf, int len)
{
  if (hasWCS(sys)) {
    int i = sys-WCS;

    if (hasWCSEqu(sys)) {
      if (sky == NATIVEWCS)
	wcsoutinit(wcs[i], ::getradecsys(wcs[i]));
      else
	wcsoutinit(wcs[i], skyFrameStr_[sky-FK4]);

      switch (format) {
      case ARCMIN:
      case ARCSEC:
      case DEGREES:
	setwcsdeg(wcs[i],1);
	wcs[i]->ndec = 5;
	break;
      case SEXAGESIMAL:
      case HMS:
	setwcsdeg(wcs[i],0);
	wcs[i]->ndec = 3;
	break;
      }
    }
    else {
      wcsoutinit(wcs[i], ::getradecsys(wcs[i]));
      setwcslin(wcs[i],2);
    }

    ::pix2wcst(wcs[i], in[0], in[1], lbuf, len);
  }

  return lbuf;
}

Vector FitsImage::pix2wcs(Vector in, CoordSystem sys)
{
  if (hasWCS(sys)) {
    int i = sys-WCS;

    wcsoutinit(wcs[i], ::getradecsys(wcs[i]));
    double x,y;
    ::pix2wcs(wcs[i], in[0], in[1], &x, &y);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::pix2wcs(Vector in, CoordSystem sys, SkyFrame sky)
{
  if (hasWCS(sys)) {
    int i = sys-WCS;

    if (hasWCSEqu(sys)) {
      if (sky == NATIVEWCS)
	wcsoutinit(wcs[i], ::getradecsys(wcs[i]));
      else
	wcsoutinit(wcs[i], skyFrameStr_[sky-FK4]);
    }
    else
      wcsoutinit(wcs[i], ::getradecsys(wcs[i]));

    double x,y;
    ::pix2wcs(wcs[i], in[0], in[1], &x, &y);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::pix2wcsDist(Vector in, CoordSystem sys)
{
  if (hasWCS(sys)) {
    Vector cd = getWCScdelt(sys);
    return Vector(in[0]*cd[0], in[1]*cd[1]).abs();
  }
  else
    return in;
}

void FitsImage::process(const char* fn)
{
  if (!fits_->isValid()) {
    reset();
    return;
  }

  if (fits_->isImage())
    load();
  else if (fits_->isBinTable()) {
    if (fits_->find("ZIMAGE")) {
      initCompress();
      if (!compress_->isValid()) {
	reset();
	return;
      }

      load();
    }
    else
      initHist();
  }

  // load() can call reset()
  if (fits_)
    setFileName(fn);
}

int FitsImage::processKeywords()
{
  // Physical to Image (LTM/LTV keywords) (with no wcsname already located)

  if (!keyLTMV) {
    if (image_->find("LTM1_1") ||
	image_->find("LTM1_2") ||
	image_->find("LTM2_1") ||
	image_->find("LTM2_2") ||
	image_->find("LTV1") ||
	image_->find("LTV2"))
      keyLTMV = 1;

    double ltm11 = image_->getReal("LTM1_1", 1);
    double ltm12 = image_->getReal("LTM1_2", 0);
    double ltm21 = image_->getReal("LTM2_1", 0);
    double ltm22 = image_->getReal("LTM2_2", 1);

    double ltv1 = image_->getReal("LTV1", 0);
    double ltv2 = image_->getReal("LTV2", 0);

    physicalToImage = Matrix(ltm11, ltm12, ltm21, ltm22, ltv1, ltv2);
    imageToPhysical = physicalToImage.invert();
  }

  // CDD to Detector (DTM/DTV keywords)

  if (image_->find("DTM1_1") ||
      image_->find("DTM1_2") ||
      image_->find("DTM2_1") ||
      image_->find("DTM2_2") ||
      image_->find("DTV1") ||
      image_->find("DTV2"))
    keyDTMV = 1;

  double dtm11 = image_->getReal("DTM1_1", 1);
  double dtm12 = image_->getReal("DTM1_2", 0);
  double dtm21 = image_->getReal("DTM2_1", 0);
  double dtm22 = image_->getReal("DTM2_2", 1);

  double dtv1 = image_->getReal("DTV1", 0);
  double dtv2 = image_->getReal("DTV2", 0);

  physicalToDetector = Matrix(dtm11, dtm12, dtm21, dtm22, dtv1, dtv2);
  detectorToPhysical = physicalToDetector.invert();

  // Physical to Amplifier (ATM/ATV keywords)

  if (image_->find("ATM1_1") ||
      image_->find("ATM1_2") ||
      image_->find("ATM2_1") ||
      image_->find("ATM2_2") ||
      image_->find("ATV1") ||
      image_->find("ATV2"))
    keyATMV = 1;

  double atm11 = image_->getReal("ATM1_1", 1);
  double atm12 = image_->getReal("ATM1_2", 0);
  double atm21 = image_->getReal("ATM2_1", 0);
  double atm22 = image_->getReal("ATM2_2", 1);

  double atv1 = image_->getReal("ATV1", 0);
  double atv2 = image_->getReal("ATV2", 0);

  physicalToAmplifier = Matrix(atm11, atm12, atm21, atm22, atv1, atv2);
  amplifierToPhysical = physicalToAmplifier.invert();

  if (DebugMosaic) {
    cerr << endl;
    cerr << "ATM/V: " << physicalToAmplifier << endl;
    cerr << "ATM/V-1: " << amplifierToPhysical << endl;
    cerr << "DTM/V: " << physicalToDetector << endl;
    cerr << "DTM/V-1: " << detectorToPhysical << endl;
    cerr << "LTM/V: " << physicalToImage << endl;
    cerr << "LTM/V-1: " << imageToPhysical << endl;
  }

  // DATASEC

  iparams.xmin = 0;
  iparams.xmax = width_;
  iparams.ymin = 0;
  iparams.ymax = height_;

  {
    char* datstr = image_->getString("DATASEC");
    // default
    Vector v1(1,1);
    Vector v2(width_,height_);

    if (datstr && *datstr && parseSection(datstr,&v1,&v2)) {
      // additional check
      if (v1[0]<1 || v1[1]<1 || v1[1]>width_ || v2[1]>height_) {
	// default
	v1 = Vector(1,1);
	v2 = Vector(width_,height_);
	keyDATASEC = 0;
      } else
	keyDATASEC = 1;
    }
    else
      keyDATASEC = 0;

    // DATASEC is in IMAGE, dparams go from 0 to width
    datasec = BBox(v1,v2);
    v1 -= Vector(1,1);
    dparams.xmin = v1[0];
    dparams.xmax = v2[0];
    dparams.ymin = v1[1];
    dparams.ymax = v2[1];

    if (datstr)
      delete [] datstr;
  }

  if (DebugMosaic) {
    cerr << "iparams " 
	 << iparams.xmin << ' ' << iparams.xmax << ' ' 
	 << iparams.ymin << ' ' << iparams.ymax << endl;

    cerr << "dparams " 
	 << dparams.xmin << ' ' << dparams.xmax << ' ' 
	 << dparams.ymin << ' ' << dparams.ymax << endl;
  }

  switch (*(parent->currentMosaic)) {
  case Base::NOMOSAIC:
    return 1;
  case Base::IRAF:
    return processKeywordsIRAF();
  case Base::WCSMOSAIC:
  case Base::WFPC2:
    return processKeywordsWCS();
  }
}

int FitsImage::processKeywordsIRAF() 
{
  // DETSIZE
  {
    char* sizestr = image_->getString("DETSIZE");
    Vector v1,v2;
    if (!(sizestr && *sizestr && parseSection(sizestr,&v1,&v2))) {
      if (sizestr)
	delete [] sizestr;
      return 0;
    }
    delete [] sizestr;
    detsize = BBox(v1,v2);
  }

  // DETSEC
  Orientation orientation = NORMAL;
  Orientation mo = NORMAL;
  {
    char* detstr =  image_->getString("DETSEC");
    Vector v1,v2;
    if (!(detstr && *detstr && parseSection(detstr,&v1,&v2))) {
      if (detstr)
	delete [] detstr;
      return 0;
    }
    delete [] detstr;
    detsec = BBox(v1,v2);

    int xx = (v1[0] < v2[0]);
    int yy = (v1[1] < v2[1]);

    if (xx && yy)
      orientation = NORMAL;
    else if (!xx && yy)
      orientation = XX;
    else if (!xx && !yy)
      orientation = XY;
    else if (xx && !yy)
      orientation = YY;
  }

  // CCDSUM
  ccdsum = Vector(1,1);
  {
    char* ccdstr = image_->getString("CCDSUM");
    if (ccdstr && *ccdstr) {
      double Ns, Np, Ns1, Np1;
      string x(ccdstr);
      istringstream str(x);

      str >> Ns >> Np >> Ns1 >> Np1;
      ccdsum = Vector(1/Ns, 1/Np);
    }

    if (ccdstr)
      delete [] ccdstr;
  }

  // origin
  Vector origin = detsec.ll * Scale(ccdsum) * Translate(-datasec.ll);

  // matrix
  // if the segment is flipped, we can have a discontinuity at
  // the edges, due to round off errors, so we 'nudge' it

  Matrix flip;
  switch (orientation) {
  case NORMAL:
    break;
  case XX:
    flip = FlipX();
    origin -= Vector(FLT_EPSILON,0);
    break;
  case YY:
    flip = FlipY();
    origin -= Vector(0,FLT_EPSILON);
    break;
  case XY:
    flip = FlipXY();
    origin -= Vector(FLT_EPSILON,FLT_EPSILON);
    break;
  }

  // internal flip
  Matrix mflip;
  switch (parent->IRAFOrientation(orientation)) {
  case NORMAL:
    break;
  case XX:
    mflip = FlipX();
    break;
  case YY:
    mflip = FlipY();
    break;
  case XY:
    mflip = FlipXY();
    break;
  }

  Vector center = datasec.center() * imageToData;
  Vector mcenter = detsize.center() * imageToData * Scale(ccdsum);

  wcsToRef =
    Translate(-center) *
    flip *
    Translate(center) *
    Translate(origin) * 
    Translate(-mcenter) *
    mflip *
    Translate(mcenter);

  if (DebugMosaic) {
    cerr << "ProcessKeywordsIRAF" << endl
	 << " datasec: " << datasec << endl
	 << " ccdsum : " << ccdsum << endl
	 << " detsize: " << detsize << endl
	 << " detsec : " << detsec << endl
	 << " matrix : " << wcsToRef << endl;
  }

  return 1;
}

int FitsImage::processKeywordsWCS()
{
  wcsToRef = parent->calcAlignWCS(*(parent->channelFits_), this,
				  *(parent->currentMosaicSystem), FK5);

  if (DebugMosaic)
    cerr << "ProcessKeywordsWCS " << endl
	 << " matrix : " << wcsToRef << endl;
 
  return 1;
}

void FitsImage::replaceWCS(FitsHead* h)
{
  // Process OBJECT keyword
  if (objectName)
    delete [] objectName;
  objectName = h->getString("OBJECT");

  // Process WCS keywords
  initWCS(h);
}

void FitsImage::reset()
{
  if (fits_)
    delete fits_;
  fits_ = NULL;

  if (compress_)
    delete compress_;
  compress_ = NULL;
  if (hist_)
    delete hist_;
  hist_ = NULL;

  if (basedata_)
    delete basedata_;
  basedata_ = NULL;

  if (smooth_)
    delete smooth_;
  smooth_ = NULL;
  if (smoothdata_)
    delete smoothdata_;
  smoothdata_ = NULL;

  base_ = NULL;
  image_ = NULL;
  data_ = NULL;
}

void FitsImage::resetWCS()
{
  // Process OBJECT keyword
  if (objectName)
    delete [] objectName;
  objectName = image_->getString("OBJECT");

  // Process WCS keywords
  initWCS(image_->head());
}

void FitsImage::resetWCS0()
{
  if (wcs[WCS0-WCS])
    wcsfree(wcs[WCS0-WCS]);
  wcs[WCS0-WCS] = NULL;
}

char* FitsImage::root(const char* fn)
{
  if (fn) {
    const char* ptr = fn;           // init the ptr
    while(*ptr++);                  // walk it forward to end of string
    ptr--;                          // backup one
    while(*ptr != '/' && ptr != fn) // walk it backward til last / or beginning
      ptr--;
    if (*ptr == '/')                // step it over the last '/'
      ptr++;
    return dupstr(ptr);             // got it!
  }
  else
    return NULL;
}

// used by incr load
void FitsImage::setDataParams(int x0, int y0, int x1, int y1)
{
  irparams.xmin = drparams.xmin = x0;
  irparams.xmax = drparams.xmax = x1;
  irparams.ymin = drparams.ymin = y0;
  irparams.ymax = drparams.ymax = y1;

  // now clip by DATASEC
  if (dparams.xmin > drparams.xmin)
    drparams.xmin = dparams.xmin;
  if (dparams.xmax < drparams.xmax)
    drparams.xmax = dparams.xmax;
  if (dparams.ymin < drparams.ymin)
    drparams.ymin = dparams.ymin;
  if (dparams.ymax < drparams.ymax)
    drparams.ymax = dparams.ymax;
}

void FitsImage::setFileName(const char* fn)
{
  if (fullBaseFileName)
    delete [] fullBaseFileName;
  fullBaseFileName = NULL;

  if (rootBaseFileName)
    delete [] rootBaseFileName;
  rootBaseFileName = NULL;

  if (iisFileName)
    delete [] iisFileName;
  iisFileName = NULL;

  // no filename to set
  if (!fn)
    return;

  // strip any '[]'
  char* ffn = strip(fn);

  FitsFile* ptr = compress_ ? compress_ : fits_;
  if (!ptr)
    return;

  if(ptr->find("EXTNAME")) {
    char* c = ptr->getString("EXTNAME");
    {
      ostringstream str;
      str << ffn << '[' << c << ']' << ends;
      fullBaseFileName = dupstr(str.str().c_str());
    }
    {
      char* m = root(ffn);
      ostringstream str;
      str << m << '[' << c << ']' << ends;
      rootBaseFileName = dupstr(str.str().c_str());
      delete [] m;
    }
    delete [] c;
  }
  else if (ptr->ext()) {
    {
      ostringstream str;
      str << ffn << '[' << ptr->ext() << ']' << ends;
      fullBaseFileName = dupstr(str.str().c_str());
    }
    {
      char* m = root(ffn);
      ostringstream str;
      str << m << '[' << ptr->ext() << ']' << ends;
      rootBaseFileName = dupstr(str.str().c_str());
      delete [] m;
    }
  }
  else {
    fullBaseFileName = dupstr(ffn);
    rootBaseFileName = root(ffn);
  }
  
  // by default, iisFileName is fullBaseFileName
  if (fullBaseFileName)
    iisFileName = dupstr(fullBaseFileName);

  delete [] ffn;
  updateFileName();
}

// used by incr load
void FitsImage::setMinMaxParams(int x0, int y0, int x1, int y1)
{
  imparams.xmin = dmparams.xmin = x0;
  imparams.xmax = dmparams.xmax = x1;
  imparams.ymin = dmparams.ymin = y0;
  imparams.ymax = dmparams.ymax = y1;

  // now clip by DATASEC
  if (dparams.xmin > dmparams.xmin)
    dmparams.xmin = dparams.xmin;
  if (dparams.xmax < dmparams.xmax)
    dmparams.xmax = dparams.xmax;
  if (dparams.ymin < dmparams.ymin)
    dmparams.ymin = dparams.ymin;
  if (dparams.ymax < dmparams.ymax)
    dmparams.ymax = dparams.ymax;

  if (data_)
    data_->resetMinMax();
}

void FitsImage::setObjectName(const char* obj)
{
  if (objectName)
    delete [] objectName;
  objectName = dupstr(obj);
}

char* FitsImage::strip(const char* fn)
{
  if (fn) {
    char* r = dupstr(fn);           // dup the string
    char* ptr = r;                  // init the ptr
    while(*ptr != '[' && *ptr)      // walk it forward til '[' or end
      ptr++;
    *ptr = '\0';                    // zero out rest

    return r;                       // got it!
  }
  else
    return NULL;
}

void FitsImage::updateFileName()
{
  if (fullFileName)
    delete [] fullFileName;
  fullFileName = NULL;

  if (rootFileName)
    delete [] rootFileName;
  rootFileName = NULL;

  char* filter = (char*)fits_->pFilter();
  char* slice = (char*)fits_->pSliceFilter();
  int doFilter = (filter && *filter);
  int doSlice = (slice && *slice);

  if (fullBaseFileName) {
    ostringstream str;
    if (doFilter && doSlice)
      str << fullBaseFileName << '[' << filter << ',' << slice << ']' <<ends;
    else if (doFilter && !doSlice)
      str << fullBaseFileName << '[' << filter << ']' << ends;
    else if (!doFilter && doSlice)
      str << fullBaseFileName << '[' << slice << ']' << ends;
    else
      str << fullBaseFileName << ends;

    fullFileName = dupstr(str.str().c_str());
  }

  if (rootBaseFileName) {
    ostringstream str;
    if (doFilter && doSlice)
      str << rootBaseFileName << '[' << filter << ',' << slice << ']' <<ends;
    else if (doFilter && !doSlice)
      str << rootBaseFileName << '[' << filter << ']' << ends;
    else if (!doFilter && doSlice)
      str << rootBaseFileName << '[' << slice << ']' << ends;
    else
      str << rootBaseFileName << ends;

    rootFileName = dupstr(str.str().c_str());
  }
}

void FitsImage::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  dataToMagnifier = dataToRef * refToMagnifier;
  magnifierToData = dataToMagnifier.invert();
}

void FitsImage::updateMagnifierMatrices(Matrix3d& refToMagnifier3d)
{
  dataToMagnifier3d = dataToRef3d * refToMagnifier3d;
  magnifierToData3d = dataToMagnifier3d.invert();
}

void FitsImage::updateMatrices(Matrix& rgbToRef,
			       Matrix& refToUser,
			       Matrix& userToWidget,
			       Matrix& widgetToCanvas,
			       Matrix& userToPanner)
{
  dataToRef = wcsToRef * rgbToRef;
  refToData = dataToRef.invert();

  // derived matrices

  refToCanvas = refToUser * userToWidget * widgetToCanvas;
  canvasToRef = refToCanvas.invert();

  physicalToData = physicalToImage * imageToData;
  dataToPhysical = physicalToData.invert();

  physicalToRef = physicalToImage * imageToData * dataToRef;
  refToPhysical = physicalToRef.invert();

  physicalToUser = physicalToImage * imageToData * dataToRef * refToUser;
  userToPhysical = physicalToUser.invert();

  physicalToCanvas = physicalToImage * imageToData * dataToRef * refToUser * 
    userToWidget * widgetToCanvas;
  canvasToPhysical = physicalToCanvas.invert();

  physicalToPanner = physicalToImage * imageToData * dataToRef * refToUser * 
    userToPanner;
  pannerToPhysical = physicalToPanner.invert();

  imageToRef = imageToData * dataToRef;
  refToImage = imageToRef.invert();

  imageToUser = imageToData * dataToRef * refToUser;
  userToImage = imageToUser.invert();

  imageToWidget = imageToData * dataToRef *  refToUser * userToWidget;
  widgetToImage = imageToWidget.invert();

  imageToCanvas = imageToData * dataToRef *  refToUser *
    userToWidget *  widgetToCanvas;
  canvasToImage = imageToCanvas.invert();

  imageToPanner = imageToData * dataToRef * refToUser * userToPanner;
  pannerToImage = imageToPanner.invert();

  dataToUser = dataToRef * refToUser;
  userToData = dataToUser.invert();

  dataToWidget = dataToRef *  refToUser *  userToWidget;
  widgetToData = dataToWidget.invert();

  dataToCanvas = dataToRef *  refToUser *  userToWidget * widgetToCanvas;
  canvasToData = dataToCanvas.invert();

  dataToPanner = dataToRef * refToUser * userToPanner;
  pannerToData = dataToPanner.invert();

  amplifierToRef = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef;
  refToAmplifier = amplifierToRef.invert();

  amplifierToUser = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser;
  userToAmplifier = amplifierToUser.invert();

  amplifierToCanvas = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser *
    userToWidget * 
    widgetToCanvas;
  canvasToAmplifier = amplifierToCanvas.invert();

  amplifierToPanner = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser * 
    userToPanner;
  pannerToAmplifier = amplifierToPanner.invert();

  amplifierToImage =
    amplifierToPhysical *
    physicalToImage;
  imageToAmplifier = amplifierToImage.invert();

  detectorToRef = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef;
  refToDetector = detectorToRef.invert();

  detectorToUser = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser;
  userToDetector = detectorToUser.invert();

  detectorToCanvas = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser *
    userToWidget * 
    widgetToCanvas;
  canvasToDetector = detectorToCanvas.invert();

  detectorToPanner = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser * 
    userToPanner;
  pannerToDetector = detectorToPanner.invert();

  detectorToImage =
    detectorToPhysical *
    physicalToImage;
  imageToDetector = detectorToImage.invert();
}

void FitsImage::updateMatrices(Matrix3d& threedToRef3d,
			       Matrix3d& refToUser3d,
			       Matrix3d& userToWidget3d,
			       Matrix3d& userToPanner3d)
{
  dataToRef3d = threedToRef3d;
  refToData3d = dataToRef3d.invert();

  // derived matrices

  dataToPanner3d = dataToRef3d * refToUser3d * userToPanner3d;
  pannerToData3d = dataToPanner3d.invert();

  dataToWidget3d = dataToRef3d *  refToUser3d *  userToWidget3d;
  widgetToData3d = dataToWidget3d.invert();
}

void FitsImage::updatePS(Matrix ps)
{
  dataToPS = dataToRef * ps;
  PSToData = dataToPS.invert();
}

void FitsImage::updatePS(Matrix3d ps)
{
  dataToPS3d = dataToRef3d * ps;
  PSToData3d = dataToPS3d.invert();
}

Vector FitsImage::wcs2pix(Vector in, CoordSystem sys)
{
  if (hasWCS(sys)) {
    int i = sys-WCS;
    wcsininit(wcs[i], ::getradecsys(wcs[i]));

    double x,y;
    int off;
    ::wcs2pix(wcs[i], in[0], in[1], &x, &y, &off);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2pix(Vector in, CoordSystem sys, SkyFrame sky)
{
  if (hasWCS(sys)) {
    int i = sys-WCS;

    if (hasWCSEqu(sys)) {
      if (sky == NATIVEWCS)
	wcsininit(wcs[i], ::getradecsys(wcs[i]));
      else
	wcsininit(wcs[i], skyFrameStr_[sky-FK4]);
    }
    else
      wcsininit(wcs[i], ::getradecsys(wcs[i]));

    double x,y;
    int off;
    ::wcs2pix(wcs[i], in[0], in[1], &x, &y, &off);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2pixDist(Vector in, CoordSystem sys)
{
  if (hasWCS(sys)) {
    Vector cd = getWCScdelt(sys);
    return Vector(in[0]/cd[0], in[1]/cd[1]).abs();
  }
  else
    return in;
}

double FitsImage::wcsdist(Vector a, Vector b, CoordSystem sys)
{
  if (hasWCS(sys))
    return ::wcsdist(a[0],a[1],b[0],b[1]);
  else
    return 0;
}

