// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "framebase.h"
#include "fitsimage.h"
#include "contour.h"
#include "ps.h"

// Public

FrameBase::FrameBase(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : Base(i, c, item)
{
  rotateSrcXM = NULL;
  rotateDestXM = NULL;
  rotatePM = 0;
  rotateGCXOR = NULL;

  // this should all be set in the derived classes
  markers = &userMarkers;
  undoMarkers = &undoUserMarkers;
  pasteMarkers = &pasteUserMarkers;
  undoMarkerType = NONE;

  editMarker = NULL;
  rotateMarker = NULL;
  markerDefaultSystem = IMAGE;
  markerDefaultSky = FK5;

  compositeMarker = NULL;
}

FrameBase::~FrameBase()
{
  if (panPM)
    Tk_FreePixmap(display, panPM);

  if (panGCXOR)
    XFreeGC(display, panGCXOR);

  if (rotateGCXOR)
    XFreeGC(display, rotateGCXOR);
}

double FrameBase::calcZoom(Vector src, Vector dest)
{
  // we need to calculate the width and height of the rotated image
  // so we can derived a zoom factor to shrink it to fit the requested size

  Vector ll(0,0);
  Vector ul(0,dest[1]);
  Matrix m = Translate(-dest/2) * Rotate(wcsRotation) * Rotate(rotation);

  Vector cc = ll * m;
  Vector dd = ul * m;

  float z1 = fabs(cc[0]) > fabs(cc[1]) ? fabs(cc[0]) : fabs(cc[1]);
  float z2 = fabs(dd[0]) > fabs(dd[1]) ? fabs(dd[0]) : fabs(dd[1]);

  return ((src[0]<src[1] ? src[0]:src[1]) / ((z1>z2 ? z1:z2)*2));
}

double FrameBase::calcPannerZoom()
{
  if (keyFits_ && *keyFits_ && pannerPixmap)
    return calcZoom(Vector(pannerWidth,pannerHeight),imageSize());
  else
    return 1;
}

void FrameBase::centerImage()
{
  // Note: cursor is in USER coords, imageCenter() in IMAGE coords
  //   always center to center of pixel
  //   we have a rounding issue, so bump it before the floor()

  //  cursor = imageCenter().floor() + Vector(.5,.5);
  Vector center = imageCenter() * imageToData * 
    Translate(FLT_EPSILON,FLT_EPSILON);
  cursor = center.floor() * Translate(.5,.5);
}

void FrameBase::getColorScaleLevel(int count, const Vector& v,
				   Matrix& (FitsImage::*matrixToData)())
{
  if (FitsImage* ptr = isInFits(v, matrixToData, NULL)) {
    Base::getColorScaleLevelCmd(count, ptr->getLowDouble(),
				ptr->getHighDouble(),
				currentScale->colorScaleType());		
  }
  else
    Base::getColorScaleLevelCmd(count, currentScale->low(), 
				currentScale->high(),
				currentScale->colorScaleType());		
}

void FrameBase::getFitsExt(const Vector& v, InternalSystem ref,
			   Matrix& (FitsImage::*matrixToData)())
{
  if (*channelFits_) {
    FitsImage* ptr = *channelFits_;
    while (ptr) {
      Vector img = v * (ptr->*matrixToData)();
      FitsBound* params = ptr->getDataParams(currentScale->scanMode());

      if (img[0]>=params->xmin && img[0]<params->xmax && 
	  img[1]>=params->ymin && img[1]<params->ymax) {
	// advance to correct slice
	for (int ss=1; ss<slice; ss++)
	  if (ptr)
	    ptr = ptr->nextSlice();
	if (!ptr)
	  break;

	ostringstream str;
	str << ptr->ext() << ends;
	Tcl_AppendResult(interp, str.str().c_str(), NULL);
	return;
      }
      ptr = ptr->nextMosaic();
    }
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getPixelTable(const Vector& v, 
			  Matrix& (FitsImage::*getMatrix)(),
			  int size, char* var)
{
  Vector r;
  if (FitsImage* ptr = isInFits(v, getMatrix, &r)) {
    getPixelTable(ptr, r, size, var);
    return;
  }

  // else return blank
  for (int i=0; i<=size; i++)
    for (int j=0; j<=size; j++) {
      ostringstream str;
      str << i << j << ends;
      Tcl_SetVar2(interp,var,str.str().c_str(),"",0);
    }
}

void FrameBase::getPixelTable(FitsImage* ptr, const Vector& img, 
			  int size, char* var)
{
  Vector half((size-1)/2,(size-1)/2);
  Vector ur = img+half;
  Vector ll = img-half;
  int i,j;

  FitsBound* params = ptr->getDataParams(currentScale->scanMode());

  // x (columns)
  for (i=0,j=0; i<size; i++) {
    ostringstream str;

    str << i+1 << j << ends;

    if (ur[0]>=params->xmin && ll[0]<params->xmax && 
	ur[1]>=params->ymin && ll[1]<params->ymax) {
      Vector pt = ((ll+Vector(i,j)) * dataToImage).round();
      if (pt[0]>params->xmin && pt[0]<=params->xmax) {
	ostringstream lstr;
	lstr << pt[0] << ends;
	Tcl_SetVar2(interp,var,str.str().c_str(),lstr.str().c_str(),0);
      }
      else
	Tcl_SetVar2(interp,var,str.str().c_str(),"",0);
    }
    else
      Tcl_SetVar2(interp,var,str.str().c_str(),"",0);
  }

  // y (rows)
  for (i=0,j=0; j<size; j++) {
    ostringstream str;
    str << i << j+1 << ends;

    if (ur[0]>=params->xmin && ll[0]<params->xmax && ur[1]>=params->ymin && ll[1]<params->ymax) {
      Vector pt = ((ll+Vector(i,j)) * dataToImage).round();
      if (pt[1]>params->ymin && pt[1]<=params->ymax) {
	ostringstream lstr;
	lstr << pt[1] << ends;
	Tcl_SetVar2(interp,var,str.str().c_str(),lstr.str().c_str(),0);
      }
      else
	Tcl_SetVar2(interp,var,str.str().c_str(),"",0);
    }
    else
      Tcl_SetVar2(interp,var,str.str().c_str(),"",0);
  }

  // body
  for (i=0; i<size; i++)
    for (j=0; j<size; j++) {
      Vector pt = ll+Vector(i,j);
      ostringstream str;      
      str << i+1 << j+1 << ends;

      if (pt[0]>=params->xmin && pt[0]<params->xmax && pt[1]>=params->ymin && pt[1]<params->ymax)
	Tcl_SetVar2(interp,var,str.str().c_str(),(char*)ptr->getValue(pt),0);
      else
	Tcl_SetVar2(interp,var,str.str().c_str(),"",0);
    }
}

FitsImage* FrameBase::isInFits(const Vector& vv, 
			       Matrix& (FitsImage::*getMatrix)(), Vector* rv)
{
  FitsImage* ptr = currentFits_;
  while (ptr) {
    Vector rr = vv * (ptr->*getMatrix)();
    FitsBound* params = ptr->getDataParams(currentScale->scanMode());

    if (rr[0]>=params->xmin && rr[0]<params->xmax && 
	rr[1]>=params->ymin && rr[1]<params->ymax) {
      if (rv)
	*rv = rr;
      return ptr;
    }
    ptr = ptr->nextMosaic();
  }

  return NULL;
}

void FrameBase::pannerPush()
{
  if (usePanner) {
    ostringstream str;

    str << pannerName << " update " << pannerPixmap << ";";

    // calculate bbox
    Vector ll = Vector(0,0) * widgetToPanner;
    Vector lr = Vector(options->width,0) * widgetToPanner;
    Vector ur = Vector(options->width,options->height) * widgetToPanner;
    Vector ul = Vector(0,options->height) * widgetToPanner;

    str << pannerName << " update bbox " << ll << lr << ur << ul << ";";

    // calculate image compass vectors
    Matrix mm = 
      FlipY() *
      wcsOrientationMatrix * 
      Rotate(wcsRotation) *
      orientationMatrix *
      Rotate(rotation);

    Vector xx = (Vector(1,0)*mm).normalize();
    Vector yy = (Vector(0,1)*mm).normalize();

    str << pannerName << " update image compass " << xx << yy << ";";

    if (hasWCS(pannerSystem)) {
      Vector orval=currentFits_->pix2wcs(currentFits_->getWCScrpix(pannerSystem),
					pannerSystem, pannerSky);
      // special case, check for poles, use image center
      if (hasWCSEqu(pannerSystem) && (orval[1] == 90 || orval[1] == -90))
	orval=currentFits_->pix2wcs(imageCenter(), pannerSystem, pannerSky);

      Vector orpix = currentFits_->wcs2pix(orval,pannerSystem,pannerSky);
      Vector origin = orpix;
      Vector delta = currentFits_->getWCScdelt(pannerSystem).abs();

      // find normalized north
      Vector north = currentFits_->wcs2pix(Vector(orval[0],orval[1]+delta[1]), 
					  pannerSystem,pannerSky);
      Vector nn = ((north-origin)*mm).normalize();

      // find normalized east
      Vector east = currentFits_->wcs2pix(Vector(orval[0]+delta[0],orval[1]), 
					 pannerSystem,pannerSky);
      Vector ee = ((east-origin)*mm).normalize();

      // and update the panner
      str << pannerName << " update wcs compass " << nn << ee << ends;
    }
    else
      str << pannerName << " update wcs compass invalid" << ends;

    Tcl_Eval(interp, str.str().c_str());
  }
}

int FrameBase::postscriptProc(int prepass)
{
  if (!visible)
    return TCL_OK;

  if (prepass)
    return TCL_OK;

  ps(&FitsImage::getmPSToData);

  switch (psLevel) {
  case 1:
    switch (psColorSpace) {
    case BW:
    case GRAY:
      psContours(GRAY);

      if (grid)
	grid->ps(GRAY);

      if (showMarkers) {
	psMarkers(&userMarkers, GRAY);
	psMarkers(&catalogMarkers, GRAY);
      }

      if (useCrosshair)
	psCrosshair(GRAY);

      break;
    case RGB:
    case CMYK:
      psContours(RGB);

      if (grid)
	grid->ps(RGB);

      if (showMarkers) {
	psMarkers(&userMarkers, RGB);
	psMarkers(&catalogMarkers, RGB);
      }

      if (useCrosshair)
	psCrosshair(RGB);

      break;
    }
    break;
  case 2:
    psContours(psColorSpace);

    if (grid)
      grid->ps(psColorSpace);

    if (showMarkers) {
      psMarkers(&userMarkers, psColorSpace);
      psMarkers(&catalogMarkers, psColorSpace);
    }

    if (useCrosshair)
      psCrosshair(psColorSpace);

    break;
  }

  return TCL_OK;
}

void FrameBase::printFromRef(FitsImage* ptr, const Vector& v,
			     CoordSystem sys, SkyFrame sky, SkyFormat format,
			     Precision p)
{
  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    printVector(ptr->mapFromRef(v, sys, sky), p);
    return;
  default:
    if (ptr->hasWCS(sys)) {
      if (ptr->hasWCSEqu(sys)) {
	switch (format) {
	case DEGREES:
	    printVector(ptr->mapFromRef(v, sys, sky), p);
	  break;
	case SEXAGESIMAL:
	case HMS:
	  {
	    char buf[64];
	    buf[0] = '\0';
	    ptr->mapFromRef(v, sys, sky, format, buf, 64);

	    // grap only the first two items
	    char* ptr = buf;
	    while (*ptr)
	      ptr++;
	    while (*ptr != ' ' && ptr >= buf)
	      ptr--;
	    *ptr = '\0';

	    Tcl_AppendResult(interp, buf, " ", NULL);
	  }
	  break;
	}
      }
      else
	printVector(ptr->mapFromRef(v, sys), p);
    }
    else
      printVector(Vector(),p);

    return;
  }
}

void FrameBase::psContours(PSColorSpace cs)
{
  // render back to front
  // aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->ps(cs);
    ptr=ptr->previous();
  }

  // main contour
  if (*currentcontour)
    (*currentcontour)->ps(cs);
}

Matrix FrameBase::psMatrix(float scale, int width, int height)
{
  Matrix mm;
  if (!isIIS()) {
    // Note: imageCenter() is in IMAGE coords
    Vector center = imageCenter() * imageToData;
    mm = Translate(-center) * FlipY() * Translate(center);
  }

  return mm *
    Translate(-cursor) *                // translate to cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    Scale(zoom_) *                      // scale about cursor position
    Scale(scale) *
    Rotate(rotation) *                  // rotate about cursor position

    FlipY() *
    Translate(Vector(width,height)/2.);
}

void FrameBase::renderContours(Pixmap pm, const Matrix& mx, 
			       int width, int height)
{
  // render from back to front
  // aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->render(pm, mx, width, height);
    ptr=ptr->previous();
  }

  // main contour
  if (hasContour())
    (*currentcontour)->render(pm, mx, width, height);
}

void FrameBase::renderMagnifierCursor(const Vector& v)
{
    // first, the inner white box
    Vector u = v * canvasToUser;
    Matrix m = Translate(-u) * userToMagnifierCursor;

    Vector c[5];
    c[0] = (u + Vector(-.5,-.5)) * m;
    c[1] = (u + Vector( .5,-.5)) * m;
    c[2] = (u + Vector( .5, .5)) * m;
    c[3] = (u + Vector(-.5, .5)) * m;
    c[4] = c[0];
    
    XPoint pts[5];
    for (int i=0; i<5; i++) {
      Vector z = c[i].round();
      pts[i].x = (short)z[0];
      pts[i].y = (short)z[1];
    }
    XSetForeground(display, magnifierGC, getColor("white"));
    XDrawLines(display, magnifierPixmap, magnifierGC, pts, 5, CoordModeOrigin);

    // now, the outer black box
    Matrix n =
      Translate(-(u * userToMagnifierCursor)) *
      Rotate(-rotation) *
      Rotate(-wcsRotation);
    Matrix o = n.invert();

    Vector d[5];
    {
      for (int i=0; i<5; i++)
	d[i] = c[i] * n;
    }

    d[0]+=Vector(-1,-1);
    d[1]+=Vector( 1,-1);
    d[2]+=Vector( 1, 1);
    d[3]+=Vector(-1, 1);
    d[4]=d[0];

    {
      for (int i=0; i<5; i++) {
	d[i] = d[i] * o;
	Vector z = d[i].round();
	pts[i].x = (int)z[0];
	pts[i].y = (int)z[1];
      }
    }
    XSetForeground(display, magnifierGC, getColor("black"));
    XDrawLines(display, magnifierPixmap, magnifierGC, pts, 5, CoordModeOrigin);
}

void FrameBase::reset()
{
  crosshair = cursor;

  orientation = NORMAL;
  orientationMatrix.identity();
  zoom_ = Vector(1,1);
  rotation = 0;
  
  wcsAlignment = 0;
  wcsSystem = WCS;
  wcsSky = FK5;
  wcsOrientation = NORMAL;
  wcsOrientationMatrix.identity();
  wcsRotation = 0;
  
  centerImage();

  unselectMarkers(&userMarkers);
  unselectMarkers(&catalogMarkers);

  Base::reset();
}

void FrameBase::resetValues()
{
  if (DebugPerf)
    cerr << "resetValues" << endl;

  alignWCS();

  if (!preservePan) {
    centerImage();
    crosshair = cursor;
  }
}

void FrameBase::unloadFits()
{
  if (DebugPerf)
    cerr << "FrameBase::unloadFits" << endl;

  if (!preserveMarkers) {
    // delete markers
    userMarkers.deleteAll();
    undoUserMarkers.deleteAll();
    pasteUserMarkers.deleteAll();
  }

  catalogMarkers.deleteAll();

  Base::unloadFits();
}

void FrameBase::updateBase()
{
  if (DebugPerf)
    cerr << "updateBase..." << endl;

  Base::updateBase();

  int& width = options->width;
  int& height = options->height;

  if (doRender())
    ximageToPixmap(basePixmap, baseXImage, &FitsImage::getmWidgetToData);
  else {
    XSetForeground(display, gc, bgColor->pixel);
    XFillRectangle(display, basePixmap, gc, 0, 0, width, height);
  }
  
  // contours
  renderContours(basePixmap, refToWidget, width, height);

  // grid
  if (grid)
    grid->render();

  // panner
  updatePanner();

  if (DebugPerf)
    cerr << "end" << endl;
}

void FrameBase::updateGCs()
{
  Base::updateGCs();

  // panGCXOR
  if (!panGCXOR)
    panGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, panGCXOR, 0, 0, rectWindow, 1, Unsorted);

  // rotateGCXOR
  if (!rotateGCXOR)
    rotateGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, rotateGCXOR, 0, 0, rectWindow, 1, Unsorted);
}

void FrameBase::updateMagnifierMatrix()
{
  // calculate partial magnifier matrix
  // This is a partial matix, it lacks the Translate at the begining

  userToMagnifier =                     
    wcsOrientationMatrix *              // flip x/y axis about center
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    Scale(zoom_) *                      // scale about cursor position
    Scale(magnifierZoom_) *             // scale
    Rotate(rotation) *                  // rotate about cursor position
    Translate(magnifierWidth/2., magnifierHeight/2.);

  userToMagnifierCursor =                     
    Rotate(wcsRotation) *
    Scale(zoom_) *
    Scale(magnifierZoom_) *
    Rotate(rotation) *
    Translate(magnifierWidth/2., magnifierHeight/2.);
}

void FrameBase::updateMagnifierZoom(const Vector& v)
{
  // v is in CANVAS coords
  Matrix refToMagnifier = refToUser * Translate(-(v * canvasToUser)) * 
    userToMagnifier;
  Matrix magnifierToRef = refToMagnifier.invert();

  updateMagnifierMatrices(refToMagnifier);

  // Update ximage
  buildXImage(magnifierXImage, &FitsImage::getmMagnifierToData);

  // copy XImage to Pixmap
  XPutImage(display, magnifierPixmap, gc, magnifierXImage, 0, 0, 0, 0, 
	    magnifierWidth, magnifierHeight);

  if (useMagnifierGraphics) {
    // render markers
    // markers bounding box is in canvas coords
    // map the magnifier to a bounding box in canvas coords
    Matrix m = magnifierToRef * refToCanvas;
    BBox bb = BBox(v,v);
    bb.bound(Vector(0,0) * m);
    bb.bound(Vector(magnifierWidth,magnifierHeight) * m);

    if (showMarkers) {
      renderMagnifierMarkers(&userMarkers, refToMagnifier, bb);
      renderMagnifierMarkers(&catalogMarkers, refToMagnifier, bb);
    }

    // render crosshair
    if (useCrosshair) {
      Vector r = crosshair * refToMagnifier;
      if (r[0]>=0 && r[0]<magnifierWidth)
	XDrawLine(display, magnifierPixmap, crosshairGC, 
		  r[0], 1, r[0], magnifierHeight);

      if (r[1]>=0 && r[1]<magnifierHeight)
	// bump Y by one, I don't know why, it just works
	XDrawLine(display, magnifierPixmap, crosshairGC,
		  1, r[1], magnifierWidth, r[1]);

    }

    // render contours
    renderContours(magnifierPixmap, refToMagnifier, 
		   magnifierWidth, magnifierHeight);
  }
}

void FrameBase::updateMatrices()
{
  if (DebugPerf)
    cerr << "updateMatrices...";

  // These are the basic tranformation matrices
  // Note: imageCenter() is in IMAGE coords
  Vector center = imageCenter() * imageToData;

  // refToUser
  if (!isIIS()) {
    refToUser = 
      Translate(-center) *                // translate to center image
      FlipY() *                           // flip y axis for X Windows
      Translate(center);                  // translate back from center
    userToRef = refToUser.invert();
  }
  else {
    refToUser.identity();
    userToRef.identity();
  }

  // userToWidget
  userToWidget =
    Translate(-cursor) *                // translate to cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    Scale(zoom_) *                      // scale about cursor position
    Rotate(rotation) *                  // rotate about cursor position
    //    Translate(cursor) *
    //    Translate(-cursor) *
    // translate to center of widget. needs to be integer
    Translate((int)(options->width/2.), (int)(options->height/2.));
  widgetToUser = userToWidget.invert();

  Base::updateMatrices();

  // Panner Matrix
  userToPanner =
    Translate(-center) *
    wcsOrientationMatrix *              // flip x/y axis about center
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    Scale(calcPannerZoom()) *
    Rotate(rotation) *                  // rotate about cursor position
    Translate(pannerWidth/2., pannerHeight/2.);
  pannerToUser = userToPanner.invert();

  refToPanner = refToUser * userToPanner;
  pannerToRef = refToPanner.invert();

  pannerToWidget = pannerToUser * userToWidget;
  widgetToPanner = pannerToWidget.invert();

  // Magnifier Matrix
  updateMagnifierMatrix();

  // Markers
  updateMarkers(&userMarkers);
  updateMarkers(&catalogMarkers);

  if (DebugPerf)
    cerr << "end" << endl;
}

void FrameBase::updatePanner()
{
  if (usePanner)
    if (doRender())
      ximageToPixmap(pannerPixmap, pannerXImage, &FitsImage::getmPannerToData);
    else {
      XSetForeground(display, gc, bgColor->pixel);
      XFillRectangle(display, pannerPixmap, gc, 0, 0, pannerWidth,pannerHeight);
    }  

  pannerPush();
}

void FrameBase::updatePM(const BBox& bbox)
{
  // bbox is in Canvas Coords

  if (DebugPerf)
    cerr << "updatePM...";

  Base::updatePM(bbox);

  // markers
  if (showMarkers) {
    renderMarkers(&userMarkers, bbox);
    renderMarkers(&catalogMarkers, bbox);
  }

  // crosshair
  renderCrosshair();

  // hightlite bbox
  renderHighLite();

  if (DebugPerf)
    cerr << "end" << endl;
}


#ifdef _MACOSX
void FrameBase::macosxPrintCmd()
{
  if (!visible)
    return;

  // init
  macosxBegin();

  // image
  macosx(&FitsImage::getmPSToData);

  // contours
  //   render back to front
  //   aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->macosx();
    ptr=ptr->previous();
  }

  // main contour
  if (*currentcontour)
    (*currentcontour)->macosx();

  // grid
  if (grid)
    grid->macosx();

  // markers
  if (showMarkers) {
    macosxMarkers(&userMarkers);
    macosxMarkers(&catalogMarkers);
  }

  if (useCrosshair)
    macosxCrosshair();

  // cleanup
  macosxEnd();
}
#endif

#ifdef _WIN32
void FrameBase::win32PrintCmd()
{
  if (!visible)
    return;

  // init
  win32Begin();

  // image
  win32(&FitsImage::getmPSToData);

  // contours
  //   render back to front
  //   aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->win32();
    ptr=ptr->previous();
  }

  // main contour
  if (*currentcontour)
    (*currentcontour)->win32();

  // grid
  if (grid)
    grid->win32();

  // markers
  if (showMarkers) {
    win32Markers(&userMarkers);
    win32Markers(&catalogMarkers);
  }

  if (useCrosshair)
    win32Crosshair();

  // cleanup
  win32End();
}
#endif
