// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#ifndef __framebase_h__
#define __framebase_h__

#include "base.h"
#include "marker.h"

class Attribute;
class BaseBox;
class BaseEllipse;
class BaseMarker;
class Composite;
class Contour;
class FVContour;
class Grid;
class Marker;

class Annulus;
class Box;
class BoxAnnulus;
class Bpanda;
class Circle;
class Compass;
class Cpanda;
class Ellipse;
class EllipseAnnulus;
class Epanda;
class Line;
class Point;
class CirclePoint;
class BoxPoint;
class DiamondPoint;
class CrossPoint;
class ExPoint;
class ArrowPoint;
class BoxCirclePoint;
class Projection;
class Polygon;
class Ruler;
class Text;
class Vect;

class Filter;

// Frame

class FrameBase : public Base {
  friend class Attribute;
  friend class BaseBox;
  friend class BaseEllipse;
  friend class BaseMarker;
  friend class Composite;
  friend class Contour;
  friend class FitsImage;
  friend class FVContour;
  friend class Grid;
  friend class Marker;

  friend class Annulus;
  friend class Box;
  friend class BoxAnnulus;
  friend class Bpanda;
  friend class Circle;
  friend class Compass;
  friend class Cpanda;
  friend class Ellipse;
  friend class EllipseAnnulus;
  friend class Epanda;
  friend class Line;
  friend class Point;
  friend class CirclePoint;
  friend class BoxPoint;
  friend class DiamondPoint;
  friend class CrossPoint;
  friend class ExPoint;
  friend class ArrowPoint;
  friend class BoxCirclePoint;
  friend class Projection;
  friend class Polygon;
  friend class Ruler;
  friend class Text;
  friend class Vect;

protected:
  Vector cursor; // current cursor position in USER coords
  Vector crosshair; // crosshair location in REF coords

  XImage* rotateSrcXM;       // rotate src ximage
  XImage* rotateDestXM;      // rotate dest ximage
  Pixmap rotatePM;           // rotate pixmap
  GC rotateGCXOR;            // GC for interactive rotation

  List<Marker>* markers;     // pointer to current marker stack
  List<Marker>* undoMarkers; // pointer to current marker stack
  List<Marker>* pasteMarkers;// pointer to current marker stack

  UndoMarkerType undoMarkerType; // type

  List<Marker> userMarkers;      // foreground marker stack
  List<Marker> undoUserMarkers;  // remember marker to undo
  List<Marker> pasteUserMarkers; // cut/copy/paste

  List<Marker> catalogMarkers;      // background marker stack
  List<Marker> undoCatalogMarkers;  // remember marker to undo
  List<Marker> pasteCatalogMarkers; // cut/copy/paste

  Vector markerBegin;        // marker move begin in image coords
  Marker* editMarker;        // pointer to marker currently being edited
  Marker* rotateMarker;      // pointer to marker currently being rotated
  Vector regionBegin;        // select region begin in canvas coords
  Vector regionEnd;          // select region end in canvas coords
  CoordSystem markerDefaultSystem; // default marker sys (RAWXY)
  SkyFrame markerDefaultSky; // default marker sky (RAWXY)

  Composite* compositeMarker; // pointer to current composite marker

  Vector iisLastCursor;      // iis cursor state info
  Vector iisSaveCursor;
  Window iisSaveRoot;
  int iisInteractive;
				  
  Matrix userToMagnifierCursor;

private:
  void bltCut(char*, char*, Orientation, const Vector&);
  void bltCutFits(double*, double*, int, Vector&, Orientation);
  void bltProjection(char*, char*, char*, char*, 
		     CoordSystem, SkyFrame,
		     const Vector&, const Vector&, double, int);
  void bltProjFits(double*, double*, double*, double*, int,
		   CoordSystem, SkyFrame,
		   const Vector&, const Vector&, double, int);

  double calcPannerZoom();
  void createMarker(Marker*);

  void drawCrosshair();

  void eraseCrosshair();

  FitsImage* isInFits(const Vector&, Matrix& (FitsImage::*)(), Vector*);

  void markerListHeader(ostream&, CoordSystem, SkyFrame, SkyFormat, int);
  void markerListXMLHeader(ostream&, CoordSystem, SkyFrame, SkyFormat);
  void markerListXMLFooter(ostream&);
  void markerListCiaoHeader(ostream&, CoordSystem, SkyFrame, SkyFormat, int);
  void markerListSAOtngHeader(ostream&,CoordSystem, SkyFrame, SkyFormat, int);
  void markerPrintCoord(const Vector&, InternalSystem);
  void markerPrintCoord(const Vector&, const Vector&,
			CoordSystem, SkyFrame, SkyFormat);

  void markerPrintDouble(double, InternalSystem);
  void markerPrintDouble(const Vector&, double, CoordSystem, SkyFormat);

  void markerPrintDouble(const Vector&, const Vector&, InternalSystem);
  void markerPrintDouble(const Vector&, const Vector&, const Vector&, 
			 CoordSystem, SkyFormat);

  void markerPrintVector(const Vector&,InternalSystem);
  void markerPrintVector(const Vector&,const Vector&,CoordSystem,SkyFormat);
  void markerUndo(Marker*, UndoMarkerType);

  void pannerPush();
  void parseMarker(MarkerFormat,istream&);
  int postscriptProc(int);
  void printFromRef(FitsImage*, const Vector&, CoordSystem, SkyFrame,
		    SkyFormat, Precision);
  void psContours(PSColorSpace);
  void psCrosshair(PSColorSpace);
  void psMarkers(List<Marker>*, int);

  void renderContours(Pixmap, const Matrix&, int, int);
  void renderCrosshair();
  void renderMagnifierCursor(const Vector&);
  void renderMagnifierMarkers(List<Marker>*, const Matrix&, const BBox& bb);
  void renderMarkers(List<Marker>*, const BBox&);

  void unhighliteMarkers();
  void unselectMarkers(List<Marker>*);
  void updateBase();
  void updateGCs();
  void updateMagnifierMatrix();
  void updateMagnifierZoom(const Vector&);
  void updateMarkers(List<Marker>*);
  void updatePM(const BBox&);

  void xmlParse(istream&);
  void xmlParseFIELD(void*, int*, char**, char**, char**, char**, int);
  void xmlParseTR(char**, int*, char**, char**, char**, char**, int);
  void xmlSetProps(unsigned short*, unsigned short, char*);
  int xmlCount(const char*);
  Vector xmlPoint(FitsImage*, const char* x, const char* y, 
		  CoordSystem sys, SkyFrame sky, SkyFormat format, int which=0);
  List<Vertex>* xmlVertex(FitsImage*, const char*, const char*, 
			  CoordSystem, SkyFrame, SkyFormat);
  double xmlAngle(const char*, int, double, AngleFormat, CoordSystem, SkyFrame);
  double* xmlAngles(const char*, int, double, int, AngleFormat, CoordSystem, SkyFrame);
  double* xmlDistance(FitsImage*, const char*, int, CoordSystem, SkyFormat);
  Vector* xmlDistance(FitsImage*, const char*, const char*, int, CoordSystem, SkyFormat);

#ifdef _MACOSX
  void macosxCrosshair();
  void macosxMarkers(List<Marker>*);
#endif

#ifdef _WIN32
  void win32Crosshair();
  void win32Markers(List<Marker>*);
#endif

protected:
  double calcZoom(Vector,Vector);
  void centerImage();
  void createTemplate(const Vector&, istream&);

  void getColorScaleLevel(int, const Vector&, Matrix& (FitsImage::*)());
  void getFitsExt(const Vector&, InternalSystem, Matrix& (FitsImage::*)());
  void getPixelTable(const Vector&, Matrix& (FitsImage::*matrixToData)(),
		     int, char*);
  void getPixelTable(FitsImage*, const Vector&, int, char*);

  virtual BBox imageBBox() =0;
  Vector imageCenter() {return imageBBox().center();}
  Vector imageSize() {return imageBBox().size();}

  Matrix psMatrix(float, int, int);

  void reset();
  void resetValues();
  virtual void rotateMotion() =0;

  void saveFitsResample(OutFitsStream&);
  void saveFitsResampleFits(OutFitsStream&);
  void saveFitsResampleKeyword(OutFitsStream&, FitsHead&);
  virtual void setBinCursor(const Vector&) =0;

  virtual void unloadFits();
  void updateCBMarkers();
  void updateMarkerCBs(List<Marker>*);
  void updateMarkerCoords(List<Marker>*, const Matrix&);
  virtual void updateMatrices();
  virtual void updateMagnifierMatrices(Matrix&) =0;
  void updatePanner();

public:
  FrameBase(Tcl_Interp*, Tk_Canvas, Tk_Item*);
  virtual ~FrameBase();

  Vector centroid(const Vector&);

  FitsImage* findFits(CoordSystem, const Vector&);
  FitsImage* findFits(const Vector&);
  FitsImage* findFits(int);
  int findFits(FitsImage*);

  Matrix getCanvasToRef() {return canvasToRef;}
  CoordSystem getMarkerDefaultSystem() {return markerDefaultSystem;}
  SkyFrame getMarkerDefaultSky() {return markerDefaultSky;}
  void resetCompositeMarker() {compositeMarker = NULL;}

  // Contour Commands
  void contourCopyCmd(CoordSystem, SkyFrame);
  void contourCreateCmd(const char*, int, int, int, float, int);
  void contourCreateCmd(const char*, int, int, float, int, const char*);
  void contourDeleteCmd();
  void contourDeleteAllCmd();
  void contourLevelCmd(const char*);
  void contourLoadCmd(const char*, int, int, const char*, CoordSystem,SkyFrame);
  void contourPasteCmd(const char*, int, int, void*, CoordSystem, SkyFrame);
  void contourSaveCmd(const char*, CoordSystem, SkyFrame);
  void contourSetDashCmd(int);
  void contourSetColorCmd(const char*);
  void contourSetLineWidthCmd(int);
  void getContourCmd(CoordSystem sys, SkyFrame sky);
  void getContourColorNameCmd();
  void getContourDashCmd();
  void getContourLevelCmd();
  void getContourLineWidthCmd();
  void getContourMethodCmd();

  // Coordinate Commands
  void getCoordCmd(const Vector&, InternalSystem, CoordSystem, 
		   SkyFrame, SkyFormat);
  void getCursorCmd(InternalSystem);
  void getCursorCmd(CoordSystem, SkyFrame, SkyFormat, Precision);

  // CrossHair Commands
  void crosshairCmd(int);
  void crosshairCmd(const Vector&, InternalSystem);
  void crosshairCmd(const Vector&, CoordSystem, SkyFrame);
  void crosshairBeginCmd(const Vector&, InternalSystem);
  void crosshairBeginCmd(const Vector&, CoordSystem);
  void crosshairMotionCmd(const Vector&, InternalSystem);
  void crosshairMotionCmd(const Vector&, CoordSystem);
  void crosshairWarpCmd(const Vector&);
  void getCrosshairCmd(InternalSystem);
  void getCrosshairCmd(CoordSystem, SkyFrame, SkyFormat, Precision);
  void getCrosshairStatusCmd();
  void getHorzCutCmd(char*, char*, const Vector&, InternalSystem);
  void getVertCutCmd(char*, char*, const Vector&, InternalSystem);

  // Data Values
  void getDataValuesCmd(int, const Vector&, CoordSystem, SkyFrame, 
			const Vector&, char*);
  void getDataValuesCmd(const Vector&, InternalSystem, const Vector&);

  // Fits Commands
  void getFitsExtCmd(const Vector&, InternalSystem);
  void getFitsFileNameCmd(const Vector&, InternalSystem, FileNameType);
  void getFitsCenterCmd(CoordSystem, SkyFrame, SkyFormat, Precision);
  void getFitsSizeCmd(CoordSystem, SkyFrame, SkyFormat, Precision);
  void saveFitsResampleFileCmd(const char*, int);
  void saveFitsResampleChannelCmd(const char*, int);
  void saveFitsResampleSocketCmd(int, int);
  void updateFitsCmd(int, BBox, int);

  // Grid Commands
  void gridCmd(CoordSystem, SkyFrame, SkyFormat, Grid::GridType, const char*);
  void gridDeleteCmd();
  void getGridCmd();
  void getGridOptionCmd();

  // IIS Commands
  void iisCursorModeCmd(int);
  void iisGetCursorCmd();
  void iisGetFileNameCmd();
  void iisGetFileNameCmd(int);
  void iisGetFileNameCmd(const Vector&);
  void iisMessageCmd(const char*);
  void iisSetCursorCmd(const Vector&, InternalSystem);
  void iisSetCursorCmd(const Vector&, CoordSystem);
  void iisSetFileNameCmd(const char*);
  void iisSetFileNameCmd(const char*,int);
  void iisUpdateCmd() {updateNow(MATRIX);}

  // Info Commands
  void getValueCmd(const Vector&, InternalSystem);

  // Marker Commands
  void createCircleCmd(const Vector&, 
		       double,
		       const char*, int*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);
  void createAnnulusCmd(const Vector&, 
			double, double, int,
			const char*, int*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createAnnulusCmd(const Vector&, 
			int, double*,
			const char*, int*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createCpandaCmd(const Vector&, 
		       double, double, int, 
		       double, double, int,
		       const char*, int*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);
  void createCpandaCmd(const Vector&, 
		       int, double*,
		       int, double*,
		       const char*, int*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);


  void createEllipseCmd(const Vector&, 
			const Vector&, 
			double, 
			const char*, int*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createEllipseAnnulusCmd(const Vector&, 
			       const Vector&, const Vector&, int, 
			       double,
			       const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);
  void createEllipseAnnulusCmd(const Vector&, 
			       int, Vector*,
			       double, 
			       const char*, int*, int,  const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);
  void createEpandaCmd(const Vector&, 
		       double, double, int, 
		       const Vector&, const Vector&, int, 
		       double,
		       const char*, int*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);
  void createEpandaCmd(const Vector&, 
		       int, double*,
		       int, Vector*,
		       double,
		       const char*, int*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);


  void createBoxCmd(const Vector&, 
		    const Vector&, 
		    double, 
		    const char*, int*, int, const char*, 
		    const char*, unsigned short, const char*,
		    const List<Tag>&, const List<CallBack>& cb);
  void createBoxAnnulusCmd(const Vector&, 
			   const Vector&, const Vector&, int, 
			   double,
			   const char*, int*, int, const char*, 
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createBoxAnnulusCmd(const Vector&, 
			   int, Vector*,
			   double, 
			   const char*, int*, int, const char*, 
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createBpandaCmd(const Vector&, 
		       double, double, int, 
		       const Vector&, const Vector&, int, 
		       double,
		       const char*, int*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);
  void createBpandaCmd(const Vector&, 
		       int, double*,
		       int, Vector*,
		       double,
		       const char*, int*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);


  void createLineCmd(const Vector&, 
		     const Vector&, 
		     int, int,
		     const char*, int*, int, const char*,
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);
  void createVectCmd(const Vector&, 
		     const Vector&, 
		     int,
		     const char*, int*, int, const char*,
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);
  void createVectCmd(const Vector&, 
		     double mag, double ang, 
		     int,
		     const char*, int*, int, const char*,
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);
  void createProjectionCmd(const Vector&, 
			   const Vector&,
			   double, 
			   const char*, const char*,
			   const char*, int*, int, const char*,
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createRulerCmd(const Vector&, 
		      const Vector&, 
		      CoordSystem, SkyFrame, CoordSystem, SkyFormat,
		      const char*, int*, int, const char*,
		      const char*, unsigned short, const char*,
		      const List<Tag>&, const List<CallBack>& cb);

  void createCirclePointCmd(const Vector&, int,
			    const char*, int*, int, const char*,
			    const char*, unsigned short, const char*,
			    const List<Tag>&, const List<CallBack>& cb);
  void createBoxPointCmd(const Vector&, int,
			 const char*, int*, int, const char*,
			 const char*, unsigned short, const char*,
			 const List<Tag>&, const List<CallBack>& cb);
  void createDiamondPointCmd(const Vector&, int,
			     const char*, int*, int, const char*, 
			     const char*, unsigned short, const char*,
			     const List<Tag>&, const List<CallBack>& cb);
  void createCrossPointCmd(const Vector&, int,
			   const char*, int*, int, const char*,
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createExPointCmd(const Vector&, int,
			const char*, int*, int, const char*,
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createArrowPointCmd(const Vector&, int,
			   const char*, int*, int, const char*,
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createBoxCirclePointCmd(const Vector&, int,
			       const char*, int*, int, const char*,
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);

  void createTextCmd(const Vector&, 
		     double, int,
		     const char*, int*, int, const char*, 
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);

  void createPolygonCmd(const Vector&, 
			const Vector&,
			const char*, int*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createPolygonCmd(const List<Vertex>&,
			const char*, int*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createContourPolygonCmd(const char*, int*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);

  void createCompassCmd(const Vector&, 
			double, 
			const char*, const char*,
			int, int,
			CoordSystem, SkyFrame,
			const char*, int*, int, const char*,
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);

  void createCompositeCmd(const Vector&, double, int,
			  const char*, int*, int, const char*, 
			  const char*, unsigned short, const char*,
			  const List<Tag>&, const List<CallBack>& cb);
  void createCompositeCmd(
			  const char*, int*, int, const char*, 
			  const char*, unsigned short, const char*,
			  const List<Tag>&, const List<CallBack>& cb);

  void createTemplateCmd(const Vector&, const char*);
  void createTemplateCmd(const Vector&, CoordSystem, SkyFrame, const char*);
  void createTemplateVarCmd(const Vector&, const char* var);

  void getMarkerAngleCmd(int);
  void getMarkerAngleCmd(int, CoordSystem, SkyFrame);
  void getMarkerAnnulusRadiusCmd(int, InternalSystem);
  void getMarkerAnnulusRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerBoxAnnulusRadiusCmd(int, InternalSystem);
  void getMarkerBoxAnnulusRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerBoxRadiusCmd(int, InternalSystem);
  void getMarkerBoxRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerBpandaAnglesCmd(int);
  void getMarkerBpandaAnglesCmd(int, CoordSystem, SkyFrame);
  void getMarkerBpandaRadiusCmd(int, InternalSystem);
  void getMarkerBpandaRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerCenterCmd(int, InternalSystem);
  void getMarkerCenterCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerCircleRadiusCmd(int, InternalSystem);
  void getMarkerCircleRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerColorCmd();
  void getMarkerColorCmd(const char*);
  void getMarkerColorCmd(int);
  void getMarkerCompassArrowCmd(int);
  void getMarkerCompassLabelCmd(int);
  void getMarkerCompassRadiusCmd(int, InternalSystem);
  void getMarkerCompassRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerCompassSystemCmd(int);
  void getMarkerCompositeCmd(int);
  void getMarkerCpandaAnglesCmd(int);
  void getMarkerCpandaAnglesCmd(int, CoordSystem, SkyFrame);
  void getMarkerCpandaRadiusCmd(int, InternalSystem);
  void getMarkerCpandaRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerEllipseRadiusCmd(int, InternalSystem);
  void getMarkerEllipseRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerEllipseAnnulusRadiusCmd(int, InternalSystem);
  void getMarkerEllipseAnnulusRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerEpandaAnglesCmd(int);
  void getMarkerEpandaAnglesCmd(int, CoordSystem, SkyFrame);
  void getMarkerEpandaRadiusCmd(int, InternalSystem);
  void getMarkerEpandaRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerFontCmd();
  void getMarkerFontCmd(const char*);
  void getMarkerFontCmd(int);
  void getMarkerHandleCmd(const Vector&);
  void getMarkerIdCmd(const char*);
  void getMarkerIdCmd(const Vector&);
  void getMarkerIdAllCmd();
  void getMarkerLineCmd(int, InternalSystem);
  void getMarkerLineCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerLineArrowCmd(int);
  void getMarkerLineLengthCmd(int, InternalSystem);
  void getMarkerLineLengthCmd(int, CoordSystem, SkyFormat);
  void getMarkerLineWidthCmd();
  void getMarkerLineWidthCmd(int);
  void getMarkerMapLenFromRefCmd(int, double, CoordSystem, SkyFormat);
  void getMarkerPointSizeCmd(int);
  void getMarkerPolygonSegmentCmd(const Vector&);
  void getMarkerProjectionCmd(int, char*, char*, char*, char*);
  void getMarkerProjectionPointsCmd(int, InternalSystem);
  void getMarkerProjectionPointsCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerProjectionLengthCmd(int, InternalSystem);
  void getMarkerProjectionLengthCmd(int, CoordSystem, SkyFormat);
  void getMarkerProjectionMethodCmd(int);
  void getMarkerProjectionWidthCmd(int, InternalSystem);
  void getMarkerProjectionWidthCmd(int, CoordSystem, SkyFormat);
  void getMarkerProjectionSystemCmd(int);
  void getMarkerPropertyCmd(unsigned short);
  void getMarkerPropertyCmd(const char*, unsigned short);
  void getMarkerPropertyCmd(int, unsigned short);
  void getMarkerRulerLengthCmd(int, InternalSystem);
  void getMarkerRulerLengthCmd(int, CoordSystem, SkyFormat);
  void getMarkerRulerPointCmd(int, InternalSystem);
  void getMarkerRulerPointCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerRulerSystemCmd(int);
  void getMarkerSelectedCmd();
  void getMarkerSelectedCmd(int);
  void getMarkerSelectedCmd(const Vector&);
  void getMarkerSelectedNumberCmd();
  void getMarkerHighlitedCmd();
  void getMarkerHighlitedCmd(int);
  void getMarkerHighlitedCmd(const Vector&);
  void getMarkerHighlitedNumberCmd();
  void getMarkerTagCmd(const char*);
  void getMarkerTagCmd(int);
  void getMarkerTagCmd(int,int);
  void getMarkerTagDefaultNameCmd();
  void getMarkerTagNumberCmd(const char*);
  void getMarkerTagsCmd();
  void getMarkerTextCmd(int);
  void getMarkerTextRotateCmd(int);
  void getMarkerTypeCmd(int);
  void getMarkerVectorCmd(int, InternalSystem);
  void getMarkerVectorCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerVectorArrowCmd(int);
  void getMarkerVectorLengthCmd(int, InternalSystem);
  void getMarkerVectorLengthCmd(int, CoordSystem, SkyFormat);

  void hasMarkerSelectedCmd();
  void hasMarkerHighlitedCmd();
  void hasMarkerPasteCmd();
  void hasMarkerUndoCmd();

  void markerLayerCmd(MarkerLayer);

  void markerAngleCmd(int, double);
  void markerAngleCmd(int, double, CoordSystem, SkyFrame);
  void markerAnnulusCreateRadiusCmd(int, const Vector&);
  void markerAnnulusDeleteRadiusCmd(int, int);
  void markerAnnulusRadiusCmd(int, double, double, int, InternalSystem);
  void markerAnnulusRadiusCmd(int, double, double, int, CoordSystem,SkyFormat);
  void markerAnnulusRadiusCmd(int, const char*, CoordSystem, SkyFormat);
  void markerBackCmd();
  void markerBackCmd(const char*);
  void markerBackCmd(int);
  void markerBoxAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				 InternalSystem);
  void markerBoxAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				 CoordSystem, SkyFormat);
  void markerBoxAnnulusRadiusCmd(int, const char*, CoordSystem, SkyFormat);
  void markerBoxAnnulusCreateRadiusCmd(int, const Vector&);
  void markerBoxAnnulusDeleteRadiusCmd(int, int);
  void markerBoxRadiusCmd(int, const Vector&, InternalSystem);
  void markerBoxRadiusCmd(int, const Vector&, CoordSystem, SkyFormat);
  void markerBpandaCreateAnglesCmd(int, const Vector&);
  void markerBpandaCreateRadiusCmd(int, const Vector&);
  void markerBpandaDeleteCmd(int, int);
  void markerBpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int);
  void markerBpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int,
			   CoordSystem, SkyFrame);
  void markerBpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,
			   CoordSystem, SkyFormat);
  void markerCallBackCmd(int, CallBack::Type, const char*, const char*);
  void markerCentroidCmd();
  void markerCentroidCmd(const char*);
  void markerCentroidCmd(int);
  void markerCentroidAutoCmd(int);
  void markerCentroidOptionCmd(int,float);
  void markerCircleRadiusCmd(int, double, InternalSystem);
  void markerCircleRadiusCmd(int, double, CoordSystem, SkyFormat);
  void markerColorCmd(const char*);
  void markerColorCmd(const char*, const char*);
  void markerColorCmd(int, const char*);
  void markerCompassArrowCmd(int, int, int);
  void markerCompassLabelCmd(int, const char*, const char*);
  void markerCompassRadiusCmd(int, double, InternalSystem);
  void markerCompassRadiusCmd(int, double, CoordSystem, SkyFormat);
  void markerCompassSystemCmd(int, CoordSystem, SkyFrame);
  void markerCompositeCmd(int, int);
  void markerCompositeDeleteCmd();
  void markerCommandCmd(MarkerFormat, const char*);
  void markerCommandVarCmd(MarkerFormat, const char*);
  void markerCpandaCreateAnglesCmd(int, const Vector&);
  void markerCpandaCreateRadiusCmd(int, const Vector&);
  void markerCpandaDeleteCmd(int, int);
  void markerCpandaEditCmd(int, double, double, int, double, double, int);
  void markerCpandaEditCmd(int, double, double, int, double, double, int,
			  CoordSystem, SkyFrame);
  void markerCpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,CoordSystem, SkyFormat);
  void markerDeleteCmd();
  void markerDeleteCmd(const char*);
  void markerDeleteCmd(int);
  void markerDeleteAllCmd();
  void markerDeleteLastCmd();
  void markerDeleteCallBackCmd(int, CallBack::Type, const char*);
  void markerDeleteTagCmd(int);
  void markerDeleteTagCmd(int, const char*);
  void markerDeleteTagCmd(int, int);
  void markerEditBeginCmd(int, int);
  void markerEditBeginCmd(const Vector&, int);
  void markerEditMotionCmd(const Vector&, int h);
  void markerEditEndCmd();
  void markerEllipseRadiusCmd(int, const Vector&, InternalSystem);
  void markerEllipseRadiusCmd(int, const Vector&, CoordSystem, SkyFormat);
  void markerEllipseAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				     InternalSystem);
  void markerEllipseAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				     CoordSystem, SkyFormat);
  void markerEllipseAnnulusRadiusCmd(int, const char*, 
				     CoordSystem, SkyFormat);
  void markerEllipseAnnulusCreateRadiusCmd(int, const Vector&);
  void markerEllipseAnnulusDeleteRadiusCmd(int, int);
  void markerEpandaCreateAnglesCmd(int, const Vector&);
  void markerEpandaCreateRadiusCmd(int, const Vector&);
  void markerEpandaDeleteCmd(int, int);
  void markerEpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int);
  void markerEpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int,
			   CoordSystem, SkyFrame);
  void markerEpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,
			   CoordSystem, SkyFormat);
  void markerFontCmd(const char*);
  void markerFontCmd(const char*, const char*);
  void markerFontCmd(int, const char*);
  void markerFrontCmd();
  void markerFrontCmd(const char*);
  void markerFrontCmd(int);
  void markerKeyCmd();
  void markerKeyCmd(const Vector&);
  void markerLineCmd(int, const Vector&, const Vector&, InternalSystem);
  void markerLineCmd(int, const Vector&, const Vector&, CoordSystem, SkyFrame);
  void markerLineArrowCmd(int, int, int);
  void markerLineWidthCmd(int);
  void markerLineWidthCmd(int, int);
  void markerListCmd(MarkerFormat, CoordSystem, SkyFrame, SkyFormat, 
		     int strip, int select,
		     unsigned short, unsigned short, List<Tag>&);
  void markerLoadCmd(MarkerFormat,const char*);
  void markerLoadCmd(MarkerFormat,const char*,CoordSystem,SkyFrame);
  void markerLoadCmd(MarkerFormat,int);
  void markerLoadCmd(MarkerFormat,int,CoordSystem,SkyFrame);
  void markerLoadFitsCmd(const char*, const char*, int*, int, const char*);
  void markerMoveCmd(const Vector&);
  void markerMoveCmd(const char*, const Vector&);
  void markerMoveCmd(int id, const Vector&);
  void markerMoveBeginCmd(const Vector&);
  void markerMoveBeginCmd(int, const Vector&);
  void markerMoveMotionCmd(const Vector&);
  void markerMoveMotionCmd(int, const Vector&);
  void markerMoveEndCmd();
  void markerMoveEndCmd(int);
  void markerMoveToCmd(const Vector&, InternalSystem);
  void markerMoveToCmd(const Vector&, CoordSystem, SkyFrame);
  void markerMoveToCmd(const char*, const Vector&, InternalSystem);
  void markerMoveToCmd(const char*, const Vector&, CoordSystem, SkyFrame);
  void markerMoveToCmd(int, const Vector&, InternalSystem);
  void markerMoveToCmd(int, const Vector&, CoordSystem, SkyFrame);
  void markerPointSizeCmd(int,int);
  void markerPolygonCreateVertexCmd(int, int, const Vector&);
  void markerPolygonDeleteVertexCmd(int, int);
  void markerPolygonResetCmd(int, const Vector&, InternalSystem);
  void markerPolygonResetCmd(int, const Vector&, CoordSystem, SkyFormat);
  void markerPropertyCmd(unsigned short, int);
  void markerPropertyCmd(const char*, unsigned short, int);
  void markerPropertyCmd(int, unsigned short, int);
  void markerPropertyCmd(unsigned short, int, const Vector&);
  void markerProjectionCmd(int, const Vector&, const Vector&, 
			   InternalSystem, double, int);
  void markerProjectionCmd(int, const Vector&, const Vector&, 
			   CoordSystem, SkyFrame, 
			   double, int, CoordSystem, SkyFormat);
  void markerProjectionSystemCmd(int, CoordSystem, SkyFrame);
  void markerRotateBeginCmd(int);
  void markerRotateBeginCmd(const Vector&);
  void markerRotateMotionCmd(const Vector&, int h);
  void markerRotateEndCmd();
  void markerRulerPointCmd(int, const Vector&, const Vector&, 
			   CoordSystem, SkyFrame);
  void markerRulerPointCmd(int, const Vector&, const Vector&, InternalSystem);
  void markerRulerSystemCmd(int, CoordSystem, SkyFrame, 
			    CoordSystem, SkyFormat);
  void markerSaveCmd(const char*, MarkerFormat, CoordSystem, SkyFrame,
		     SkyFormat, int strip);
  void markerSaveTemplateCmd(const char*);

  void markerVectorCmd(int, const Vector&, InternalSystem, double, double);
  void markerVectorCmd(int, const Vector&, CoordSystem, SkyFrame, 
		       double, CoordSystem, SkyFormat, double);
  void markerVectorArrowCmd(int, int);

  void markerSelectAllCmd();
  void markerSelectCmd(int); // not used
  void markerSelectCmd(const char*);
  void markerSelectToggleCmd();
  void markerSelectToggleCmd(const Vector&);
  void markerSelectOnlyCmd(int);
  void markerSelectOnlyCmd(const char*);
  void markerSelectOnlyCmd(const Vector&);
  void markerUnselectAllCmd();
  void markerUnselectCmd(int); // not used
  void markerUnselectCmd(const char*); // not used

  void markerHighliteAllCmd(); // not used
  void markerHighliteCmd(int); // not used
  void markerHighliteCmd(const char*); // not used
  void markerHighliteToggleCmd(const Vector&);
  void markerHighliteOnlyCmd(int);
  void markerHighliteOnlyCmd(const char*); // not used
  void markerHighliteOnlyCmd(const Vector&);  // not used
  void markerUnhighliteAllCmd();
  void markerUnhighliteCmd(int); // not used
  void markerUnhighliteCmd(const char*);

  void markerShowCmd(int);
  void markerShowTextCmd(int);
  void markerTagCmd(const char*);
  void markerTagCmd(int, const char*);
  void markerTagEditCmd(const char*, const char*);
  void markerTagDeleteCmd(const char*);
  void markerTagDeleteAllCmd();
  void markerTagUpdateCmd(const char*);
  void markerTextCmd(int, const char*);
  void markerTextRotateCmd(int, int);

  void markerCopyCmd();
  void markerCopyCmd(const char*);
  void markerCutCmd();
  void markerCutCmd(const char*);
  void markerPasteCmd();
  void markerPasteCmd(CoordSystem, SkyFrame);
  void markerUndoCmd();

  void regionHighliteBeginCmd(const Vector& v) {regionSelectBeginCmd(v);}
  void regionHighliteMotionCmd(const Vector& v) {regionSelectMotionCmd(v);}
  void regionHighliteEndCmd();
  void regionHighliteShiftEndCmd();

  void regionSelectBeginCmd(const Vector&);
  void regionSelectMotionCmd(const Vector&);
  void regionSelectEndCmd();
  void regionSelectShiftEndCmd();

  // Pan Zoom Rotate Orient Commands
  void panCmd(const Vector&, const Vector&);
  void panCmd(const Vector&, InternalSystem);
  void panCmd(const Vector&, CoordSystem, SkyFrame);
  void panBBoxCmd(const Vector&);
  void panBeginCmd(const Vector&);
  void panMotionCmd(const Vector&);
  void panEndCmd(const Vector&);
  void panToCmd(const Vector&, InternalSystem);
  void panToCmd(const Vector&, CoordSystem, SkyFrame);
  void rotateBeginCmd();
  void rotateMotionCmd(double);
  void rotateEndCmd();
  void zoomAboutCmd(const Vector&, const Vector&, InternalSystem);
  void zoomAboutCmd(const Vector&, const Vector&, CoordSystem, SkyFrame);
  void zoomToAboutCmd(const Vector&, const Vector&, InternalSystem);
  void zoomToAboutCmd(const Vector&, const Vector&, CoordSystem, SkyFrame);
  void zoomToFitCmd(double =1);

  // Pixel Table Commands
  void getPixelTableCmd(const Vector&, InternalSystem, int, char*);

  // Print Functions
#ifdef _MACOSX
  void macosxPrintCmd();
#endif
#ifdef _WIN32
  void win32PrintCmd();
#endif

  // Scale Commands
  void getColorScaleLevelCmd(int, const Vector&, InternalSystem);

  // 3d
  void get3dViewCmd();
  void get3dRenderMethodCmd();
  void get3dThresholdCmd();
};

#endif
