// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "framergb.h"
#include "fitsimage.h"
#include "ps.h"
#include "contour.h"
#include "NaN.h"

// Frame Member Functions

FrameRGB::FrameRGB(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : FrameBase(i,c,item)
{
  channel = 0;
  rgbSystem = WCS;

  for (int ii=0; ii<3; ii++) {
    view[ii] = 1;

    fits[ii] = NULL;
    cfits[ii] = NULL;

    mosaicCount[ii] = 0;
    sliceCount[ii] = 0;

    bias[ii] = .5;
    contrast[ii] = 1.0;

    colorScale[ii] = NULL;
    
    mosaicType[ii] = NOMOSAIC;
    mosaicSystem[ii] = WCS;

    contour[ii] = NULL;
  }

  colorCount = 0;
  colorCells = NULL;

  keyFits_ = NULL;
  currentFits_ = NULL;
  cfits[channel] = NULL;

  setChannel(channel);
}

FrameRGB::~FrameRGB()
{
  for (int ii=0; ii<3; ii++) {
    if (colorScale[ii])
      delete colorScale[ii];
  }

  if (colorCells)
    delete [] colorCells;
}

void FrameRGB::alignWCS()
{
  if (!wcsAlignment ||
      !(keyFits_ && *keyFits_) ||
      !(*keyFits_)->hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
  }
  else
    calcAlignWCS(*keyFits_, wcsSystem, wcsSky,
		 &wcsOrientation, &wcsOrientationMatrix, &wcsRotation);

  // image,pysical,amplifier,detector are ok, check for wcs
  if (rgbSystem >= WCS) {
    for (int ii=0; ii<3; ii++) {
      if (fits[ii] && !fits[ii]->hasWCS(rgbSystem)) {
	// ok, don't have requested coordinate system
	// down grade to image
	rgbSystem = IMAGE;
	break;
      }
    }
  }

  // rgb align
  for (int ii=0; ii<3; ii++) {
    rgb[ii].identity();

    if (fits[ii] && keyFits_ && *keyFits_) {
      switch (rgbSystem) {
      case IMAGE:
	// nothing to do here
	break;
      case PHYSICAL:
	if (fits[ii] != *keyFits_) 
	  rgb[ii] = 
	    fits[ii]->getImageToPhysical()*(*keyFits_)->getPhysicalToImage();
	break;
      case AMPLIFIER:
	if (fits[ii] != *keyFits_) 
	  rgb[ii] = fits[ii]->getImageToAmplifier() * 
	    (*keyFits_)->getAmplifierToImage();
	break;
      case DETECTOR:
	if (fits[ii] != *keyFits_) 
	  rgb[ii] = fits[ii]->getImageToDetector() * 
	    (*keyFits_)->getDetectorToImage();
	break;
      default:
	rgbAlignWCS(ii);
	break;
      }
    }

    if (DebugRGB) 
      cerr << "rgb[" << ii << "] " << rgb[ii] << endl;
  }
}   

void FrameRGB::alignWCS(FitsImage* ptr)
{
  if (!wcsAlignment ||
      !(keyFits_ && *keyFits_) ||
      !ptr ||
      !(*keyFits_)->hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
  }
  else
    calcAlignWCS(ptr, *keyFits_, wcsSystem, wcsSky,
		 &wcsOrientation, &wcsOrientationMatrix, &wcsRotation, &zoom_);

  // image,pysical,amplifier,detector are ok, check for wcs
  if (rgbSystem >= WCS) {
    for (int ii=0; ii<3; ii++) {
      if (fits[ii] && !fits[ii]->hasWCS(rgbSystem)) {
	// ok, don't have requested coordinate system
	// down grade to image
	rgbSystem = IMAGE;
	break;
      }
    }
  }

  // rgb align
  for (int ii=0; ii<3; ii++) {
    rgb[ii].identity();

    if (fits[ii] && keyFits_ && *keyFits_) {
      switch (rgbSystem) {
      case IMAGE:
	// nothing to do here
	break;
      case PHYSICAL:
	if (fits[ii] != *keyFits_) 
	  rgb[ii] = 
	    fits[ii]->getImageToPhysical()*(*keyFits_)->getPhysicalToImage();
	break;
      case AMPLIFIER:
	if (fits[ii] != *keyFits_) 
	  rgb[ii] = fits[ii]->getImageToAmplifier() * 
	    (*keyFits_)->getAmplifierToImage();
	break;
      case DETECTOR:
	if (fits[ii] != *keyFits_) 
	  rgb[ii] = fits[ii]->getImageToDetector() * 
	    (*keyFits_)->getDetectorToImage();
	break;
      default:
	rgbAlignWCS(ii);
	break;
      }
    }

    if (DebugRGB) 
      cerr << "rgb[" << ii << "] " << rgb[ii] << endl;
  }
}

void FrameRGB::getInfo(const Vector& vv, InternalSystem ref,
		       Matrix& (FitsImage::*matrixToData)(),
		       Matrix& (FitsImage::*matrixToImage)(),
		       SkyFrame sky, SkyFormat format, char* var)
{
  Base::getInfo(vv, ref, matrixToData, matrixToImage, sky, format, var);

  char* array[3] = {"value,red","value,green","value,blue"};
  for (int ii=0; ii<3; ii++) {

    if (!cfits[ii])
      continue;

    FitsImage* sptr = cfits[ii];
    int mosaic = mosaicCount[ii]>1 ? 1:0;

    FitsBound* params = sptr->getDataParams(frScale[ii].scanMode());

    do {
      Vector img = vv * (sptr->*matrixToData)();

      if (img[0]>=params->xmin && img[0]<params->xmax && 
	  img[1]>=params->ymin && img[1]<params->ymax) {

	Tcl_SetVar2(interp,var,array[ii],(char*)sptr->getValue(img),0);

	break;
      }
      else {
	if (mosaic) {
	  sptr = sptr->nextMosaic();
	  if (sptr)
	    params = sptr->getDataParams(currentScale->scanMode());
	}
      }
    }
    while (mosaic && sptr);
  }
}

unsigned char* FrameRGB::fillImage(int width, int height,
				   double* (FitsImage::*getMatrix)())
{
  // img
  unsigned char* img = new unsigned char[width*height*3];
  memset(img,0,width*height*3);

  // mk
  char* mk = new char[width*height];
  memset(mk,0,width*height);

  // one channel at a time
  for (int kk=0; kk<3; kk++) {
    if (!view[kk] || !fits[kk])
      continue;

    // basics
    int length = colorScale[kk]->size() - 1;
    const unsigned char* table = colorScale[kk]->psColors();

    FitsImage* sptr = cfits[kk];
    int mosaic = mosaicCount[kk]>1 ? 1:0;

    // variable
    double* mm = (sptr->*getMatrix)();
    FitsBound* params = sptr->getDataParams(frScale[kk].scanMode());
    int srcw = sptr->width();

    double ll = sptr->getLowDouble();
    double hh = sptr->getHighDouble();
    double diff = hh - ll;

    // main loop
    unsigned char* dest = img;
    char* mkptr = mk;
    for (long jj=0; jj<height; jj++) {
      for (long ii=0; ii<width; ii++, dest+=3, mkptr++) {

	if (mosaic) {
	  sptr = cfits[kk];

	  mm = (sptr->*getMatrix)();
	  params = sptr->getDataParams(frScale[kk].scanMode());
	  srcw = sptr->width();

	  ll = sptr->getLowDouble();
	  hh = sptr->getHighDouble();
	  diff = hh - ll;
	}

	do {
	  double xx = ii*mm[0] + jj*mm[3] + mm[6];
	  double yy = ii*mm[1] + jj*mm[4] + mm[7];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax) {
	    double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	    if (!isnand(value)) {
	      if (value <= ll)
		*(dest+kk) = *table;
	      else if (value >= hh)
		*(dest+kk) = *(table+length);
	      else
		*(dest+kk) = *(table+((int)(((value - ll)/diff * length) +.5)));
	      *mkptr =2;
	    }
	    else if (*mkptr < 2)
	      *mkptr =1;

	    break;
	  }
	  else {
	    if (mosaic) {
	      sptr = sptr->nextMosaic();

	      if (sptr) {
		mm = (sptr->*getMatrix)();
		params = sptr->getDataParams(frScale[kk].scanMode());
		srcw = sptr->width();

		ll = sptr->getLowDouble();
		hh = sptr->getHighDouble();
		diff = hh - ll;
	      }
	    }
	  }
	}
	while (mosaic && sptr);
      }
    }
  }

  // now fill in bg
  {
    unsigned char* dest = img;
    char* mkptr = mk;
    for (int jj=0; jj<height; jj++)
      for (int ii=0; ii<width; ii++, dest+=3, mkptr++) {
	if (*mkptr == 2) // good value
	  ;
	else if (*mkptr == 1) { // nan
	  *dest = (unsigned char)nanColor->red;
	  *(dest+1) = (unsigned char)nanColor->green;
	  *(dest+2) = (unsigned char)nanColor->blue;
	}
	else { // bg
	  *dest = (unsigned char)bgColor->red;
	  *(dest+1) = (unsigned char)bgColor->green;
	  *(dest+2) = (unsigned char)bgColor->blue;
	}
      }	
  }

  return img;
}

BBox FrameRGB::imageBBox()
{
  // returns imageBBox in IMAGE coords
  //   and extends edge to edge

  // make sure we have the rgb matrices defined
  alignWCS();

  BBox rr;
  int first=1;
  for (int ii=0; ii<3; ii++) {
    if (fits[ii]) {
      FitsImage* ptr = fits[ii];
      while (ptr) {
	FitsBound* params = ptr->getDataParams(frScale[ii].scanMode());
	Matrix mm = ptr->getWCSToRef() * rgb[ii] * dataToImage;

	Vector aa = Vector(params->xmin,params->ymin) * mm;
	if (first) {
	  rr = BBox(aa,aa);
	  first = 0;
	}
	else
	  rr.bound(aa);

	rr.bound(Vector(params->xmax,params->ymin) * mm);
	rr.bound(Vector(params->xmax,params->ymax) * mm);
	rr.bound(Vector(params->xmin,params->ymax) * mm);

	ptr = ptr->nextMosaic();
      }
    }
  }

  return rr;
}

void FrameRGB::loadRGBCube(MemType which, const char* fn, FitsImage* img)
{
  if (!img || !img->isValid() || !(img->isImage() || img->isCompress()) || (img->depth() != 3))
    goto error;

  fits[0] = img;

  if (img->isCompress())
    which = COMPRESS;

  switch (which) {
  case ALLOC:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextAlloc(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextAlloc(this, fn, fits[1]->fitsFile());
    break;
  case ALLOCGZ:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextAllocGZ(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextAllocGZ(this, fn, fits[1]->fitsFile());
    break;
  case CHANNEL:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextChannel(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextChannel(this, fn, fits[1]->fitsFile());
    break;
  case MMAP:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextMMap(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextMMap(this, fn, fits[1]->fitsFile());
    break;
  case SMMAP:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextSMMap(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextSMMap(this, fn, fits[1]->fitsFile());
    break;
  case MMAPINCR:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextMMapIncr(this, fn,fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextMMapIncr(this, fn,fits[1]->fitsFile());
    break;
  case SHARE:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextShare(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextShare(this, fn, fits[1]->fitsFile());
    break;
  case SSHARE:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextSShare(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextSShare(this, fn, fits[1]->fitsFile());
    break;
  case SOCKET:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextSocket(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextSocket(this, fn, fits[1]->fitsFile());
    break;
  case SOCKETGZ:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextSocketGZ(this, fn,fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextSocketGZ(this, fn,fits[1]->fitsFile());
    break;
  case VAR:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextVar(this, fn, fits[0]->fitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextVar(this, fn, fits[1]->fitsFile());
    break;
    case COMPRESS:
      if (fits[0] && fits[0]->isValid())
	fits[1] = new FitsImageFitsNextCompress(this, img,fits[0]->baseFile());
      if (fits[1] && fits[1]->isValid())
	fits[2] = new FitsImageFitsNextCompress(this, img,fits[1]->baseFile());
      break;
  }

  // is everything ok?
  if (fits[0] && fits[0]->isValid() &&
      (fits[0]->isImage() || fits[0]->isCompress()) &&
      fits[1] && fits[1]->isValid() &&
      (fits[1]->isImage() || fits[1]->isCompress()) &&
      fits[2] && fits[2]->isValid() &&
      (fits[2]->isImage() || fits[2]->isCompress())) {

    loadRGBFinish();
    return;
  }

 error:
  unloadAllFits();
  reset();
  resetClip(0);
  update(MATRIX);
    
  Tcl_AppendResult(interp, "Unable to load rgb cube file", NULL);
  result = TCL_ERROR;
  return;
}

void FrameRGB::loadRGBImage(MemType which, const char* fn, FitsImage* img)
{
  FitsImage* r = img;
  FitsImage* g = NULL;
  FitsImage* b = NULL;

  if (!img || !img->isValid() || !(img->isImage() || img->isCompress()))
    goto error;

  switch (which) {
  case ALLOC:
    if (r && r->isValid())
      g = new FitsImageMosaicNextAlloc(this, fn, r->fitsFile(), 
				       FitsFile::NOFLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextAlloc(this, fn, g->fitsFile(), 
				       FitsFile::NOFLUSH);
    break;
  case ALLOCGZ:
    if (r && r->isValid())
      g = new FitsImageMosaicNextAllocGZ(this,fn,r->fitsFile(), 
					 FitsFile::NOFLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextAllocGZ(this,fn,g->fitsFile(), 
					 FitsFile::NOFLUSH);
    break;
  case CHANNEL:
    if (r && r->isValid())
      g = new FitsImageMosaicNextChannel(this,fn,r->fitsFile(), 
					 FitsFile::NOFLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextChannel(this,fn,g->fitsFile(), 
					 FitsFile::NOFLUSH);
    break;
  case MMAP:
    if (r && r->isValid())
      g = new FitsImageMosaicNextMMap(this, fn, r->fitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextMMap(this, fn, g->fitsFile());
    break;
  case MMAPINCR:
    if (r && r->isValid())
      g = new FitsImageMosaicNextMMapIncr(this, fn, r->fitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextMMapIncr(this, fn, g->fitsFile());
    break;
  case SHARE:
    if (r && r->isValid())
      g = new FitsImageMosaicNextShare(this,fn, r->fitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextShare(this,fn, g->fitsFile());
    break;
  case SOCKET:
    if (r && r->isValid())
      g = new FitsImageMosaicNextSocket(this,fn,r->fitsFile(), 
					FitsFile::FLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextSocket(this,fn,g->fitsFile(), 
					FitsFile::FLUSH);
    break;
  case SOCKETGZ:
    if (r && r->isValid())
      g = new FitsImageMosaicNextSocketGZ(this,fn,r->fitsFile(), 
					  FitsFile::FLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextSocketGZ(this,fn,g->fitsFile(), 
					  FitsFile::FLUSH);
    break;
  case VAR:
    if (r && r->isValid())
      g = new FitsImageMosaicNextVar(this, fn, r->fitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextVar(this, fn, g->fitsFile());
    break;
  }

  // ok, figure out which is which channel
  fits[0] = fits[1] = fits[2] = NULL;

  if (r->findKeyword("EXTNAME")) {
    char* c = r->getKeyword("EXTNAME");
    if (!strncmp(c,"RED",3))
      fits[0] = r;
    else if (!strncmp(c,"GREEN",3))
      fits[1] = r;
    else if (!strncmp(c,"BLUE",3))
      fits[2] = r;

    if (c)
      delete [] c;
  }
  else
    fits[0] = r;

  if (g->findKeyword("EXTNAME")) {
    char* c = g->getKeyword("EXTNAME");
    if (!strncmp(c,"RED",3))
      fits[0] = g;
    else if (!strncmp(c,"GREEN",3))
      fits[1] = g;
    else if (!strncmp(c,"BLUE",3))
      fits[2] = g;

    if (c)
      delete [] c;
  }
  else
    fits[1] = g;

  if (b->findKeyword("EXTNAME")) {
    char* c = b->getKeyword("EXTNAME");
    if (!strncmp(c,"RED",3))
      fits[0] = b;
    else if (!strncmp(c,"GREEN",3))
      fits[1] = b;
    else if (!strncmp(c,"BLUE",3))
      fits[2] = b;

    if (c)
      delete [] c;
  }
  else
    fits[2] = b;

  // is everything ok?
  if (fits[0] && fits[0]->isValid() && 
      (fits[0]->isImage() || fits[0]->isCompress()) &&
      fits[1] && fits[1]->isValid() && 
      (fits[1]->isImage() || fits[1]->isCompress()) &&
      fits[2] && fits[2]->isValid() && 
      (fits[2]->isImage() || fits[2]->isCompress())) {

    loadRGBFinish();
    return;
  }

 error:
  unloadAllFits();
  reset();
  resetClip(0);
  update(MATRIX);
    
  Tcl_AppendResult(interp, "Unable to load rgb image file", NULL);
  result = TCL_ERROR;
  return;
}

void FrameRGB::loadRGBFinish()
{
  for (int ii=0; ii<3; ii++) {
    cfits[ii] = fits[ii];

    mosaicCount[ii] = 1;
    sliceCount[ii] = 1;
    mosaicType[ii] = NOMOSAIC;
    mosaicSystem[ii] = WCS;
  }

  // update clip values
  channel = 0;
  slice = 1;
  for (int ii=0; ii<3; ii++) {
    currentFits_ = fits[ii];

    setChannel(ii);
    resetClip(0);
  }

  channel = 0;
  slice = 1;
  currentFits_ = fits[channel];
  setChannel(channel);

  keyFits_ = channelFits_;

  resetValues();
  update(MATRIX);
}

void FrameRGB::reset()
{
  for (int ii=0; ii<3; ii++) {
    bias[ii] = 0.5;
    contrast[ii] = 1.0;
  }

  FrameBase::reset();
}

void FrameRGB::rgbAlignWCS(int ii)
{
  if (keyFits_ && *keyFits_  && (*keyFits_)->hasWCS(rgbSystem))
    rgb[ii] = calcAlignWCS(*keyFits_, fits[ii], rgbSystem, FK5);

  if (DebugRGB)
    cerr << "rgbAlignWCS " << rgb[ii] << endl;
}

void FrameRGB::setBinCursor(const Vector& v)
{
  for (int ii=0; ii<3; ii++)
    if (fits[ii])
      fits[ii]->setBinCursor(cursor);
}

void FrameRGB::setChannel(int ii)
{
  //  currentFits_ = fits[ii];
  //  cfits[ii] = fits[ii];

  channelFits_ = &fits[ii];

  currentScale = &frScale[ii];
  currentMosaicCount = &mosaicCount[ii];
  currentSliceCount = &sliceCount[ii];
  currentMosaic = &mosaicType[ii];
  currentMosaicSystem = &mosaicSystem[ii];

  currentcontour = &contour[ii];
}

void FrameRGB::setCurrentFits(FitsImage* f)
{
  currentFits_ = f;
  cfits[channel] = f;
}

void FrameRGB::setKeyFits()
{
  if (!keyFits_)
    keyFits_ = channelFits_;
}  

void FrameRGB::updateBin(const Matrix& mx)
{
  // Note: cursor is in USER coords, imageCenter() in IMAGE coords
  cursor = imageCenter();

  if (keyFits_ && *keyFits_ && (*keyFits_ == currentFits_)) {
    // update markers,crosshair,aux contours
    crosshair *= mx;

    updateMarkerCoords(&userMarkers, mx);
    updateMarkerCoords(&catalogMarkers, mx);

    // aux contours
    {
      Contour* ptr=auxcontours.head();
      while (ptr) {
	ptr->updateCoords(mx);
	ptr=ptr->next();
      }
    }
  }

  FitsImage* currentSave = currentFits_;

  for (int ii=0; ii<3; ii++) 
    if (fits[ii] && fits[ii]->isHist()) {
      // update bin

      if (fits[ii] == currentSave && contour[ii])
	contour[ii]->updateCoords(mx);

      currentFits_ = fits[ii];
      cfits[ii] = fits[ii];
      setChannel(ii);

      // delete any previous slices
      {
	FitsImage* ptr = (*channelFits_)->nextSlice();
	(*channelFits_)->setNextSlice(NULL);
	while (ptr) {
	  FitsImage* tmp = ptr->nextSlice();
	  delete ptr;
	  ptr = tmp;
	}
      }

      // finish bin
      *currentMosaicCount = 1;
      *currentSliceCount = 1;

      // bin data cube
      int bd = (*channelFits_)->binDepth();
      if (bd > 1) {
	FitsImage* ptr = *channelFits_;
	for (int i=1; i<bd; i++) {
	  FitsImage* next = NULL;
	  next = 
	    new FitsImageFitsNextHist(this,*channelFits_, ptr->baseFile());
	  if (next && next->isValid()) {
	    ptr->setNextSlice(next);
	    ptr = next;
	    (*currentSliceCount)++;
	  }
	  else {
	    if (next)
	      delete next;
	    break;
	  }
	}
      }

      updateBinFileNames();
      resetClip(0);
    }

  slice = 1;
  currentFits_ = fits[channel];
  cfits[channel] = fits[channel];
  setChannel(channel);

  alignWCS();
  updateNow(MATRIX); // because we have changed zoom

  // update marker call backs
  // wait til matrices have been updated so that any dialogs will print
  // the correct coords

  updateMarkerCBs(&userMarkers);
  updateMarkerCBs(&catalogMarkers);
}

void FrameRGB::updateColorCells(unsigned char* cells, int cnt)
{
  if (DebugRGB) 
    cerr << "updateColorCells" << endl;

  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }

  memcpy(colorCells, cells, cnt*3);
}

void FrameRGB::updateColorScale()
{
  // we need colors before we can construct a scale
  if (!colorCells)
    return;

  if (DebugRGB) 
    cerr << "updateColorScale" << endl;

  // since we need to know about bytes per pixel and byte order, 
  // we may need to wait til we have an XImage available and try again
  //  if (!baseXImage)
  //    return;

  for (int ii=0; ii<3; ii++) {
    if (colorScale[ii])
      delete colorScale[ii];

    switch (frScale[ii].colorScaleType()) {
    case FrScale::LINEARSCALE:
      colorScale[ii] = 
	new LinearScaleRGB(ii, colorCount, colorCells, colorCount);
      break;
    case FrScale::LOGSCALE:
      colorScale[ii] =
	new LogScaleRGB(ii, SCALESIZE, colorCells, colorCount,
			frScale[ii].low(),
			frScale[ii].high());
      break;
    case FrScale::POWSCALE:
      colorScale[ii] =
	new PowScaleRGB(ii, SCALESIZE, colorCells, colorCount,
			frScale[ii].low(),
			frScale[ii].high());
      break;
    case FrScale::SQRTSCALE:
      colorScale[ii] =
	new SqrtScaleRGB(ii, SCALESIZE, colorCells, colorCount);
      break;
    case FrScale::SQUAREDSCALE:
      colorScale[ii] =
	new SquaredScaleRGB(ii, SCALESIZE, colorCells, colorCount);
      break;
    case FrScale::HISTEQUSCALE:
      calcHistEqu();
      colorScale[ii] = 
	new HistEquScaleRGB(ii, SCALESIZE, colorCells, colorCount, 
			    frScale[ii].histequ(), HISTEQUSIZE);
      break;
    }
  }
}

void FrameRGB::updateMatrices()
{
  // very important! do this first
  FrameBase::updateMatrices();

  // alway identity
  Matrix3d threedToRef3d;
  Matrix3d refToUser3d; 
  Matrix3d userToWidget3d;
  Matrix3d userToPanner3d;

  for (int ii=0; ii<3; ii++) {
    FitsImage* ptr = fits[ii];
    while (ptr) {
      FitsImage* sptr = ptr;
      while (sptr) {
	sptr->updateMatrices(rgb[ii], refToUser, userToWidget,
			     widgetToCanvas, userToPanner);
	sptr = sptr->nextSlice();
      }
      ptr = ptr->nextMosaic();
    }
  }
}

void FrameRGB::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  for (int ii=0; ii<3; ii++) {
    FitsImage* ptr = fits[ii];
    while (ptr) {
      FitsImage* sptr = ptr;
      while (sptr) {
	sptr->updateMagnifierMatrices(refToMagnifier);
	sptr = sptr->nextSlice();
      }
      ptr = ptr->nextMosaic();
    }
  }
}

void FrameRGB::updatePS(float scale, int width, int height)
{
  Matrix mx = psMatrix(scale, width, height);

  for (int kk=0; kk<3; kk++)
    if (fits[kk]) {
      FitsImage* ptr = cfits[kk];
      while (ptr) {
	ptr->updatePS(mx);
	ptr = ptr->nextMosaic();
      }
    }
}

void FrameRGB::unloadAllFits()
{
  if (DebugPerf)
    cerr << "Frame::RGBunloadAllFits" << endl;

  for (int ii=0; ii<3; ii++) {
    // fits
    {
      FitsImage* ptr = fits[ii];
      while (ptr) {
	FitsImage* sptr = ptr->nextSlice();
	while (sptr) {
	  FitsImage* stmp = sptr->nextSlice();
	  delete sptr;
	  sptr = stmp;
	}

	FitsImage* tmp = ptr->nextMosaic();
	delete ptr;
	ptr = tmp;
      }
    }

    fits[ii] = NULL;
    cfits[ii] = NULL;

    mosaicCount[ii] = 0;
    sliceCount[ii] = 0;
    mosaicType[ii] = NOMOSAIC;
    mosaicSystem[ii] = WCS;

    if (contour[ii])
      delete contour[ii];
    contour[ii] = NULL;

    rgb[ii].identity();

    // reset user mode
    // if (frScale[ii].clipMode() == FrScale::USERCLIP)
    // frScale[ii].setClipMode(FrScale::MINMAX);

    frScale[ii].deleteHistogramX();
    frScale[ii].deleteHistogramY();
    frScale[ii].deleteHistequ();

    // always (for HISTEQU and LOG)
    updateColorScale();
  }

  keyFits_ = NULL;
  FrameBase::unloadFits();
}

void FrameRGB::unloadFits()
{
  if (DebugPerf)
    cerr << "FrameRGB::unloadFits" << endl;

  // now, delete any fits for current channel
  {
    FitsImage* ptr = fits[channel];
    while (ptr) {
      FitsImage* sptr = ptr->nextSlice();
      while (sptr) {
	FitsImage* stmp = sptr->nextSlice();
	delete sptr;
	sptr = stmp;
      }

      FitsImage* tmp = ptr->nextMosaic();
      delete ptr;
      ptr = tmp;
    }
    fits[channel] = NULL;
    cfits[channel] = NULL;
  }

  if (contour[channel])
    delete contour[channel];
  contour[channel] = NULL;

  rgb[channel].identity();

  keyFits_ = NULL;
  currentFits_ = NULL;
  cfits[channel] = NULL;

  slice = 1;
  *currentMosaicCount = 0;
  *currentSliceCount = 0;
  *currentMosaic = NOMOSAIC;

  // reset user mode
  // if (frScale[channel].clipMode() == FrScale::USERCLIP)
  // frScale[channel].setClipMode(FrScale::MINMAX);

  resetClip(0);
}

// Commands

void FrameRGB::getColormapCmd()
{
  ostringstream str;

  str << setiosflags(ios::fixed);
  for (int ii=0; ii<3; ii++)
    str << bias[ii] << ' ';
  for (int ii=0; ii<3; ii++)
    str << contrast[ii] << ' ';
  str << invert << ' ' << ends;

  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void FrameRGB::getRGBChannelCmd()
{
  switch (channel) {
  case 0:
    Tcl_AppendResult(interp, "red", NULL);
    return;
  case 1:
    Tcl_AppendResult(interp, "green", NULL);
    return;
  case 2:
    Tcl_AppendResult(interp, "blue", NULL);
    return;
  }
}

void FrameRGB::getRGBSystemCmd()
{
  printCoordSystem(rgbSystem);
}

void FrameRGB::getRGBViewCmd()
{
  for (int ii=0; ii<3; ii++)
    Tcl_AppendElement(interp, view[ii] ? "1" : "0");
}

void FrameRGB::getTypeCmd()
{
  Tcl_AppendResult(interp, "rgb", NULL);
}

void FrameRGB::getWCSAlignCmd()
{
  ostringstream str;
  if (keyFits_ && *keyFits_)
    str << (unsigned short*)*keyFits_ << ends;
  else
    str << (unsigned short*)NULL << ends;

  Tcl_AppendResult(interp, (wcsAlignment ? "1" : "0"), " ", 
		   coordSystemStr_[wcsSystem-IMAGE], " ",
		   skyFrameStr_[wcsSky-FK4], " ",
		   str.str().c_str(), NULL);
}

void FrameRGB::loadRGBCubeAllocCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsAlloc(this, ch, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOC,fn,img);
}

void FrameRGB::loadRGBCubeAllocGZCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsAllocGZ(this, ch, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOCGZ,fn,img);
}

void FrameRGB::loadRGBCubeChannelCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsChannel(this, interp, ch, fn, 
					    FitsFile::NOFLUSH);
  loadRGBCube(CHANNEL,fn,img);
}

void FrameRGB::loadRGBCubeMMapCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsMMap(this, fn);
  setScanModeIncr(lm);
  loadRGBCube(MMAP,fn,img);
}

void FrameRGB::loadRGBCubeSMMapCmd(const char* hdr, const char* fn, 
				   LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsSMMap(this, hdr, fn);
  setScanModeIncr(lm);
  loadRGBCube(SMMAP,fn,img);
}

void FrameRGB::loadRGBCubeMMapIncrCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsMMapIncr(this, fn);
  setScanModeIncr(lm);
  loadRGBCube(MMAPINCR,fn,img);
}

void FrameRGB::loadRGBCubeShareCmd(ShmType type, int id, const char* fn,
				   LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsShare(this, type, id, fn);
  setScanModeIncr(lm);
  loadRGBCube(SHARE,fn,img);
}

void FrameRGB::loadRGBCubeSShareCmd(ShmType type, int hdr, int id,
				    const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsSShare(this, type, hdr, id, fn);
  setScanModeIncr(lm);
  loadRGBCube(SSHARE,fn,img);
}

void FrameRGB::loadRGBCubeSocketCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsSocket(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKET,fn,img);
}

void FrameRGB::loadRGBCubeSocketGZCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsSocketGZ(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKETGZ,fn,img);
}

void FrameRGB::loadRGBCubeVarCmd(const char* ch, const char* fn,
				     LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsVar(this, interp, ch, fn);
  setScanModeIncr(lm);
  loadRGBCube(VAR,fn,img);
}

void FrameRGB::loadRGBImageAllocCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicAlloc(this, ch, fn, FitsFile::NOFLUSH);
  loadRGBImage(ALLOC,fn,img);
}

void FrameRGB::loadRGBImageAllocGZCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicAllocGZ(this, ch, fn, FitsFile::NOFLUSH);
  loadRGBImage(ALLOCGZ,fn,img);
}

void FrameRGB::loadRGBImageChannelCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicChannel(this, interp, ch, fn, 
					      FitsFile::NOFLUSH);
  loadRGBImage(CHANNEL,fn,img);
}

void FrameRGB::loadRGBImageMMapCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicMMap(this, fn);
  setScanModeIncr(lm);
  loadRGBImage(MMAP,fn,img);
}

void FrameRGB::loadRGBImageMMapIncrCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicMMapIncr(this, fn);
  setScanModeIncr(lm);
  loadRGBImage(MMAPINCR,fn,img);
}

void FrameRGB::loadRGBImageShareCmd(ShmType type, int id, const char* fn,
				    LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicShare(this, type, id, fn);
  setScanModeIncr(lm);
  loadRGBImage(SHARE,fn,img);
}

void FrameRGB::loadRGBImageSocketCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicSocket(this, s, fn, FitsFile::FLUSH);
  loadRGBImage(SOCKET,fn,img);
}

void FrameRGB::loadRGBImageSocketGZCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicSocketGZ(this, s, fn, FitsFile::FLUSH);
  loadRGBImage(SOCKETGZ,fn,img);
}

void FrameRGB::loadRGBImageVarCmd(const char* ch, const char* fn,
				      LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicVar(this, interp, ch, fn);
  setScanModeIncr(lm);
  loadRGBImage(VAR,fn,img);
}

void FrameRGB::loadArrRGBCubeAllocCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrAlloc(this, ch, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOC,fn,img);
}

void FrameRGB::loadArrRGBCubeAllocGZCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrAllocGZ(this, ch, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOCGZ,fn,img);
}

void FrameRGB::loadArrRGBCubeChannelCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrChannel(this, interp, ch, fn, 
					   FitsFile::NOFLUSH);
  loadRGBCube(CHANNEL,fn,img);
}

void FrameRGB::loadArrRGBCubeMMapCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrMMap(this, fn);
  loadRGBCube(MMAP,fn,img);
}

void FrameRGB::loadArrRGBCubeMMapIncrCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrMMapIncr(this, fn);
  loadRGBCube(MMAPINCR,fn,img);
}

void FrameRGB::loadArrRGBCubeShareCmd(ShmType type, int id, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrShare(this, type, id, fn);
  loadRGBCube(SHARE,fn,img);
}

void FrameRGB::loadArrRGBCubeSocketCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrSocket(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKET,fn,img);
}

void FrameRGB::loadArrRGBCubeSocketGZCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrSocketGZ(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKETGZ,fn,img);
}

void FrameRGB::loadArrRGBCubeVarCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrVar(this, interp, ch, fn);
  loadRGBCube(VAR,fn,img);
}

void FrameRGB::setRGBChannelCmd(const char* c)
{
  if (!strncmp(c,"red",3))
    channel = 0;
  else if (!strncmp(c,"gre",3))
    channel = 1;
  else if (!strncmp(c,"blu",3))
    channel = 2;
  else
    channel = 0;

  currentFits_ = cfits[channel];

  setChannel(channel);

  // execute any update callbacks
  updateCBMarkers();

 // always update
  update(BASE);
}

void FrameRGB::setRGBSystemCmd(CoordSystem sys)
{
  rgbSystem = sys;

  // save current matrix
  Matrix old[3];
  for (int ii=0; ii<3; ii++)
    old[ii] = rgb[ii];

  alignWCS();

  // fix any contours
  for(int ii=0; ii<3; ii++)
    if (contour[ii]) {
      Matrix mm = old[ii].invert() * rgb[ii];
      contour[ii]->updateCoords(mm);
    }

  update(MATRIX);
}

void FrameRGB::setRGBViewCmd(int r, int g, int b)
{
  view[0] = r ? 1 : 0;
  view[1] = g ? 1 : 0;
  view[2] = b ? 1 : 0;

  update(BASE); // always update
}

void FrameRGB::sliceCmd(int s)
{
  // fits
  for (int ii=0; ii<3; ii++) {
    cfits[ii] = fits[ii];
    if (cfits[ii] && (s>0) && (s<=sliceCount[ii])) {
      for (int i=1; i<s; i++)
	cfits[ii] = cfits[ii]->nextSlice();
    }
  }

  currentFits_ = cfits[channel];
  slice = s;

  // execute any update callbacks
  updateCBMarkers();

  update(BASE);
}


