// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "base.h"
#include "framebase.h"
#include "fitsimage.h"
#include "NaN.h"
#include "blt.h"

void FrameBase::bltProjection(char* xname, char* yname, 
			      char* xcname, char* ycname,
			      CoordSystem sys, SkyFrame sky,
			      const Vector& p1, const Vector& p2, 
			      double width, int avg)
{
  if (!currentFits_)
    return;

  int num = (p2-p1).length() +1;

  double* x = (double*)malloc(num*sizeof(double));
  double* y = (double*)malloc(num*sizeof(double));
  double* xc = (double*)malloc(num*sizeof(double));
  double* yc = (double*)malloc(num*sizeof(double));

  bltProjFits(x, y, xc, yc, num, sys, sky, p1, p2, width, avg);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, y, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* xxc;
  if (Blt_GetVector(interp, xcname, &xxc) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xxc, xc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yyc;
  if (Blt_GetVector(interp, ycname, &yyc) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yyc, yc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  return;

 error:
  result = TCL_ERROR;
  return;
}

void FrameBase::bltProjFits(double* x, double* y, 
			    double* xc, double* yc, int num,
			    CoordSystem sys, SkyFrame sky,
			    const Vector& p1, const Vector& p2, 
			    double width, int avg)
{
  int* cnt = new int[num];
  Vector s = (p2-p1).normalize();

  // basics
  FitsImage* sptr = currentFits_;
  int mosaic = isMosaic();

  // variable
  FitsBound* params = sptr->getDataParams(currentScale->scanMode());

  // main loop
  for (long ii=0; ii<num; ii++) {
    Vector t = p1 + s*ii;

    x[ii] = ii+1;

    if (mosaic) {
      sptr = currentFits_;
      params = sptr->getDataParams(currentScale->scanMode());
    }

    do {
      Vector tv = sptr->mapFromRef(t,sys,sky);
      xc[ii] = tv[0];
      yc[ii] = tv[1];

      Vector z = t * sptr->getRefToData();

      if (z[0]>=params->xmin && z[0]<params->xmax && 
	  z[1]>=params->ymin && z[1]<params->ymax) {

	// check for nan
	double v = sptr->getValueDouble(z);
	if (!isnand(v)) {
	  y[ii] = v;
	  cnt[ii] = 1;
	}
	else
	  y[ii] = getnand();

	FitsImage* spptr = currentFits_;
	FitsBound* pparams = spptr->getDataParams(currentScale->scanMode());

	Vector ss = Vector(-s[1],s[0]);

	for (long jj=1; jj<width; jj++) {
	  Vector tt = p1 + s*ii + ss*jj;

	  if (mosaic) {
	    spptr = currentFits_;
	    pparams = spptr->getDataParams(currentScale->scanMode());
	  }

	  do {
	    Vector zz = tt * spptr->getRefToData();
		
	    if (zz[0]>=pparams->xmin && zz[0]<pparams->xmax && 
		zz[1]>=pparams->ymin && zz[1]<pparams->ymax) {

	      double vvalue = spptr->getValueDouble(zz);
	      // check for nan
	      if (!isnand(vvalue) && !isnand(y[ii])) {
		y[ii] += vvalue;
		cnt[ii]++;
	      }
	      else
		y[ii] = getnand();

	      break;
	    }
	    else
	      if (mosaic) {
		spptr = spptr->nextMosaic();
		if (spptr)
		  params = spptr->getDataParams(currentScale->scanMode());
	      }
	  }
	  while (mosaic && spptr);

	  if (!spptr)
	    y[ii] = getnand();
	}

	break;
      }
      else
	if (mosaic) {
	  sptr = sptr->nextMosaic();
	  if (sptr)
	    params = sptr->getDataParams(currentScale->scanMode());
	}
    }
    while (mosaic && sptr);

    if (!sptr)
      y[ii] = getnand();
  }

  // average if needed
  if (avg)
    for (long ii=0; ii<num; ii++)
      if (!isnand(y[ii]))
	y[ii] /= cnt[ii];

      // convert nand's to low
  for (int j=0; j<num; j++)
    if (isnand(y[j]))
      y[j] = currentScale->low();

  delete [] cnt;
}

void FrameBase::bltCut(char* xname, char* yname, 
		       Orientation axis, const Vector& rr)
{
  if (!currentFits_)
    return;

  // Vector r is in Widget coords
  Vector r = rr;

  int size;
  if (axis == XX)
    size = options->width;
  else
    size = options->height;

  long length = (size+1) * 2;
  double* x = (double*)malloc(sizeof(double)*length);
  double* y = (double*)malloc(sizeof(double)*length);

  if (*currentMosaicCount == 0 && *currentSliceCount == 0) {
    for (int i=0; i<=size; i++) {
      x[i*2] = i;
      x[i*2+1] = i;
      y[i*2] = 0;
      y[i*2+1] = 0;
    }
  }
  else
    bltCutFits(x, y, size, r, axis);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK)
    goto error;

  if (Blt_ResetVector(yy, y, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  return;

 error:
    result = TCL_ERROR;
    return;
}

void FrameBase::bltCutFits(double* x, double* y, int size, Vector& r, 
			   Orientation axis)
{
  // basics
  FitsImage* sptr = currentFits_;
  int mosaic = isMosaic();

  // variable
  FitsBound* params = sptr->getDataParams(currentScale->scanMode());
  double prev = currentScale->low();

  // main loop
  for (int ii=0; ii<=size; ii++) {
    double v = currentScale->low();

    if (mosaic) {
      sptr = currentFits_;
      params = sptr->getDataParams(currentScale->scanMode());
    }

    do {
      Vector img;
      if (axis == XX)
	img = Vector(1+ii,r[1]) * sptr->getWidgetToData();
      else
	img = Vector(r[0],1+ii) * sptr->getWidgetToData();

      if (img[0]>=params->xmin && img[0]<params->xmax && 
	  img[1]>=params->ymin && img[1]<params->ymax) {
	double value = sptr->getValueDouble(img);

	if (!isnand(value))
	  v = value;

	break;
      }
      else {
	if (mosaic) {
	  sptr = sptr->nextMosaic();
	  if (sptr)
	    params = sptr->getDataParams(currentScale->scanMode());
	}
      }
    }
    while (mosaic && sptr);

    x[2*ii] = ii;
    x[2*ii +1] = ii;
    
    y[2*ii] = prev;
    y[2*ii +1] = v;
    prev = v;
  }
}

#ifdef USE_FUNTOOLS_FILTER
// NB: this code is left as an example of filtering, its not fully debugged!!!
#include "fitsy.h"

#define FPRINTF(x) fprintf x
#define XPRINTF(x)

// can't include filter.h, because it conflicts with defs in base.h
typedef struct filtmask {
  int region;
  int y;
  int xstart, xstop;
} *FilterMask, FilterMaskRec;

extern "C" {
void *FilterOpen(void  *fitsy, char *filtstr, char *mode);
void *FilterClose(void *filter);
int   FilterImage(void *filter,
		  int x0, int x1, int y0, int y1, int block,
		  FilterMask *masks, int *nreg);
}

char *FrameBase::proj2Box(const Vector& p1, const Vector& p2, int h, 
			  char *s, int l)
{
  int type;
  double minsize=1.0;
  double xwidth, yheight;
  double xcen, ycen;
  double slope;
  double angl, dangl;
  double half_width, half_height;
  double cosangl, sinangl;
  double hw_cos, hw_sin;
  double hh_cos, hh_sin;
  Vector tp1=p1;
  Vector tp2=p2;

  // start out pessimistically
  *s = '\0';

  // sanity check on valid width and height
  xwidth = sqrt((tp1[0]-tp2[0]) * (tp1[0]-tp2[0]) + 
		(tp1[1]-tp2[1]) * (tp1[1]-tp2[1]));
  yheight = (double)h;
  if( xwidth < minsize )  xwidth  = minsize;
  if( yheight < minsize ) yheight = minsize;

  // the slope is the tangent of the angle we need
  slope = (double)(tp2[1] - tp1[1])/(double)(tp2[0] - tp1[0]);
  if( tp2[0] != tp1[0] )
    angl = atan(slope);
  else
    angl = M_PI_2;
  while(angl >= (M_PI*2.0)) angl -= (M_PI*2.0);

  // the region will output the angle in degrees
  dangl = angl / M_PI * 180.0;

  // some useful variables
  sinangl = sin(angl);
  cosangl = cos(angl);
  half_width = xwidth / 2.0;
  half_height = yheight / 2.0;
  hw_cos = half_width * cosangl;
  hw_sin = half_width * sinangl;
  hh_cos = half_height * cosangl;
  hh_sin = half_height * sinangl;

  // set type based on angle and y direction
  // (dunno why this is so confusing!)
  if( (dangl > 0) ){
    if( tp2[1] > tp1[1] ){
      type = 1;
    }
    else{
      type = 3;
    }
  }
  else{
    if( tp2[1] > tp1[1] ){
      type = 3;
    }
    else{
      type = 1;
    }
  }

  // determine center
  switch(type){
  case 1:
    xcen = tp1[0] + hw_cos - hh_sin;
    ycen = tp1[1] + hh_cos + hw_sin;
    break;
  case 2:
    xcen = tp1[0] + hw_cos + hh_sin;
    ycen = tp1[1] - hh_cos + hw_sin;
    break;
  case 3:
    xcen = tp1[0] - hw_cos + hh_sin;
    ycen = tp1[1] - hh_cos - hw_sin;
    break;
  case 4:
    xcen = tp1[0] - hw_cos - hh_sin;
    ycen = tp1[1] + hh_cos - hw_sin;
    break;
  }

  // compose the box region
  snprintf(s, l-1, "image;box(%.2f %.2f %.2f %.2f %.2f)\n", 
	   xcen+1, ycen, xwidth+1, yheight, dangl);

  // return the region
  if( !s || !*s )
    return NULL;
  else
    return s;
}

int FrameBase::createFilter(FitsFile* fits, char *filtstr, char *mode)
{
  FitsHead* srcHead = fits->head();

  if (filtstr && *filtstr) {
    if( !(fitsy = ft_headinit(srcHead->cards(), srcHead->headbytes())) ){
      cerr << "Error: Fitsy++ internal error- bad filter head" << endl;
      return 0;
    } else {
      if( !(filter = FilterOpen(fitsy, filtstr, mode)) ){
	cerr << "Error: Fitsy++ internal error- unable to build filter"<<endl;
	return 0;
      }
    }
    return 1;
  } else{
    return 0;
  }
}

void FrameBase::deleteFilter()
{
  if (filter) {
    FilterClose(filter);
    filter = NULL;
  }
  if (fitsy) {
    ft_headfree((FITSHead)fitsy,0);
    fitsy = NULL;
  }
}

void FrameBase::bltProjection2(char* xname, char* yname, 
			      char* xcname, char* ycname,
			      CoordSystem sys, SkyFrame sky,
			      const Vector& p1, const Vector& p2, 
			      double width, int avg)
{
  if (!currentFits_)
    return;

  int num = (p2-p1).length() +1;

  double* x  = (double*)calloc(num, sizeof(double));
  double* y  = (double*)calloc(num, sizeof(double));
  double* xc = (double*)calloc(num, sizeof(double));
  double* yc = (double*)calloc(num, sizeof(double));

  num = bltProjFits2(x, y, xc, yc, num, sys, sky, p1, p2, width, avg);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, y, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* xxc;
  if (Blt_GetVector(interp, xcname, &xxc) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xxc, xc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yyc;
  if (Blt_GetVector(interp, ycname, &yyc) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yyc, yc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  return;

 error:
  result = TCL_ERROR;
  return;
}

// NB: mosaic not implemented!!!
int FrameBase::bltProjFits2(double* x, double* y, 
			    double* xc, double* yc, int num,
			    CoordSystem sys, SkyFrame sky,
			    const Vector& p1, const Vector& p2, 
			    double width, int avg)
{
  // basics
  int mosaic = isMosaic();
  int *cnt = new int[num];
  int ii, hii=0;
  int xmin, xmax, ymin, ymax, block, nmask, plen;
  char filtstr[SZ_LINE];
  FitsImage* sptr = currentFits_;
  FitsFile *fits = sptr->imageFile();
  FitsBound* params = sptr->getDataParams(currentScale->scanMode());
  FilterMask masks=NULL;

  // prepare for determining the projection array offset
  Vector tp1=p1;
  Vector tp2=p2;
  double Ax=tp1[0];
  double Ay=tp1[1];
  // translate B so that A is on the origin
  double Bx=tp2[0]-Ax;
  double By=tp2[1]-Ay;
  // now rotate B onto positive X axis
  double distAB=sqrt(Bx*Bx+By*By);
  double theCos=Bx/distAB;
  double theSin=By/distAB;

  // convert the projection region to a box region
  if( !proj2Box(p1, p2, width, filtstr, SZ_LINE) )  return 0;
  XPRINTF((stderr, "filter: %s\n", filtstr));
  // open an image box filter
  if( !createFilter(fits, filtstr, "type=image") ) return 0;

  // set min and max for the filter call
  xmin = 1;
  xmax = params->xmax+1;
  ymin = 1;
  ymax = params->ymax+1;
  block = 1;
  plen = num;

  // seed the x array
  for(int i=0; i<num; i++){
    x[i] = i+1;
    cnt[i] = 0;
  }

  // get region segments for the current filter
  nmask = FilterImage(filter, xmin, xmax, ymin, ymax, block, &masks, NULL);
  for(int i=0; i<nmask; i++){
    for(int j=masks[i].xstart; j<masks[i].xstop; j++){
      // convert position to 0-based system
      int ix = j-1;
      int iy = masks[i].y-1;
      // translate x, y so that A is on origin
      double Cx = ix-Ax;
      double Cy = iy-Ay;
      // find x intercept in rotated cooridinate system
      // (remember that line seg from x,y to AB is orthogonal to AB)
      double dx = Cx*theCos + Cy*theSin;
      // if( dx < 0 ) continue;
      // the x intercept is the index of the projection
      ii=(int)dx;
      // get data value and store in projection arrays
      // x[ii] = ii+1;
      XPRINTF((stderr, "%d,%d -> %.2f,%.2f -> %.2f -> %d (%d)\n", ix, iy, Cx, Cy, dx, ii, plen));
      Vector t = Vector(ix,iy);

      Vector tv = sptr->mapFromRef(t,sys,sky);
      xc[ii] = tv[0];
      yc[ii] = tv[1];

      Vector z = t * sptr->getRefToData();
      if (z[0]>=params->xmin && z[0]<params->xmax && 
	  z[1]>=params->ymin && z[1]<params->ymax) {
	// get pixel value
	double v = sptr->getValueDouble(z);
	// store value in projection array (checking for nan)
	if (!isnand(v)) {
	  y[ii] += v;
	  hii = max(hii,ii);
	  XPRINTF((stderr, "yy[%d] = %.2f (%.2f %.2f)\n", ii, y[ii], v, x[ii]));
	  cnt[ii] += 1;
	}
	else{
	  y[ii] = getnand();
	}
      } else {
	XPRINTF((stderr, "%.2f,%.2f out of bounds: %d-> %d %d->%dn", 
		 z[0], z[1], 
		 params->xmin, params->xmax, params->ymin, params->ymax));
      }
    }
  }

  // average if needed
  if (avg){
    for (long ii=0; ii<num; ii++){
      if (!isnand(y[ii])){
	y[ii] /= cnt[ii];
      }
    }
  }

  // convert nand's to low
  for (int j=0; j<num; j++){
    if (isnand(y[j]))
      y[j] = currentScale->low();
  }

  // clean up
  delete [] cnt;
  if( masks ) free(masks);
  deleteFilter();

  // return length (i.e. largest array element)
  return hii;
}

#endif
