// Copyright (C) 1999-2009
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "truecolor24.h"
#include "util.h"

#ifdef _MACOSX
extern void macosxGetMasks(unsigned long*, unsigned long*, 
			   unsigned long*, unsigned long*);
#endif

TrueColor24::TrueColor24(Visual* visual)
{
#if !(_WIN32 || _MACOSX)
  rs_ = decodeMask((unsigned long)visual->red_mask);
  gs_ = decodeMask((unsigned long)visual->green_mask);
  bs_ = decodeMask((unsigned long)visual->blue_mask);
  as_ = 0;
#endif
#if _MACOSX
  // tk visual masks are bogus
  unsigned long red_mask, green_mask, blue_mask, alpha_mask;
  macosxGetMasks(&red_mask, &green_mask, &blue_mask, &alpha_mask);

  rs_ = decodeMask((unsigned long)red_mask);
  gs_ = decodeMask((unsigned long)green_mask);
  bs_ = decodeMask((unsigned long)blue_mask);
  as_ = decodeMask((unsigned long)alpha_mask);
#endif
#if _WIN32
  // windows masks are plain wrong
  rs_ = decodeMask((unsigned long)0x00FF0000);
  gs_ = decodeMask((unsigned long)0x0000FF00);
  bs_ = decodeMask((unsigned long)0x000000FF);
  as_ = 0;
#endif
}

int TrueColor24::decodeMask(unsigned long mask)
{
  switch (mask) {
  case 0xff:
    return 0;
  case 0xff00:
    return 8;
  case 0xff0000:
    return 16;
  case 0xff000000:
    return 24;
  }
}

void TrueColor24::encodeTrueColor(XColor* src, char* dest, XImage* ximage)
{
  if (!ximage)
    return;

  switch (ximage->bits_per_pixel) {
  case 24:
    encodeTrueColor24(src, dest, ximage);
    break;
  case 32:
    encodeTrueColor32(src, dest, ximage);
    break;
  }
}

void TrueColor24::encodeTrueColor24(XColor* src, char* dest, XImage* ximage)
{
  int msb = ximage->byte_order;

  unsigned int r = (unsigned char)src->red;
  unsigned int g = (unsigned char)src->green;
  unsigned int b = (unsigned char)src->blue;
  unsigned int v = 0;
  v |= r << rs_;
  v |= g << gs_;
  v |= b << bs_;

  // we need to check to byteswap when we have cross platforms
  if ((!msb && lsb()) || (msb && !lsb()))
    memcpy(dest, &v, 3);
  else {
    unsigned char* rr = (unsigned char*)(&v);
    *(dest) = *(rr+3);
    *(dest+1) = *(rr+2);
    *(dest+2) = *(rr+1);
  }
}

void TrueColor24::encodeTrueColor32(XColor* src, char* dest, XImage* ximage)
{
  int msb = ximage->byte_order;

  unsigned int r = (unsigned char)src->red;
  unsigned int g = (unsigned char)src->green;
  unsigned int b = (unsigned char)src->blue;
  unsigned int v = 0;
  v |= r << rs_;
  v |= g << gs_;
  v |= b << bs_;

  // we need to check to byteswap when we have cross platforms
  if ((!msb && lsb()) || (msb && !lsb()))
    memcpy(dest, &v, 4);
  else {
    unsigned char* rr = (unsigned char*)(&v);
    *(dest) = *(rr+3);
    *(dest+1) = *(rr+2);
    *(dest+2) = *(rr+1);
    *(dest+3) = *(rr);
  }
}

void TrueColor24::encodeTrueColor(unsigned char* src, XImage* ximage)
{
  if (!ximage)
    return;

  switch (ximage->bits_per_pixel) {
  case 24:
    encodeTrueColor24(src, ximage);
    break;
  case 32:
    encodeTrueColor32(src, ximage);
    break;
  }
}

void TrueColor24::encodeTrueColor24(unsigned char* src, XImage* ximage)
{
  int& width = ximage->width;
  int& height = ximage->height;
  char* data = XImageData(ximage);
  int bytesPerPixel = ximage->bits_per_pixel/8;
  int msb = ximage->byte_order;

  const unsigned char* ptr = src;
  if ((!msb && lsb()) || (msb && !lsb())) {
    for (int jj=0; jj<height; jj++) {
      char* dest = data + jj*ximage->bytes_per_line;

      for (int ii=0; ii<width; ii++, dest+=bytesPerPixel, ptr+=3) {
	unsigned int r = ptr[0];
	unsigned int g = ptr[1];
	unsigned int b = ptr[2];
	unsigned int v = 0;
	v |= r << rs_;
	v |= g << gs_;
	v |= b << bs_;

	memcpy(dest, &v, 3);
      }
    }
  }
  else {
    for (int jj=0; jj<height; jj++) {
      char* dest = data + jj*ximage->bytes_per_line;

      for (int ii=0; ii<width; ii++, dest+=bytesPerPixel, ptr+=3) {
	unsigned int r = ptr[0];
	unsigned int g = ptr[1];
	unsigned int b = ptr[2];
	unsigned int v = 0;
	v |= r << rs_;
	v |= g << gs_;
	v |= b << bs_;

	unsigned char* rr = (unsigned char*)(&v);
	*(dest) = *(rr+3);
	*(dest+1) = *(rr+2);
	*(dest+2) = *(rr+1);
      }
    }
  }
}

void TrueColor24::encodeTrueColor32(unsigned char* src, XImage* ximage)
{
  int& width = ximage->width;
  int& height = ximage->height;
  char* data = XImageData(ximage);
  int bytesPerPixel = ximage->bits_per_pixel/8;
  int msb = ximage->byte_order;

  const unsigned char* ptr = src;
  if ((!msb && lsb()) || (msb && !lsb())) {
    for (int jj=0; jj<height; jj++) {
      // the line may be padded at the end
      char* dest = data + jj*ximage->bytes_per_line;

      for (int ii=0; ii<width; ii++, dest+=bytesPerPixel, ptr+=3) {
	unsigned int r = ptr[0];
	unsigned int g = ptr[1];
	unsigned int b = ptr[2];
	unsigned int v = 0;
#ifdef _MACOSX
	unsigned int a = 0xff;
	v |= a << as_;
#endif
	v |= r << rs_;
	v |= g << gs_;
	v |= b << bs_;

	memcpy(dest, &v, 4);
      }
    }
  }
  else {
    for (int jj=0; jj<height; jj++) {
      // the line may be padded at the end
      char* dest = data + jj*ximage->bytes_per_line;

      for (int ii=0; ii<width; ii++, dest+=bytesPerPixel, ptr+=3) {
	unsigned int r = ptr[0];
	unsigned int g = ptr[1];
	unsigned int b = ptr[2];
	unsigned int v = 0;
#ifdef _MACOSX
	unsigned int a = 0xff;
	v |= a << as_;
#endif
	v |= r << rs_;
	v |= g << gs_;
	v |= b << bs_;

	unsigned char* rr = (unsigned char*)(&v);
	*(dest) = *(rr+3);
	*(dest+1) = *(rr+2);
	*(dest+2) = *(rr+1);
	*(dest+3) = *(rr);
      }
    }
  }
}
